/*
 * @(#)DrawApplication.java 5.1
 *
 */

package CH.ifa.draw.application;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.contrib.*;
import CH.ifa.draw.util.*;

// for class DrawingLoadServer:
import java.io.*;
import java.net.*;
import java.util.*;

import de.renew.util.StringUtil;

/**
 * DrawApplication defines a standard presentation for
 * standalone drawing editors. The presentation is
 * customized in subclasses.
 * The application is started as follows:
 * <pre>
 * public static void main(String[] args) {
 *	   MyDrawApp window = new MyDrawApp();
 * }
 * </pre>
 */

public class DrawApplication
	implements DrawingEditor, PaletteListener, AutosaveSaver {

	private static int windowCount = 0;

	protected Frame       menuFrame, toolFrame;
	private Panel         toolPanel;
	private final String  defaultDrawExtension="draw";
	public String         drawExtension;
	private Font          fMenuFont;
	private boolean       floating;
	private Vector        fDrawings = new Vector();
	private Tool          fTool;
	private Iconkit       fIconkit;
	private WindowsMenu   fDrawingsMenu;
	private TextField     fStatusLine;
	private Vector        fViewFrames = new Vector();
	private DrawingViewFrame fViewFrame;
	private boolean       fToolButtonSticky=false;
	private boolean       fAlwaysSticky=false;
	private ToolButton    fDefaultToolButton;
	private ToolButton    fSelectedToolButton = null;
	private ToolButton    fConnTextTB, fTextTB;

	// the image resource path
	private static final String fgDrawPath = "/CH/ifa/draw/";
	public static final String IMAGES = fgDrawPath+"images/";

	/**
	 * The index of the file menu in the menu bar.
	 */
	public static final int	   FILE_MENU = 0;
	/**
	 * The index of the edit menu in the menu bar.
	 */
	public static final int	   EDIT_MENU = 1;
	/**
	 * The index of the alignment menu in the menu bar.
	 */
	public static final int	   LAYOUT_MENU = 2;
	/**
	 * The index of the attributes menu in the menu bar.
	 */
	public static final int	   ATTRIBUTES_MENU = 3;

	/**
	 * The index of the drawings menu in the menu bar.
	 */
	public static final int	   DRAWINGS_MENU = 5;

	private String fLastPath = System.getProperty("user.dir");

	/** Lookup table for default directories keyed by file extensions. **/
	private Hashtable dirByExt = new Hashtable();

	/** Default directory used if the lookup table has no entry for some extension. **/
	private String fallbackDir = null;

	/**
	 * The minimum time interval that has to elapse before a
	 * drawing is saved automatically.
	 */
	private static final int AUTOSAVE_INTERVAL = 120000;

	/**
	 * The autosave manager that is associated to this
	 * DrawingEditor.
	 */
	private AutosaveManager autosaveManager =
		new AutosaveManager(this,AUTOSAVE_INTERVAL);


	/**
	 * The undo and redo history manager which keeps
	 * track of undo/redo snapshots.
	 **/
	private UndoRedoManager undoManager = new UndoRedoManager(this);

	/**
	 * Constructs a drawing window with a default title.
	 */
	public DrawApplication() {
		this("JHotDraw");
	}

	/**
	 * Constructs a drawing window with the given title.
	 */
	public DrawApplication(String title) {
		this(title, null, new String[0], null);
	}

	public DrawApplication(String title, String[] drawings) {
		this(title,null,drawings,null);
	}

  protected Component wrapStatusLine(Component statusLine) {
		return statusLine;
  }

	public DrawApplication(String title, String drawExtension, String[] drawings, String icon) {

		DrawingLoadServer dls=null;
		int loadServerPort=GUIProperties.loadServerPort();
		if (loadServerPort != -1) {
		  try {
				dls=new DrawingLoadServer(this, loadServerPort);
		  } catch (IOException ioe) {
				System.out.println("Server Socket is occupied!");
				System.exit(-1);
		  }
		} /* endif */

		menuFrame = new Frame(title);
		fStatusLine = createStatusLine();
		menuFrame.add(wrapStatusLine(fStatusLine),"South");

//		toolFrame = new Frame("Tools");
//		toolFrame.setLayout(new BorderLayout());
		toolFrame = menuFrame;
		floating = toolFrame == menuFrame;
		fIconkit = new Iconkit(toolFrame);
		if (icon != null) {
			Image iconImage = fIconkit.loadImageResource(icon);
			if (iconImage == null) {
				System.err.println("Resource icon "+icon+" could not be loaded!");
			} else {
				menuFrame.setIconImage(iconImage);
			}
		}

		toolPanel = new Panel();
		toolPanel.setBackground(Color.lightGray);
		toolPanel.setLayout(new DynamicFlowLayout(FlowLayout.LEFT,2,0));
		toolFrame.add(toolPanel,"Center");

		createTools(toolPanel);

		MenuBar mb = new MenuBar();
		fMenuFont=mb.getFont();
		int menuFontSize = GUIProperties.menuFontSize();
		if (menuFontSize != -1) {
			System.out.println("Setting menu font size to "+menuFontSize+" pt.");
			fMenuFont=Fontkit.getFont("SansSerif",Font.PLAIN,menuFontSize);
		}
		createMenus(mb);
		menuFrame.setMenuBar(mb);

		menuFrame.pack();

		// Workaround for Windows JDK pack()-bug:
		// The menu-frame is too small or too large.
		//menuFrame.setResizable(GUIProperties.windowResizable());

		menuFrame.show();
		if (! GUIProperties.avoidFrameReshape()) {
		  // pack() has to be done after show() on Windows-Systems.
		  // However, as it would involve a call to setBounds(),
		  // avoid it when frame positioning is not wanted.
		  menuFrame.pack();
		} 
		
		// This should show the main window. However, there were some problems
		// in the very early versions of Renew, where the appearance of the
		// main window would depend crucially on the order of some
		// commands. It is not clear whether the errors can occur again.
		// This must be monitored.

		if (toolFrame != menuFrame) {
			toolFrame.pack();
			toolFrame.show();
		} /* endif */

		windowCount++;
		addListeners();

		if (drawExtension==null)
			this.drawExtension=defaultDrawExtension;
		else	
			this.drawExtension=drawExtension;

		//System.out.println("number of params: "+drawings.length);
		Drawing drawing;
		for (int i=0; i<drawings.length; ++i)
			loadAndOpenDrawing(drawings[i]);
		if (fDrawings.size() == 0)
			promptNew();

		toolDone();

		if (dls!=null) {
		  dls.start();
		}
	}

	/**
	 * Registers the listeners for the main windows
	 */
	protected void addListeners() {
		menuFrame.addWindowListener(
			new WindowAdapter() {
					public void windowClosing(WindowEvent event) {
						exit();
					}
				}
			);
		if (toolFrame != menuFrame) {
			toolFrame.addWindowListener(
				new WindowAdapter() {
						public void windowClosing(WindowEvent event) {
							toolFrame.setVisible(false);
						}
					}
				);
		} /* endif */
		menuFrame.addComponentListener(
			new ComponentAdapter() {
					public void componentResized(ComponentEvent event) {
						if (! GUIProperties.avoidFrameReshape()) {
							Dimension dim=menuFrame.getPreferredSize();
							Dimension is=menuFrame.getSize();
							menuFrame.setSize(Math.max(dim.width,is.width),dim.height); 
						}				  
						//menuFrame.pack();
					}
				}
			);
	}

	public Frame menuFrame() { return menuFrame; }

	protected DrawApplication newWindow() { return new DrawApplication(menuFrame.getTitle()); }

//	   private Menu titleSubstitute;

//	   public synchronized void setTitle(String title) {
//		 menuFrame.setTitle(title);
//		 titleSubstitute.setLabel(title);
//	   }

	private void updateName(Drawing drawing, String filename) {
	  showDrawingViewFrame(drawing);
	  MenuItem item = fDrawingsMenu.getItemForDrawing(drawing);
		fDrawingsMenu.removeDrawing(drawing);
	  filename=StringUtil.makeCanonical(filename);
	  fLastPath = StringUtil.getPath(filename);
	  String nameOnly = StringUtil.getFilename(filename);
	  fViewFrame.setTitle(nameOnly);
	  drawing.setName(nameOnly);
	  drawing.setFilename(filename);
		item.setLabel(nameOnly);
		fDrawingsMenu.addDrawing(drawing, item);
	}

	/**
	 * Creates the standard menus. Clients override this
	 * method to add additional menus.
	 */
	protected void createMenus(MenuBar mb) {
		mb.add(createFileMenu());
		mb.add(createEditMenu());
		mb.add(createAlignmentMenu());
		mb.add(createAttributesMenu());
		mb.add(createDebugMenu());
		mb.add(createDrawingsMenu());
	}

	void addDrawing(Drawing drawing) {
		if (!fDrawings.contains(drawing)) {
			fDrawings.addElement(drawing);
			CheckboxMenuItem mi = new CheckboxMenuItem(drawing.getName());
			mi.setFont(fMenuFont);
			mi.addItemListener(new DrawingMenuListener(this,drawing));
			fDrawingsMenu.addDrawing(drawing, mi);

			// initialize undo history management for the new drawing
			undoManager.newUndoHistory(drawing);

			drawingAdded(drawing);
		} /* endif */
	}

	protected boolean closeViewFrame(DrawingViewFrame viewFrame) {
	  Drawing drawing=viewFrame.view().drawing();

	  if (drawing.isModified()) {
			showDrawingViewFrame(drawing);
		  int answer = MessageDialog.invokeDialog(
				menuFrame, "Renew: Confirm Close",
				new String[] {"The drawing \""+drawing.getName()+"\"","you are about to close","has been modified."},
				MessageDialog.LEFT, new String[] {" Save now "," Close ", " Cancel "}
				);
		  if (answer == 0) {
				if (!saveDrawing(drawing))
					return false;
		  } else if (answer == 2) {
				return false;
		  }
	  }
	  viewFrame.setVisible(false);
	  viewFrame.dispose();
	  fViewFrames.removeElement(viewFrame);
	  if (viewFrame == fViewFrame) {
			int numViews=fViewFrames.size();
			if (numViews > 0) {
				setCurrentDrawing((DrawingViewFrame)fViewFrames.elementAt(numViews-1));
				fViewFrame.show();
				fViewFrame.requestFocus(); // !!!
			} else
				setCurrentDrawing(null);
	  }

	  int pos = fDrawings.indexOf(drawing);
	  fDrawings.removeElementAt(pos);
	  fDrawingsMenu.removeDrawing(drawing);

	  undoManager.removeUndoHistory(drawing);

	  drawing.release();
	  drawingReleased(drawing);

	  viewFrame.discard();
	  autosaveManager.removeDrawing(drawing);

	  return true;
	}

	protected void drawingAdded(Drawing drawing) {
	  // My subclasses will add special behaviour here.
	}

	protected void drawingReleased(Drawing drawing) {
	  // My subclasses will add special behaviour here.
	}

	protected Menu createMenu(String name) {
		Menu menu=new Menu(name);
		menu.setFont(fMenuFont);
		return menu;
	}

	public MenuItem createMenuItem(String name, MenuShortcut shortcut,
	                               ActionListener action) {
		MenuItem mi;
		if (shortcut==null)
			mi=new MenuItem(name);
		else
			mi=new MenuItem(name,shortcut);
		mi.setFont(fMenuFont);
		mi.addActionListener(action);
		return mi;
	}

	public MenuItem createMenuItem(String name, ActionListener action) {
		return createMenuItem(name,null,action);
	}

	public CommandMenu createCommandMenu(String name) {
		CommandMenu cm=new CommandMenu(name);
		cm.setFont(fMenuFont);
		return cm;
	}

	protected Menu createDrawingsMenu() {
		fDrawingsMenu = new WindowsMenu("Windows", this);
		fDrawingsMenu.setFont(fMenuFont);
		return fDrawingsMenu;
	}

	public Enumeration drawings() {
		//cloned to avoid inconsistencies in local drawing Enumerations(!)
	  return ((Vector)fDrawings.clone()).elements();
	}

	/**
	 * Creates a new drawing view and a new drawing view frame.
	 * The default size and location as implemented by the
	 * <code>DrawingViewFrame</code> constructor will be used.
	 *
	 * @param drawing  the drawing to be displayed in the view.
	 *
	 * @return the new drawing view frame
	 **/
	protected DrawingViewFrame newDrawingViewFrame(Drawing drawing) {
		return newDrawingViewFrame(drawing, null, null);
	}

	/** 
	 * Creates a new drawing view and a new drawing view frame. 
	 * 
	 * @param drawing  the drawing to be displayed in the view.
	 * @param loc	   the location of the new drawing view frame
	 *				   on the screen. 
	 *				   If <code>null</code>, the view frame
	 *				   defaults will be used.
	 * @param size	   the size of the new drawing view frame.
	 *				   If <code>null</code>, the view frame
	 *				   defaults will be used.
	 *
	 * @return the new drawing view frame
	 **/
	protected DrawingViewFrame newDrawingViewFrame(Drawing drawing, Point loc, Dimension size) {
		Dimension d = drawing.defaultSize();
		StandardDrawingView drawingView = createDrawingView(d.width, d.height);
		addDrawing(drawing);
		DrawingViewFrame viewFrame=new DrawingViewFrame(this, drawingView, drawing, loc, size);
		fViewFrames.addElement(viewFrame);
		setCurrentDrawing(viewFrame);
		return viewFrame;
	}
		  
	/**
	 * Creates the file menu. Clients override this
	 * method to add additional menu items.
	 */
	protected Menu createFileMenu() {
		Menu menu = createMenu("File");

		menu.add(createMenuItem("New Drawing", new MenuShortcut('N'),
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						promptNew();
					}
				}
			));

		//				menu.add(createMenuItem("New Frame", new MenuShortcut('F'),
		//					new ActionListener() {
		//						public void actionPerformed(ActionEvent event) {
		//						  newDrawingViewFrame(drawing());
		//						}
		//					}
		//						));

		menu.add(createMenuItem("Open Drawing...", new MenuShortcut('O'),
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						promptOpen();
					}
				}
			));

		menu.add(createMenuItem("Insert Drawing...", 
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						promptInsert();
					}
				}
			));

		menu.add(createMenuItem("Save Drawing", new MenuShortcut('S'),
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						saveDrawing(drawing());
					}
				}
			));

		menu.add(createMenuItem("Save Drawing As...",
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						Drawing drawing=drawing();
						if (promptSaveAs(drawing))
							saveDrawing(drawing);
					}
				}
			));

		menu.add(createMenuItem("Save All Drawings",
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						Enumeration drawings=drawings();
						while (drawings.hasMoreElements()) {
							Drawing drawing=(Drawing)drawings.nextElement();
							if (drawing.getFilename()==null) {
								showDrawingViewFrame(drawing);
								if (!promptSaveAs(drawing))
									return;
							}
						}
						drawings=drawings();
						while (drawings.hasMoreElements()) {
							saveDrawing((Drawing)drawings.nextElement());
						}
					}
				}
			));

		menu.add(createMenuItem("Close Drawing",
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						if (fViewFrame != null) {
							closeViewFrame(fViewFrame);
						}
					}
				}
			));

		//		menu.add(createMenuItem("Close Frame",
		//		   new ActionListener() {
		//			  public void actionPerformed(ActionEvent event) {
		//				fViewFrame.hide();
		//				   fViewFrame.dispose();
		//			  }
		//		   }
		//				));

		menu.addSeparator();

		menu.add(createMenuItem("Print...", new MenuShortcut('P'),
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						print();
					}
				}
			));

		menu.addSeparator();

		menu.add(createMenuItem("Exit",
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						exit();
					}
				}
			));

		return menu;
	}

	/**
	 * Creates the edit menu. Clients override this
	 * method to add additional menu items.
	 */
	protected Menu createEditMenu() {
		CommandMenu menu = createCommandMenu("Edit");
		menu.add(new UndoRedoCommand("Undo", this, UndoRedoCommand.UNDO), 
		         new MenuShortcut('Z'));
		menu.add(new UndoRedoCommand("Redo", this, UndoRedoCommand.REDO));
		menu.addSeparator();
		menu.add(new CutCommand("Cut", this), new MenuShortcut('X'));
		menu.add(new CopyCommand("Copy", this), new MenuShortcut('C'));
		menu.add(new PasteCommand("Paste", this), new MenuShortcut('V'));
		menu.add(new DuplicateCommand("Duplicate", this), new MenuShortcut('D'));
		menu.add(new DeleteCommand("Delete", this));
		menu.addSeparator();
		menu.add(new SearchCommand("Search...", this));
		menu.add(new ReplaceCommand("Search & Replace...", this));
		menu.addSeparator();
		menu.add(new GroupCommand("Group", this));
		menu.add(new UngroupCommand("Ungroup", this));
		menu.addSeparator();
		// The menues "Send To Back" and "Bring To Front" were
		// moved to the "Layout" (formerly "Align") menu.
		menu.add(new SelectCommand("Select All", this),
		         new MenuShortcut('A'));
		menu.add(new SelectCommand("Invert Selection", this,
															 Figure.class, SelectCommand.INVERT));
		//menu.addSeparator();
		//menu.add(new ToggleStickyCommand("Toggle Sticky Tools", this));

		return menu;
	}

	/**
	 * Creates the alignment menu. Clients override this
	 * method to add additional menu items.
	 */
	protected Menu createAlignmentMenu() {
		CommandMenu menu = createCommandMenu("Layout"); // "Align"
		menu.add(new ToggleGridCommand("Toggle Snap to Grid",
		                               this, new Point(5,5)));
		menu.add(new SnapToGridCommand("Snap to Grid now",
		                               this, new Point(5,5)));
		menu.addSeparator();
		CommandMenu alignMenu = createCommandMenu("Align");
		alignMenu.add(new AlignCommand("Lefts", this, AlignCommand.LEFTS));
		alignMenu.add(new AlignCommand("Centers", this, AlignCommand.CENTERS));
		alignMenu.add(new AlignCommand("Rights", this, AlignCommand.RIGHTS));
		alignMenu.addSeparator();
		alignMenu.add(new AlignCommand("Tops", this, AlignCommand.TOPS));
		alignMenu.add(new AlignCommand("Middles", this, AlignCommand.MIDDLES));
		alignMenu.add(new AlignCommand("Bottoms", this, AlignCommand.BOTTOMS));
		menu.add(alignMenu);
		CommandMenu spreadMenu = createCommandMenu("Spread");
		spreadMenu.add(new SpreadCommand("Lefts", this, SpreadCommand.LEFTS));
		spreadMenu.add(new SpreadCommand("Centers", this, SpreadCommand.CENTERS));
		spreadMenu.add(new SpreadCommand("Rights", this, SpreadCommand.RIGHTS));
		spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.HORIZONTAL_DISTANCE));
		spreadMenu.addSeparator();
		spreadMenu.add(new SpreadCommand("Tops", this, SpreadCommand.TOPS));
		spreadMenu.add(new SpreadCommand("Middles", this, SpreadCommand.MIDDLES));
		spreadMenu.add(new SpreadCommand("Bottoms", this, SpreadCommand.BOTTOMS));
		spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.VERTICAL_DISTANCE));

		spreadMenu.addSeparator();
		spreadMenu.add(new SpreadCommand("Diagonal", this, SpreadCommand.DIAGONAL_CENTERS));
		//spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.DIAGONAL_DISTANCE));
		
		menu.add(spreadMenu);
		// the following menus used to be in "Edit":
		menu.addSeparator();
		menu.add(new SendToBackCommand("Send to Back", this));
		menu.add(new BringToFrontCommand("Bring to Front", this));
		return menu;
	}

	/**
	 * Creates the debug menu. Clients override this
	 * method to add additional menu items.
	 */
	protected Menu createDebugMenu() {
		Menu menu = createMenu("Debug");

		menu.add(createMenuItem("Simple Update",
		  new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						view().setDisplayUpdate(new SimpleUpdateStrategy());
					}
				}
			));

		menu.add(createMenuItem("Buffered Update",
			new ActionListener() {
					public void actionPerformed(ActionEvent event) {
						view().setDisplayUpdate(new BufferedUpdateStrategy());
					}
				}
			));
		return menu;
	}

	/**
	 * Creates the attributes menu and its submenus. Clients override this
	 * method to add additional menu items.
	 */
	protected Menu createAttributesMenu() {
		Menu menu = createCommandMenu("Attributes");
		menu.add(createColorMenu("Fill Color", "FillColor"));
		menu.add(createColorMenu("Pen Color", "FrameColor"));
		menu.add(createVisibilityMenu());
		menu.add(createArrowMenu());
		menu.add(createLineStyleMenu());
		menu.add(createLineShapeMenu());
		menu.add(createArcScaleMenu());

		menu.addSeparator();

		menu.add(createFontMenu());
		menu.add(createFontSizeMenu());
		menu.add(createFontStyleMenu());
		menu.add(createTextAlignmentMenu());
		menu.add(createColorMenu("Text Color", "TextColor"));
		return menu;
	}

	/**
	 * Creates the color menu.
	 */
	protected Menu createColorMenu(String title, String attribute) {
		CommandMenu menu = createCommandMenu(title);
		for (int i=0; i<ColorMap.size(); i++)
			menu.add(
				new ChangeAttributeCommand(
					ColorMap.name(i),
					attribute,
					ColorMap.color(i),
					this
				)
			);
		menu.add(new QueryAttributeCommand(
			"Color by Numeric Value",
			"other...",
			attribute,
			Color.class, this));
		return menu;
	}

	/**
	 * Creates the visibility menu.
	 */
	protected Menu createVisibilityMenu() {
		CommandMenu menu = createCommandMenu("Visibility");
		menu.add(new ChangeAttributeCommand("visible", "Visibility",
		                                    Boolean.TRUE, this));
		menu.add(new ChangeAttributeCommand("invisible", "Visibility",
		                                    Boolean.FALSE, this));
		return menu;
	}

	/**
	 * Creates the arrows menu.
	 */
	protected Menu createArrowMenu() {
		CommandMenu menu = createCommandMenu("Arrow");
		menu.add(new ChangeAttributeCommand("none",		"ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_NONE),  this));
		menu.add(new ChangeAttributeCommand("at Start", "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_START), this));
		menu.add(new ChangeAttributeCommand("at End",	"ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_END),	  this));
		menu.add(new ChangeAttributeCommand("at Both",	"ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_BOTH),  this));
		return menu;
	}

	/**
	 * Creates the line style menu.
	 */
	protected Menu createLineStyleMenu() {
		CommandMenu menu = createCommandMenu("Line Style");
		menu.add(new ChangeAttributeCommand("normal", "LineStyle", PolyLineFigure.LINE_STYLE_NORMAL,  this));
		menu.add(new ChangeAttributeCommand("dotted", "LineStyle", PolyLineFigure.LINE_STYLE_DOTTED,  this));
		menu.add(new ChangeAttributeCommand("dashed", "LineStyle", PolyLineFigure.LINE_STYLE_DASHED,  this));
		menu.add(new ChangeAttributeCommand("medium dashed", "LineStyle", PolyLineFigure.LINE_STYLE_MEDIUM_DASHED,	this));
		menu.add(new ChangeAttributeCommand("long dashed", "LineStyle", PolyLineFigure.LINE_STYLE_LONG_DASHED,	this));
		menu.add(new ChangeAttributeCommand("dash-dotted", "LineStyle", PolyLineFigure.LINE_STYLE_DASH_DOTTED,	this));
		menu.add(new QueryAttributeCommand("Line style (sequence of dash/gap lengths)",
		                                   "other...", "LineStyle", String.class, this));
		return menu;
	}

	/**
	 * Creates the line shape menu.
	 */
	protected Menu createLineShapeMenu() {
		CommandMenu menu = createCommandMenu("Line Shape");
		menu.add(new ChangeAttributeCommand("straight", "LineShape", new Integer(PolyLineFigure.LINE_SHAPE),  this));
		menu.add(createBSplineMenu());
		return menu;
	}

	/**
	 * Creates the B-Spline submenu
	 */
	protected Menu createBSplineMenu() {
		CommandMenu menu = createCommandMenu("B-Spline");
		menu.add(new SplineAttributeCommand("standard", "standard", 0, this));
		menu.add(createBSplineSegmentsMenu());
		menu.add(createBSplineDegreeMenu());
		return menu;
	}

	protected Menu createBSplineSegmentsMenu() {
		CommandMenu menu = createCommandMenu("Segments");
		for (int i = 5; i < 46; i+=5) {
			menu.add(
				new SplineAttributeCommand(
					Integer.toString(i),
					"BSplineSegments",
					i, this)
			);
		}
		return menu;
	}

	protected Menu createBSplineDegreeMenu() {
		CommandMenu menu = createCommandMenu("Degree");
		for (int i = 2; i < 10; i++) {
			menu.add(
				new SplineAttributeCommand(
					Integer.toString(i),
					"BSplineDegree",
					i, this)
			);
		}
		return menu;
	}

	protected Menu createArcScaleMenu() {
		CommandMenu menu = createCommandMenu("Round corners");
		menu.add(new ChangeAttributeCommand("scale with size", 
		                                    RoundRectangleFigure.ARC_SCALE_ATTR, 
		                                    Boolean.TRUE, 
		                                    this));
		menu.add(new ChangeAttributeCommand("fixed radius", 
		                                    RoundRectangleFigure.ARC_SCALE_ATTR, 
		                                    Boolean.FALSE, 
		                                    this));
		return menu;
	}
	
	/**
	 * Creates the fonts menus. It installs all available fonts
	 * supported by the toolkit implementation.
	 */
	protected Menu createFontMenu() {
		CommandMenu menu = createCommandMenu("Font");
		String fonts[] = Toolkit.getDefaultToolkit().getFontList();
		for (int i = 0; i < fonts.length; i++) {
			menu.add(new ChangeAttributeCommand(fonts[i], "FontName", fonts[i],	 this));
		}
		menu.add(new QueryAttributeCommand(
			"Font Name",
			"other...",
			"FontName",
			String.class, this));
		return menu;
	}

	/**
	 * Creates the font style menu with entries (Plain, Italic, Bold).
	 */
	protected Menu createFontStyleMenu() {
		CommandMenu menu = createCommandMenu("Font Style");
		menu.add(new ChangeAttributeCommand("Plain", "FontStyle", new Integer(Font.PLAIN), this));
		menu.add(new ChangeAttributeCommand("Italic","FontStyle", new Integer(Font.ITALIC),this));
		menu.add(new ChangeAttributeCommand("Bold",	 "FontStyle", new Integer(Font.BOLD),  this));
		menu.add(new ChangeAttributeCommand("Underlined", "FontStyle", new Integer(ExtendedFont.UNDERLINED),  this));
		return menu;
	}

	/**
	 * Creates the font size menu.
	 */
	protected Menu createFontSizeMenu() {
		CommandMenu menu = createCommandMenu("Font Size");
		int sizes[] = { 9, 10, 11, 12, 14, 18, 24, 36, 48, 72 };
		for (int i = 0; i < sizes.length; i++) {
			menu.add(
				new ChangeAttributeCommand(
					Integer.toString(sizes[i]),
					"FontSize",
					new Integer(sizes[i]),	this)
			);
		}
		menu.add(new QueryAttributeCommand(
			"Font Size",
			"other...",
			"FontSize",
			Integer.class, this));
		return menu;
	}

	protected Menu createTextAlignmentMenu() {
		CommandMenu menu = createCommandMenu("Text Alignment");
		menu.add(new ChangeAttributeCommand("Left", 
		                                    TextFigure.ALIGN_ATTR,
		                                    new Integer(TextFigure.LEFT), 
		                                    this));
		menu.add(new ChangeAttributeCommand("Center",
		                                    TextFigure.ALIGN_ATTR,
		                                    new Integer(TextFigure.CENTER),
		                                    this));
		menu.add(new ChangeAttributeCommand("Right",
		                                    TextFigure.ALIGN_ATTR,
		                                    new Integer(TextFigure.RIGHT),
		                                    this));
		return menu;
	}


	/**
	 * Creates the tools. By default all standard figure creation tools are added.
	 * Override this method to remove the standard tools.
	 * Call the inherited method to include the standard tools.
	 * @param toolPanel the panel where the tool palette is added.
	 */
	protected void createTools(Panel toolPanel) {
		Panel palette = new Palette(menuFrame,"Drawing Tools");

		Tool tool = createSelectionTool();
		fDefaultToolButton = createToolButton(IMAGES+"SEL", "Selection Tool", tool);
		palette.add(fDefaultToolButton);

		palette.add(new java.awt.Canvas());

		tool = new CreationTool(this, new RectangleFigure());
		palette.add(createToolButton(IMAGES+"RECT", "Rectangle Tool", tool));

		tool = new CreationTool(this, new RoundRectangleFigure());
		palette.add(createToolButton(IMAGES+"RRECT", "Round Rectangle Tool", tool));

		tool = new CreationTool(this, new EllipseFigure());
		palette.add(createToolButton(IMAGES+"ELLIPSE", "Ellipse Tool", tool));

		tool = new CreationTool(this, new DiamondFigure());
		palette.add(createToolButton(IMAGES+"DIAMOND", "Diamond Tool", tool));

		tool = new CreationTool(this, new TriangleFigure());
		palette.add(createToolButton(IMAGES+"TRIANGLE", "Triangle Tool", tool));

		tool = new CreationTool(this, new LineFigure());
		palette.add(createToolButton(IMAGES+"LINE", "Line Tool", tool));

		tool = new ConnectionTool(this, new LineConnection());
		palette.add(createToolButton(IMAGES+"CONN", "Connection Tool", tool));

		tool = new ConnectionTool(this, new ElbowConnection());
		palette.add(createToolButton(IMAGES+"OCONN", "Elbow Connection Tool", tool));

		tool = new ScribbleTool(this);
		palette.add(createToolButton(IMAGES+"SCRIBBL", "Scribble Tool", tool));

		tool = new PolygonTool(this);
		palette.add(createToolButton(IMAGES+"POLYGON", "Polygon Tool", tool));

		tool = new ImageFigureCreationTool(this);
		palette.add(createToolButton(IMAGES+"IMAGE", "Image Tool", tool));

		TextFigure prototype = new TextFigure(false);
		prototype.setAlignment(TextFigure.LEFT);
		tool = new TextTool(this, prototype);
		fTextTB = createToolButton(IMAGES+"TEXT", "Text Tool", tool);
		palette.add(fTextTB);

		prototype = new TextFigure();
		prototype.setAlignment(TextFigure.CENTER);
		tool = new ConnectedTextTool(this, prototype);
		fConnTextTB = createToolButton(IMAGES+"ATEXT","Connected Text Tool",tool);
		palette.add(fConnTextTB);

		toolPanel.add(palette);
	}


	/**
	 * Creates the selection tool used in this editor. Override to use
	 * a custom selection tool.
	 */
	protected Tool createSelectionTool() {
		return new SelectionTool(this);
	}

	/**
	 * Creates a tool button with the given image, tool, and text
	 */
	public ToolButton createToolButton(String iconName, String toolName, Tool tool) {
		return new ToolButton(this, iconName, toolName, tool);
	}

  /**
   * Creates the drawing view used in this application.
   * You need to override this method to use a DrawingView
   * subclass in your application. By default a standard
   * DrawingView is returned.
   * @param width  the default width of the new view.
   * @param height the default height of the new view.
   * @return the new view.
   */
  protected StandardDrawingView createDrawingView(int width, int height) {
		return new StandardDrawingView(this, width, height);
	}

	void unsetCurrentDrawing() { 
	  if (fViewFrame!=null) {
			CheckboxMenuItem item = (CheckboxMenuItem)fDrawingsMenu.getItemForDrawing(drawing());
			if (item != null) {			 
				item.setState(false);
			}
			fViewFrame=null;
	  }
	}

	void setCurrentDrawing(DrawingViewFrame viewFrame) {
	  if (viewFrame==null) {
			unsetCurrentDrawing();
	  } else if (fViewFrame == viewFrame) {
			// do nothing, as the frame is already current.
	  } else {
			Drawing frameDrawing=viewFrame.view().drawing();
			//System.out.println("Trying to activate drawing "+frameDrawing.getName());
			int pos=fDrawings.indexOf(frameDrawing);
		
			// Is the drawing still active or do we process a delayed event?
			if (pos>=0 && fViewFrames.removeElement(viewFrame)) {
				fViewFrames.addElement(viewFrame);
				unsetCurrentDrawing();
				// new: move viewFrame to last position in vector
				// end new
				fViewFrame = viewFrame;
				((CheckboxMenuItem)fDrawingsMenu.getItemForDrawing(frameDrawing)).setState(true);
				//System.out.println("Activated drawing "+frameDrawing.getName());
				//displayZOrder();
				// Adapt menu entries to the new situation
				menuStateChanged();
			}
	  }
	}

	/* *****
private void displayZOrder() {
   for (int i=0; i<fViewFrames.size(); ++i) {
	  System.out.println(i+". "+((DrawingViewFrame)fViewFrames.elementAt(i)).view().drawing().getName());
   }
}
***** */

	protected DrawingViewFrame getViewFrame(Drawing drawing) {
	  Enumeration viewFrames = fViewFrames.elements();
	  while (viewFrames.hasMoreElements()) {
		  DrawingViewFrame viewFrame =
				(DrawingViewFrame)viewFrames.nextElement();
		  if (viewFrame.view().drawing() == drawing)
				return viewFrame;
		}
		return null;
	}

	public DrawingView getView(Drawing drawing) {
		DrawingViewFrame viewFrame=getViewFrame(drawing);
		if (viewFrame!=null)
			return viewFrame.view();
		return null;
	}

	/**
	 * Sets the specified drawing, its view and view frame as
	 * the current drawing and view. 
	 * A view for the drawing will be created if no one exists.
	 * Default position and size for the new view will be taken
	 * from the view frame.
	 *
	 * @param drawing	  the drawing to be shown.
	 *
	 * @return the view frame that is now current.
	 **/
	public DrawingViewFrame showDrawingViewFrame(Drawing drawing) {
		return showDrawingViewFrame(drawing, null, null);
	}

	/**
	 * Sets the specified drawing, its view and view frame as
	 * the current drawing and view. 
	 * A view for the drawing will be created if no one exists.
	 * Default position and size for the new view can be
	 * specified.
	 *
	 * @param drawing	    the drawing to be shown.
	 * @param defaultLoc  the default location for a new view
	 *                    frame, if one has to be created.
	 *                    If <code>null</code>, the view frame
	 *                    defaults will be used.
	 * @param defaultSize the default size of a new view
	 *                    frame, if one has to be created.
	 *                    If <code>null</code>, the view frame
	 *                    defaults will be used.
	 *
	 * @return the view frame that is now current.
	 **/
	public DrawingViewFrame showDrawingViewFrame(Drawing drawing, 
	                                             Point defaultLoc, 
	                                             Dimension defaultSize) {
		DrawingViewFrame viewFrame = getViewFrame(drawing);
		if (viewFrame != null) {
			setCurrentDrawing(viewFrame);
			viewFrame.setState(Frame.NORMAL);
			viewFrame.show();
			viewFrame.requestFocus(); // was missing!!!
		} else {
		  // Nothing was found. Create a new frame.
		  viewFrame = newDrawingViewFrame(drawing, defaultLoc, defaultSize);
		}
		return viewFrame;
	}

	public boolean closeDrawing(Drawing drawing) {
		DrawingViewFrame viewFrame = getViewFrame(drawing);
		if (viewFrame != null) {
			return closeViewFrame(viewFrame);
		}
		return false;
	}

	/**
	 * Creates the drawing used in this application.
	 * You need to override this method to use a Drawing
	 * subclass in your application. By default a standard
	 * Drawing is returned.
	 */
	protected Drawing createDrawing() {
		return new StandardDrawing();
	}

	/**
	 * Creates the status line.
	 */
	protected TextField createStatusLine() {
		TextField field = new TextField("No Tool", 55);
		field.setEditable(false);
		return field;
	}

	/**
	 * Handles a user selection in the palette.
	 * @see PaletteListener
	 */
	public void paletteUserSelected(PaletteButton button,
																	boolean doubleclick) {
		ToolButton toolButton = (ToolButton) button;
		setSelected(toolButton);
		setTool(toolButton.tool(), toolButton.name());
		fToolButtonSticky=!fAlwaysSticky &&	 doubleclick ||
			fAlwaysSticky && !doubleclick;
	}

	/**
	 * Handles when the mouse enters or leaves a palette button.
	 * @see PaletteListener
	 */
	public void paletteUserOver(PaletteButton button, boolean inside) {
		ToolButton toolButton = (ToolButton) button;
		if (inside)
			showStatus(toolButton.name());
		else if (fSelectedToolButton != null)
			showStatus(fSelectedToolButton.name());
	}

	/**
	 * Gets the current drawing.
	 * @see DrawingEditor
	 */
	public Drawing drawing() {
		return view().drawing();
	}

	/**
	 * Gets the current tool.
	 * @see DrawingEditor
	 */
	public Tool tool() {
		return fTool;
	}

	/**
	 * Gets the current drawing view.
	 * @see DrawingEditor
	 */
	public DrawingView view() {
	  DrawingViewFrame viewFrame=fViewFrame;
	  if (viewFrame==null) {
			return NullDrawingView.INSTANCE;
	  } else {
			return viewFrame.view();
	  }
	}

	/**
	 * Gets the drawing view that was active before the current drawing view.
	 * @see DrawingEditor
	 */
	public DrawingView previousView() {
	  int numViews=fViewFrames.size();
	  if (numViews >= 2) {
		  return ((DrawingViewFrame)fViewFrames.elementAt(numViews-2)).view();
	  }
	  return null;
	}

	/**
	 * Toggle the sticky mode.
	 */
	public void toggleAlwaysSticky() {
	  fAlwaysSticky=!fAlwaysSticky;
	  fToolButtonSticky=fAlwaysSticky;
	}

	/**
	 * Gets the sticky tool mode.
	 */
	public boolean isStickyTools() {
	  return fToolButtonSticky;
	}

	/**
	 * Set the sticky tools mode.
	 */
	public void setStickyTools(boolean sticky) {
	  fToolButtonSticky=sticky;
	}

	/**
	 * Gets the editor's default tool.
	 * This is usually the selection tool.
	 */
	public Tool defaultTool() {
	  return fDefaultToolButton.tool();
	}

	/**
	 * Sets the default tool of the editor.
	 * @see DrawingEditor
	 */
	public void toolDone() {
	  // not while initializing!
	  if (windowCount > 0) {
			if (fToolButtonSticky && tool()!=null) {
				tool().deactivate();
				tool().activate();
			} else if (fDefaultToolButton!=null) {
				setTool(fDefaultToolButton.tool(), fDefaultToolButton.name());
				setSelected(fDefaultToolButton);
			}
	  }
	}

	/**
	 * Handles a change of the current selection. Updates all
	 * menu items that are selection sensitive.
	 * @see DrawingEditor
	 */
	public void selectionChanged(DrawingView view) {
		menuStateChanged();
	}

	public void menuStateChanged() {
		MenuBar mb = menuFrame.getMenuBar();
		CommandMenu editMenu = (CommandMenu)mb.getMenu(EDIT_MENU);
		editMenu.checkEnabled();
		CommandMenu layoutMenu = (CommandMenu)mb.getMenu(LAYOUT_MENU);
		layoutMenu.checkEnabled();
	}

	/**
	 * Shows a status message.
	 * @see DrawingEditor
	 */
	public void showStatus(String string) {
		fStatusLine.setText(string);
	}

	private void setTool(Tool t, String name) {
		if (fTool != null)
			fTool.deactivate();
		fTool = t;
		if (fTool != null) {
			fStatusLine.setText(name);
			fTool.activate();
		}
	}

	private void setSelected(ToolButton button) {
		if (fSelectedToolButton != null)
			fSelectedToolButton.reset();
		fSelectedToolButton = button;
		if (fSelectedToolButton != null)
			fSelectedToolButton.select();
	}


	public ToolButton toolButtonForTextFigure(TextFigure figure) {
		if (figure.parent() == null) {
		  return fTextTB;
		} else {
		  return fConnTextTB;
		}
	}

	public void doTextEdit(TextFigure textFigure, int line, int column) {
		ToolButton tb=toolButtonForTextFigure(textFigure);
		paletteUserSelected(tb,fToolButtonSticky);
		TextTool tt = (TextTool)tb.tool();
		tt.beginEdit(textFigure);
		if (line > 0 && column > 0) {
			tt.setCaretPosition(line,column);
		}
	}

	public void doTextEdit(TextFigure textFigure) {
		doTextEdit(textFigure, 0,0);
	}

	/**
	 * Exits the application. You should never override this method
	 */
	public void exit() {
		// Close all DrawingViewFrames:
		while (true) {
			Enumeration viewFrames = fViewFrames.elements();
			if (!viewFrames.hasMoreElements()) {
				break;
			} else {
			  DrawingViewFrame viewFrame =
					(DrawingViewFrame)viewFrames.nextElement();
			  if (!closeViewFrame(viewFrame))
				  return;
			}
		}
		destroy();
		toolFrame.setVisible(false);
		toolFrame.dispose();
		menuFrame.setVisible(false);	  // hide the Frame
		menuFrame.dispose();   // tell windowing system to free resources
		if (--windowCount <= 0)
			System.exit(0);
	}

	/**
	 * Handles additional clean up operations. Override to destroy
	 * or release drawing editor resources.
	 */
	protected void destroy() {
	}



	/**
	 * Creates a new drawing view with a new empty drawing.
	 */
	public void promptNew() {
		Drawing drawing = createDrawing();
		newDrawingViewFrame(drawing);
		autosaveManager.addDrawing(drawing);
	}

	/**
	 * Shows a file dialog and opens a drawing.
	 */
	public void promptOpen() {
	  String path = getLoadPath("Open File...", 
		                          null,
		                          new String[]{drawExtension,"ser"});
	  if (path != null) {
			loadAndOpenDrawing(path);
	  }
	}

	/**
	 * Shows a file dialog and inserts a drawing.
	 */
	public void promptInsert() {
	  String path = getLoadPath("Insert Drawing...", 
		                          null,
		                          new String[]{drawExtension,"ser"});
	  if (path != null) {
			Drawing drawing;
			loadAndInsertDrawing(path);
	  }
	}

	public boolean saveDrawing(Drawing drawing) {
	  String filename=drawing.getFilename();
	  if (filename==null) {
			if (!promptSaveAs(drawing))
				return false;
			filename=drawing.getFilename();
	  }
	  saveDrawing(drawing,filename);
	  return true;
	}

	// Returns null, if no backup file can be determined.
	private File makeBackupFile(String filename) {
		String backupFilename=
			StringUtil.stripFilenameExtension(filename)+".bak";
		if (backupFilename.equals(filename)) {
			return null;
		} else {
			return new File(backupFilename);
		}
	}

	public void saveDrawing(Drawing drawing, String filename) {
		String text=" drawing "+drawing.getName()+" as "+filename;
		showAndPrint("Saving"+text+"...");

		if (!drawing.getBackupStatus()) {
			// Create a backup.
			File mainFile=new File(filename);
			File backupFile=makeBackupFile(filename);
			if (backupFile!=null) {
				mainFile.renameTo(backupFile);
				drawing.setBackupStatus(true);
			}
		}

		if (drawExtension.equals(StringUtil.getExtension(filename))) {
			saveAsStorableOutput(drawing,filename);
		} else {
			saveAsObjectOutput(drawing,filename);
		}
		showStatus("Saved"+text+".");
		autosaveManager.renameDrawing(drawing);
	}

	/**
	 * Shows a file dialog for saving a drawing.
	 * If a file name is determined, the drawing's name is set accordingly.
	 * @return <code>true</code>,  if a file name was chosen and set.<br>
	 *		   <code>false</code>, if the action was cancelled for
	 *							   some reason.
	 */
	boolean promptSaveAs(Drawing drawing) {
		String filename=drawing.getFilename();
		if (filename==null) {
			filename=fLastPath
				+File.separator+drawing.getName()
				+"."+drawExtension;
		}
		toolDone();
		filename = getSavePath("Save File...",
		                       filename,
		                       new String[]{drawExtension, "ser"});
		// was: {drawExtension,"ser"}
		if (filename == null)
			return false;

		// If the file exists already, ask for permission to overwrite
		if (new File(filename). exists()) {
			int answer = MessageDialog.invokeDialog(
				menuFrame, "Renew: Confirm overwrite",
				new String[] {"The file \""+filename+"\"","does already exist."},
				MessageDialog.LEFT, new String[] {" Overwrite "," Cancel "}
			);
			if (answer == 1) {
			  return false;
			}
		}

		// Set the new file name for the drawing
		updateName(drawing,filename);

		// OK: Make sure to create another backup.
		drawing.setBackupStatus(false);
		return true;
	}

	/**
	 * Prints the drawing.
	 */
	public void print() {
		fTool.deactivate();
		PrintJob printJob = menuFrame.getToolkit().getPrintJob(fViewFrame, "Print Drawing", null);

		if (printJob != null) {
			Graphics pg = printJob.getGraphics();

			if (pg != null) {
				// OK: This was 
				// fViewFrame.drawingView.printAll(pg);
				// which relied on a standard drawing view.
				// There is no reason why a call to the standard
				// paint method should not succeed.
				view().paint(pg);
				pg.dispose(); // flush page
			}
			printJob.end();
		}
		fTool.activate();
	}

	/**
	 * Opens a save file dialog according to the parameters
	 * and returns the selected/entered file name.
	 * All parameters are handled as described in<code>
	 * getLoadOrSavePath()</code>.
	 * @see #getLoadOrSavePath
	 **/
	public String getSavePath(String title, 
	                          String defaultName,
	                          String[] defaultExtensions) {
	  return getLoadOrSavePath(title,defaultName,defaultExtensions,FileDialog.SAVE);
	}

	/**
	 * Opens a load file dialog according to the parameters
	 * and returns the selected/entered file name.
	 * All parameters are handled as described in<code>
	 * getLoadOrSavePath()</code>.
	 * @see #getLoadOrSavePath
	 **/
	public String getLoadPath(String title, 
	                          String defaultName,
	                          String[] defaultExtensions) {
	  return getLoadOrSavePath(title,defaultName,defaultExtensions,FileDialog.LOAD);
	}

	/**
	 * Opens a file dialog according to the parameters
	 * and returns the selected/entered file name.
	 * @param title	       Title to display in the dialog.
	 * @param defaultName  File name to propose to the user.
   *                     The file name is checked and completed
   *                     to end with one of the default Extensions.
   *                     If the file name is prefixed with a
   *                     directory, the dialog initially switches
   *                     its actual directory to this prefix.
   *                     In the other case, the actual directory is
   *                     set to the last directory used for files
   *                     with the file's extension.
   * @param defaultExtensions  Array of possible extensions.
   *                     The file filter is set up to show only
   *                     files with one of these extensions.
   *                     See <code>checkAndAddExtension()</code>
   *                     about the way the default extensions
   *                     affect proposed and returned file names.
   *                     If no default extensions are specified
   *                     (<code>null</code> or empty array), all
   *                     files are shown and no extension is required.
   * @param loadOrSave   One of the <code>LOAD</code> or
   *                     <code>SAVE</code> constants from
   *                     <code>java.awt.FileDialog</code>.
   *                     Use <code>getLoadPath()</code> or
   *                     <code>getSavePath()</code> to avoid
   *                     the need for these constants.
   * @return             Full path name or <code>null</code>, if
   *                     the dialog was cancelled.
   *                     The path is checked and completed to
   *                     end with one of the default extensions
   *                     before it is returned - the user has no
   *                     possibility to avoid this check.
	 * @see #getLoadPath
	 * @see #getSavePath
	 * @see #checkAndAddExtension
	 * @see java.awt.FileDialog
	 **/
	protected String getLoadOrSavePath(String title, 
	                                   String defaultName,
	                                   String[] defaultExtensions,
	                                   int loadOrSave) {
		String path = null;
		String defaultDirectory = null;
		FileDialog dialog = new FileDialog(menuFrame, title, loadOrSave);
		if (defaultName != null) {
		  //System.out.println("Given file name:	"+defaultName); 
		  defaultName = checkAndAddExtension(defaultName, defaultExtensions);
		  //System.out.println("After ext. check:	"+defaultName); 
		  defaultDirectory = StringUtil.getPath(defaultName);
		  //System.out.println("Given dir. name:	"+defaultDirectory); 
		  if ("".equals(defaultDirectory)) {
				defaultDirectory = getDirForExt(StringUtil.getExtension(defaultName));
				//System.out.println("Looked up dir.:	  "+defaultDirectory); 
		  } else {
				defaultName = StringUtil.getExtendedFilename(defaultName);
				//System.out.println("Short file name:	  "+defaultName); 
		  }
		} else if (defaultExtensions != null) {
		  if (defaultExtensions.length > 0) {
				//System.out.println("No default name."); 
				//System.out.println("First default ext.: "+defaultExtensions[0]); 
				defaultDirectory = getDirForExt(defaultExtensions[0]);
				//System.out.println("Looked up dir.:	  "+defaultDirectory); 
		  }
		}

		if (defaultDirectory != null)
		  dialog.setDirectory(defaultDirectory);
		if (defaultName != null)
		  dialog.setFile(defaultName);
		// Commented out. Might screw Macs.
		//if (defaultExtensions!=null) {
		//	dialog.setFilenameFilter(
		//		   new FileByExtensionFilter(defaultExtensions));
		//}

		dialog.show();

		String filename = dialog.getFile();
		if (filename != null) {
			filename = stripTrailingAsterisks(filename);
			//System.out.println("Entered file name:  "+filename); 
			String dirname = dialog.getDirectory();
			//System.out.println("Selected directory: "+filename); 
			if (!dirname.endsWith(File.separator)) {
				dirname += File.separator;
			}
			path = dirname + filename;
			path = checkAndAddExtension(path, defaultExtensions);
			path = StringUtil.makeCanonical(path);
			//System.out.println("Complete file name: "+path); 
			updateDirByExt(path);
		}
		dialog.dispose();
		return path;
	}

	/**
	 * Checks if the given path ends with one of the given
	 * extensions and adds one, if this was not the case.
	 * @param path	       Filename to check (and modify).
	 * @param defaultExtensions	 Array of possible extensions.
	 *				             If the given path has none of these extensions
	 *				             appended to its end, the first default
	 *				             extension is automatically appended.
	 *				             The given extension list is enforced, e.g.
	 *				             the result is guaranteed to end with one
	 *				             of these extensions.
	 *				             If no default extensions are specified
	 *				             (<code>null</code> or empty array), 
	 *				             no extension is required or appended.
	 * @return		         Completed filename (with one of the
	 *				             given extensions).
	 **/
	protected String checkAndAddExtension(String path, 
	                                      String[] defaultExtensions) {
	  if (defaultExtensions != null) {
			boolean hasExtension = false;
			String ext=StringUtil.getExtension(path);
			for (int i = 0; i < defaultExtensions.length; i++)
				if (defaultExtensions[i].equals(ext))
					hasExtension = true;
			String result = path;
			if ((defaultExtensions.length > 0) && !hasExtension)
				result += "."+defaultExtensions[0];
			return result;
	  } else {
			return path;
	  }
	}

	protected String stripTrailingAsterisks(String filename) {
		// workaround for bug on NT
		if (filename.endsWith("*.*"))
			return filename.substring(0, filename.length() - 4);
		else
			return filename;
	}

	/**
	 * Updates the directory-by-extension lookup table by
	 * extracting path and extension from the given filename.
	 * Then behaves like updateDirByExt(String ext, String path).
	 **/
	protected void updateDirByExt(String fullFilename) {
	  updateDirByExt(StringUtil.getExtension(fullFilename),
		               StringUtil.getPath(fullFilename));
	}

	/**
	 * Updates the directory-by-extension lookup table.
	 * Sets the default directory for the given extension
	 * to the given path.
	 * Also sets the fallback directory to the given path.
	 **/
	protected void updateDirByExt(String ext, String path) {
	  dirByExt.put(ext, path);
	  fallbackDir = path;
	}

	/**
	 * Queries the directory-by-extension lookup table
	 * for the default directory for the given extension.
	 * If the table does not contain an entry for the
	 * given extension, returns the fallback directory.
	 * May return <code>null</code> if the fallback
	 * directory was not set, either.
	 **/
	protected String getDirForExt(String ext) {
	  String path = (String)dirByExt.get(ext);
	  if (path == null) 
			path = fallbackDir;
	  return path;
	}
  
	// Changes version 0->1:
	//  - PolyLineFigure is now an AttributeFigure.
	// Changes version 4->5:
	//  - Saving parameters of arrow tips.
	// Changes version 5->6:
	//  - FSFigures are now CPNTextFigures.
	// Changes version 6->7:
	// - FSFigures now store closed nodes (as path list)
	// Changes version 7->8:
	// - PolylineFigures now store arrow tip type.
	// - Default font changed from "Helvetica" to "SansSerif"
	//   (patching of old files needed)
	private static final int FILEVERSION=8;

	static void trySaveAsStorableOutput(Drawing drawing, String filename,
	                                    Point loc, Dimension size,
	                                    boolean clearModified)
		throws IOException
	{
		drawing.lock();
		// TBD: should write a MIME header
		StorableOutput output = new StorableOutput(filename);
		// Write the current version. This is a
		// substitute for a MIME header.
		output.writeInt(FILEVERSION);
		// Write the drawing.
		output.writeStorable(drawing);
		// New in 1.2 beta 5 (version 4): Write Window Position:
		output.writeInt(loc.x);
		output.writeInt(loc.y);
		output.writeInt(size.width);
		output.writeInt(size.height);
		output.close();
		if (clearModified) {
			drawing.clearModified();
		}
		drawing.unlock();
	}

	public void saveAutosaveFile(Drawing drawing, String filename,
	                             Point loc, Dimension size)
		throws IOException
	{
		trySaveAsStorableOutput(drawing,filename,loc,size,false);
	}

	private void saveAsStorableOutput(Drawing drawing, String filename) {
		DrawingViewFrame viewFrame=getViewFrame(drawing);
		Point loc=viewFrame.getLocation();
		Dimension size=viewFrame.getSize();
		try {
			trySaveAsStorableOutput(drawing,filename,loc,size,true);
		} catch (IOException e) {
			System.err.println(e);
			showStatus(e.toString());
		}
	}

	private void saveAsObjectOutput(Drawing drawing, String file) {
		// TBD: should write a MIME header
		try {
			FileOutputStream stream = new FileOutputStream(file);
			ObjectOutput output = new ObjectOutputStream(stream);
			output.writeObject(drawing);
			output.close();
		} catch (IOException e) {
			System.err.println(e);
			showStatus(e.toString());
		}
	}

	private Point newWindowLoc;
	private Dimension newWindowDim;

	public synchronized void loadAndOpenDrawing(String file) {
		newWindowLoc=null;
		newWindowDim=null;
		Drawing drawing	 = loadDrawing(file);
		if (drawing != null) {
			DrawingViewFrame viewFrame;		
			if (newWindowLoc!=null && newWindowDim != null) {
				// restore old window position
				viewFrame=newDrawingViewFrame(drawing, newWindowLoc, newWindowDim);
			} else {
				viewFrame=newDrawingViewFrame(drawing);
			}
			autosaveManager.addDrawing(drawing);
		}
	}

	public synchronized void loadAndInsertDrawing(String file) {
		Drawing existingDrawing	 = drawing();
		Drawing newDrawing	= loadDrawing(file);
		DrawingView existingView = view();
		if (newDrawing != null) {
			FigureEnumeration figures = newDrawing.figures();
			existingView.clearSelection();
			while (figures.hasMoreElements()){
			  Figure fig = figures.nextFigure();
			  existingDrawing.add(fig);
			  existingView.addToSelection(fig);
			}
			existingView.checkDamage();
		}
	}

	private void showAndPrint(String msg) {
		showStatus(msg);
		System.out.println(msg);
	}

	private Drawing loadDrawing(String pathname) {
		File file;
		String msg;
		pathname = StringUtil.makeCanonical(pathname);
		showAndPrint("Loading drawing from "+pathname+"...");
		Drawing drawing = null;
		//		  toolDone();
		String type = guessType(pathname);
		if (type.equals("storable")) {
			drawing = readFromStorableInput(pathname);
			showStatus("Loaded drawing from "+pathname+".");
		} else if (type.equals("serialized")) {
			drawing = readFromObjectInput(pathname);
			showStatus("Loaded serialized drawing from "+pathname+".");
		} else {
			showAndPrint("Unknown file type of file \""+pathname+"\".");
		}
		if (drawing != null) {
			drawing.setName(StringUtil.getFilename(pathname));
			drawing.setFilename(pathname);
			fLastPath = StringUtil.getPath(pathname);
			if (drawing instanceof Figure) {
			  ((Figure)drawing).invalidate();
			}
			// As it is possible to load drawings without having
			// previously used the file dialog, update the directory
			// lookup table here, too.
			updateDirByExt(pathname);
		}
		return drawing;
	}

	/**
	 * Override to establish a new policy to create
	 * input handlers.
	 */

	protected StorableInput makeStorableInput(String file, boolean useUFT)
		throws FileNotFoundException {
	  return new StorableInput(file,useUFT);
	}

	private Drawing readFromStorableInput(String filename) {
		Drawing drawing = null;
		try {
			StorableInput input = makeStorableInput(filename,true);
			try {
			  drawing = readStorableDrawing(input);
			} catch (sun.io.MalformedInputException e) {
				input.close();
				input = makeStorableInput(filename,false);
				drawing = readStorableDrawing(input);
			}
			try {
				// is there a window position to restore?
				newWindowLoc=new Point(input.readInt(),input.readInt());
				newWindowDim=new Dimension(input.readInt(),input.readInt());
				//System.out.println("Window Coords: "
				//					 +newWindowLoc+" / "+newWindowDim);
			} catch (Exception e) {}
		} catch (IOException e) {
			showStatus("Error: " + e);
			System.out.println("Error: " + e);
		}
		return drawing;
	}

	private Drawing readStorableDrawing(StorableInput input)
	  throws IOException
	{
	  if (input.canReadInt()) {
			input.setVersion(input.readInt());
	  } else {
			input.setVersion(0);
	  }
	  return (Drawing)input.readStorable();
	}

	private Drawing readFromObjectInput(String file) {
		Drawing drawing = null;
		try {
			FileInputStream stream = new FileInputStream(file);
			ObjectInput input = new ObjectInputStream(stream);
			drawing = (Drawing)input.readObject();
		} catch (IOException e) {
			showStatus("Error: " + e);
			System.out.println("Error: " + e);
		} catch (ClassNotFoundException e) {
			showStatus("Class not found: " + e);
			System.out.println("Error: " + e);
		}
		return drawing;
	}

	private String guessType(String file) {
		String ext=StringUtil.getExtension(file);
		if (drawExtension.equals(ext))
			return "storable";
		if ("rnw".equals(ext))
			return "storable";
		if ("ser".equals(ext))
			return "serialized";
		return "unknown";
	}

	public void windowOpened(WindowEvent e) {
		//DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
		//fViewFrames.addElement(viewFrame);
		//System.out.println("New Drawing View Window: "+viewFrame.getTitle());
	}

	public void windowActivated(WindowEvent e) {
	  DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
	  if (viewFrame != null && viewFrame != fViewFrame) {
			//System.out.println("Current Drawing View Window: "+viewFrame.getTitle());
	  }
	  setCurrentDrawing(viewFrame);
	}

	public void windowDeactivated(WindowEvent e) {	  }

	public void windowClosing(WindowEvent e) {
	  DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
	  // !!!!!!!!!!! This was:
	  // removeDrawing(viewFrame.drawingview.drawing());
	  closeViewFrame(viewFrame);
	}

	public void windowClosed(WindowEvent e) {
	}

	public void windowIconified(WindowEvent e) {
	}

	public void windowDeiconified(WindowEvent e) {
	}

	public UndoRedoManager getUndoRedoManager() {
		return undoManager;
	}

	public void prepareUndoSnapshot() {		 
		Drawing currentDrawing = drawing();
		undoManager.prepareUndoSnapshot(currentDrawing);
	}

	public void commitUndoSnapshot() {
		Drawing currentDrawing = drawing();
		undoManager.commitUndoSnapshot(currentDrawing) ;
	}

	/**
	 * Prohibits the management of undo and redo snapshots
	 * for the given drawing.
	 **/
	protected void noUndoHistoryFor(Drawing drawing) {
		undoManager.removeUndoHistory(drawing);
	}

	public boolean selectOffendingElements(FigureException e) {
		if (e.errorDrawing != null && !e.errorFigures.isEmpty()) {
			view().clearSelection();
			showDrawingViewFrame(e.errorDrawing);
			FigureEnumeration errorFigures=new FigureEnumerator(e.errorFigures);
			view().addToSelectionAll(errorFigures);

			// Redraw the newly selected elements.
			view().repairDamage();
			if (e.textErrorFigure != null) {
				doTextEdit(e.textErrorFigure,e.line,e.column);
			}
			return true;
	  } else {
			return false;
	  }
	}

	//-- main -----------------------------------------------------------

	public static void main(String[] args) {
		new DrawApplication("JHotDraw",args);
	}

}

class DrawingMenuListener implements ItemListener {
	private DrawApplication fEditor;
	private Drawing fDrawing;

	DrawingMenuListener(DrawApplication editor, Drawing drawing) {
	  fEditor = editor;
	  fDrawing = drawing;
	}

	public void itemStateChanged(ItemEvent event) {
	  fEditor.showDrawingViewFrame(fDrawing);
	}
}

class DrawingLoadServer extends Thread {

	private DrawApplication editor;
	private ServerSocket s=null;

	public DrawingLoadServer(DrawApplication editor, int port) throws IOException {
		this.editor=editor;
		System.out.println("Drawing Load Server setting up server socket at port "+port+"...");
		s=new ServerSocket(port);
	}

	public void run() {
		System.out.println("Drawing Load Server waiting for parameters...");
		while (true) {
			Socket client=null;
			try {
				client=s.accept();
				System.out.println("Parameter server accepted client.");
				BufferedReader in=new BufferedReader(new InputStreamReader(client.getInputStream()));
				try {
					String drawingFileName;
					do {
						drawingFileName=in.readLine();
						if (drawingFileName!=null) {
							System.out.println("Received Parameter "+drawingFileName);
							editor.loadAndOpenDrawing(drawingFileName);
						}
					} while (drawingFileName != null);
				} catch (IOException e) { }
				System.out.println("Connection closed.");
				in.close();
				client.close();
			} catch (Exception e) {
				System.out.println("Drawing Load Server threw exception: "+e);
			}
		}
	}



}
