package CH.ifa.draw.application;

import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.Window;

import CH.ifa.draw.framework.Drawing;

import java.util.Hashtable;
import java.util.Vector;

/**
 * A windows menu which groups its menu items by category.
 * <p></p>
 * WindowsMenu.java
 * Created: Tue Mar 20	2001
 * @author Michael Duvigneau
 */
public class WindowsMenu extends Menu {

	private DrawApplication editor;

	/**
	 * Table to organize the menus in categories.
	 * Contains pairs of the type (<category>, <menu list>).
	 **/
	private Hashtable categoryMap = new Hashtable();

	/**
	 * Table for reverse lookups.
	 * Contains pairs of the type (<item object>, <category>).
	 **/
	private Hashtable objectMap = new Hashtable();

	private static final String NO_CATEGORY = "Other windows";

	public WindowsMenu (String name, DrawApplication editor){
		super(name);
		this.editor = editor;
		setEnabled(false);
	}

	public void addDrawing(Drawing drawing, MenuItem item) {
		String category = drawing.getWindowCategory();
		addCategorizedItem(category, item, drawing);
	}

	public MenuItem getItemForDrawing(Drawing drawing) {
		return getItemFor(drawing);
	}

	public void removeDrawing(Drawing drawing) {
		removeItemFor(drawing);
	}

	public void addCategorizedItem(String category, 
																 MenuItem item,
																 Object object) {
		if (category == null) {
			category = NO_CATEGORY;
		}
		
		MenuList ml = (MenuList)categoryMap.get(category);
		if (ml == null) {
			Menu menu = editor.createMenu(category);
			ml = new MenuList(menu, new Vector());
			categoryMap.put(category, ml);
			int pos = findSortedPosition(this, menu);
			insert(menu, pos);
			setEnabled(true);
		} 

		int pos = findSortedPosition(ml.menu, item);
		ml.menu.insert(item, pos);
		ml.list.insertElementAt(object, pos);
		objectMap.put(object, category);
	}

	public MenuItem getItemFor(Object object) {
		MenuItem result = null;
		try {
			String category = (String)objectMap.get(object);
			MenuList ml = (MenuList)categoryMap.get(category);
			int pos = ml.list.indexOf(object);
			result = ml.menu.getItem(pos);
		} catch (NullPointerException e) {
		} catch (ArrayIndexOutOfBoundsException e) {
		}
		return result;
	}

	public void removeItemFor(Object object) {
		try {
			String category = (String)objectMap.get(object);
			MenuList ml = (MenuList)categoryMap.get(category);
			int pos = ml.list.indexOf(object);
			ml.menu.remove(pos);
			ml.list.removeElementAt(pos);
			objectMap.remove(object);
			if (ml.list.isEmpty()) {
				remove(ml.menu);
				categoryMap.remove(category);
				if (categoryMap.isEmpty()) {
					setEnabled(false);
				} 
			}				
		} catch (NullPointerException e) {
		} catch (ArrayIndexOutOfBoundsException e) {
		}
	}

	/**
	 * Data structure to group a menu with a corresponding
	 * list of objects.
	 **/
	private class MenuList {
		public final Menu menu;
		public final Vector list;
		public MenuList(Menu menu, Vector list) {
			this.menu = menu;
			this.list = list;
		}
	}

	/**
	 * Returns the position of the first <code>menu</code> item
	 * whose label is greater than (lexicographically, ignoring
	 * case) the given <code>newItem</code>'s label. If there
	 * aren't any existing items in the given menu, returns 0. If
	 * the new item's label is greater than all existing menu
	 * items, returns <code>menu.getItemCount()</code>.  
	 **/
	private static int findSortedPosition(Menu menu, MenuItem newItem) {
		int min = 0;
		int max = menu.getItemCount();
		int pos, comp;
		MenuItem item;
		String searchString = newItem.getLabel().toLowerCase();
		// System.out.println("Sorting in \""+searchString+"\"");
		while (max > min) {
			pos = (max + min) / 2;
			item = menu.getItem(pos);
			comp = searchString.compareTo(item.getLabel().toLowerCase());
			// System.out.println
			// 	("  try: max "+max+", min "+min+", pos "+pos+", comp "+comp
			// 	 +", item \""+item.getLabel().toLowerCase()+"\"");
			if (comp < 0) {
				// searchString is less than current item label
				max = pos;
			} else if (comp >= 0) {
				// searchString is greater than or equal to current item label
				min = pos + 1;
			}
		}
		return min;
	}
}
