/*
 * @(#)AttributeFigure.java 5.1
 *
 */

package CH.ifa.draw.figures;

import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;

import java.awt.*;
import java.util.*;
import java.io.*;

/**
 * A figure that can keep track of an open ended set of attributes.
 * The attributes are stored in a dictionary implemented by
 * FigureAttributes.
 *
 * @see Figure
 * @see Handle
 * @see FigureAttributes
 */

// It is assume by PolyLineFigure that AttributeFigure
// inherits directly from AbstractFigure.

public abstract class AttributeFigure extends AbstractFigure
                                      implements FigureWithID {

    /**
     * The attributes of a figure. Each figure can have
     * an open ended set of attributes. Attributes are
     * identified by name.
     * @see #getAttribute
     * @see #setAttribute
     * @serial
     */
    private FigureAttributes        fAttributes;

    /**
     * The default attributes associated with a figure.
     * If a figure doesn't have an attribute set, a default
     * value from this shared attribute set is returned.
     * @see #getAttribute
     * @see #setAttribute
     */
    private static FigureAttributes fgDefaultAttributes = null;

    /*
     * Serialization support.
     */
    private static final long serialVersionUID = -10857585979273442L;
    private int attributeFigureSerializedDataVersion = 1;

    protected AttributeFigure() { }

    public void internalDraw(Graphics g) {
      Color fill = getFillColor();
      if (!ColorMap.isTransparent(fill)) {
	g.setColor(fill);
	drawBackground(g);
      }
      Color frame = getFrameColor();
      if (!ColorMap.isTransparent(frame)) {
	g.setColor(frame);
	drawFrame(g);
      }
    }

    /**
     * Draws the figure in the given graphics. Draw is a template
     * method calling drawBackground followed by drawFrame.
     *
     * If it is sometimes required to override this method,
     * make internalDraw() more public and override that method.
     */
    public final void draw(Graphics g) {
	if (isVisible()) {
	  internalDraw(g);
	}
    }

    /**
     * Draws the figure in an  appearance according to the DrawingContext.
     * @param g the Graphics to draw into
     * @param dc the DrawingContext to obey
     */
    public void draw(Graphics g, DrawingContext dc) {
      if (!dc.isVisible(this)) {
	    return;
      }
	if (dc.isHighlighted(this)) {
	    Color fill = getFillColor();
	    Color frame = getFrameColor();
	    Color text = null;
	    if (ColorMap.isTransparent(fill)
		|| ColorMap.isBackground(fill)
		|| this instanceof PolyLineFigure) {
		if (this instanceof TextFigure) {
		    text = (Color)getAttribute("TextColor");
		    setTextColor(ColorMap.hilight(text));
		} else if (!ColorMap.isTransparent(frame)
			   && !ColorMap.isBackground(fill)) {
		    setFrameColor(ColorMap.hilight(frame));
		}
	    } else {
		setFillColor(ColorMap.hilight(fill));
	    }
	    internalDraw(g);
	    setFillColor(fill);
	    setFrameColor(frame);
	    if (text!=null)
		setTextColor(text);
	} else {
	  internalDraw(g);
	}
    }

    /**
     * Draws the background of the figure.
     * @see #draw
     */
    protected void drawBackground(Graphics g) {
    }

    /**
     * Draws the frame of the figure.
     * @see #draw
     */
    protected void drawFrame(Graphics g) {
    }

    /**
     * Gets the fill color of a figure. This is a convenience
     * method.
     @see getAttribute
     */
    public Color getFillColor() {
        return (Color) getAttribute("FillColor");
    }

    /**
     * Gets the frame color of a figure. This is a convenience
     * method.
     @see getAttribute
     */
    public Color getFrameColor() {
        return (Color) getAttribute("FrameColor");
    }

    /**
     * Sets the fill color of a figure <b>without</b> issuing
     * a <tt>changed</tt> Event.
     */
    public void setFillColor(Color color) {
        if (fAttributes == null)
            fAttributes = new FigureAttributes();
        fAttributes.set("FillColor", color);
    }

    /**
     * Sets the frame color of a figure <b>without</b> issuing
     * a <tt>changed</tt> Event.
     */
    public void setFrameColor(Color color) {
        if (fAttributes == null)
            fAttributes = new FigureAttributes();
        fAttributes.set("FrameColor", color);
    }

    /**
     * Sets the text color of a figure <b>without</b> issuing
     * a <tt>changed</tt> Event.
     */
    public void setTextColor(Color color) {
        if (fAttributes == null)
            fAttributes = new FigureAttributes();
        fAttributes.set("TextColor", color);
    }

    //---- figure attributes ----------------------------------

    private static void initializeAttributes() {
        fgDefaultAttributes = new FigureAttributes();
        fgDefaultAttributes.set("FrameColor", Color.black);
        fgDefaultAttributes.set("FillColor",  new Color(0x70DB93));
        fgDefaultAttributes.set("TextColor",  Color.black);
        fgDefaultAttributes.set(TextFigure.ALIGN_ATTR, new Integer(TextFigure.LEFT));
        fgDefaultAttributes.set("ArrowMode",  new Integer(0));
        fgDefaultAttributes.set("FontName",  "Helvetica");
        fgDefaultAttributes.set("FontSize",   new Integer(12));
        fgDefaultAttributes.set("FontStyle",  new Integer(Font.PLAIN));
        fgDefaultAttributes.set("LineShape",  new Integer(PolyLineFigure.LINE_SHAPE));
	fgDefaultAttributes.set("BSplineSegments", new Integer(BSpline.DEFSEGMENTS));
	fgDefaultAttributes.set("BSplineDegree", new Integer(BSpline.DEFDEGREE));

	// Initialize the ID of this Figure to NOID.
	// (See comments to getID(), setID() and _idAttr below.)
        fgDefaultAttributes.set(_idAttr,      new Integer(FigureWithID.NOID));
    }

    /**
     * Gets a the default value for a named attribute
     * @see getAttribute
     */
    public static Object getDefaultAttribute(String name) {
        if (fgDefaultAttributes == null)
            initializeAttributes();
        return fgDefaultAttributes.get(name);
    }

    /**
     * Returns the named attribute or null if a
     * a figure doesn't have an attribute.
     * All figures support the attribute names
     * FillColor and FrameColor
     */
    public Object getAttribute(String name) {
        if (fAttributes != null) {
            if (fAttributes.hasDefined(name))
                return fAttributes.get(name);
        }
	Object supAttr=super.getAttribute(name);
	if (supAttr==null)
	    return getDefaultAttribute(name);
	else
	    return supAttr;
    }

    /**
     * Sets the named attribute to the new value
     */
    public void setAttribute(String name, Object value) {
	super.setAttribute(name,value);
        if (fAttributes == null)
            fAttributes = new FigureAttributes();
        fAttributes.set(name, value);
        changed();
    }

    /**
     * Stores the Figure to a StorableOutput.
     */
    public void write(StorableOutput dw) {
        super.write(dw);
	// The write() method of PolyLineFigure assumes that
	// the first thing that this method writes is a string
	// and not a number.
        if (fAttributes == null)
            dw.writeString("no_attributes");
        else {
            dw.writeString("attributes");
            fAttributes.write(dw);
        }
    }

    /**
     * Reads the Figure from a StorableInput.
     */
    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        String s = dr.readString();
        if (s.toLowerCase().equals("attributes")) {
	    if (fAttributes == null)
		fAttributes = new FigureAttributes();
            fAttributes.read(dr);
	    Enumeration attrenum=fAttributes.definedAttributes();
	    while (attrenum.hasMoreElements()) {
		String attr=(String)attrenum.nextElement();
		Object val=fAttributes.get(attr);
		super.setAttribute(attr,val);
	    }
        }
    }


    /**
     * Name of the attribute holding the ID required by
     * interface FigureWithID.
     * The ID is stored as an attribute to avoid changing
     * the data format written to streams.
     * @see CH.ifa.draw.framework.FigureWithID
     **/
    private static final String _idAttr = "FigureWithID";
    // An additional line concerning this attribute has
    // been added to initializeDefaultAttributes.

    /**
     * Get the ID as required by interface FigureWithID.
     * @see CH.ifa.draw.framework.FigureWithID
     **/
    public int getID() {
	return ((Integer)getAttribute(_idAttr)).intValue();
    }

    /**
     * Set the ID as required by interface FigureWithID.
     * @see CH.ifa.draw.framework.FigureWithID
     **/
    public void setID(int id) {
	setAttribute(_idAttr, new Integer(id));
    }
}
