package CH.ifa.draw.util;

import CH.ifa.draw.framework.*;
import de.renew.util.Scheduler;
import de.renew.util.StringUtil;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.io.FileOutputStream;
import java.io.File;
import java.io.IOException;

public class AutosaveTask implements Runnable {
  private Drawing drawing;
  private long lastSaved;
  private long interval;
  /**
   * <code>autosaveFilename</code> is null, until the first
   * time that the drawing was saved. Afterwards, it keeps track
   * of the name of the autosave file.
   */
  private String autosaveFilename=null;
  private AutosaveSaver saver;

  public AutosaveTask(AutosaveSaver saver,
		      Drawing drawing,
		      long interval)
  {
    this.saver=saver;
    this.drawing=drawing;
    this.interval=interval;
    reset();

    Scheduler.instance().executeIn(this,interval);
  }

  private void touch(String filename) {
    try {
      FileOutputStream stream=new FileOutputStream(filename);
      stream.close();
    } catch (IOException e) {
      // Never try to save again.
      interval=0;
    }
  }

  public synchronized void reset() {
    lastSaved=System.currentTimeMillis();
  }

  /**
   * Terminate the autosave process, removing the autosave file,
   * if it exists.
   */
  public synchronized void terminate() {
    interval=0;
    if (autosaveFilename!=null) {
      File autosaveFile=new File(autosaveFilename);
      if (autosaveFile.exists()) {
	autosaveFile.delete();
      }
    }
  }

  public synchronized void run() {
    if (interval>0) {
      // Ok, this autosave task is still associated to an
      // open drawing.
      long now=System.currentTimeMillis();
      long remainingDelay=lastSaved+interval-now;
      if (remainingDelay<=0) {
	if (drawing.isModified()) {
	  if (autosaveFilename==null) {
	    String filename=drawing.getFilename();
	    if (filename!=null) {
	      autosaveFilename=
		StringUtil.stripFilenameExtension(filename)+".aut";
	    }
	    if (autosaveFilename!=null &&
		!new File(autosaveFilename).exists())
	      {
		touch(autosaveFilename);
	      } else {
		autosaveFilename=null;
		for (int i=0;i<25 && autosaveFilename==null;i++) {
		  int autosaveNumber=(int)Math.floor(Math.random()*100000);
		  String testFilename="rnw"+autosaveNumber+".aut";
		  File testFile=new File(testFilename);
		  // Should use File.createNewFile(), but that is
		  // only available from Java 1.2 on.
		  if (!testFile.exists()) {
		    // Make sure that the file exists, so that
		    // further autosave files are saved under a different name.
		    touch(testFilename);
		    // Remember the filename.
		    autosaveFilename=testFilename;
		  }
		}
	      }
	  }

	  if (autosaveFilename!=null) {
	    Rectangle rect=drawing.displayBox();
	    Dimension size=new Dimension(rect.width+rect.x+200,
					 rect.height+rect.y+200);
	    try {
	      saver.saveAutosaveFile(drawing,
				     autosaveFilename,
				     new Point(),
				     size);
	    } catch (IOException e) {
	      // An error occurred. Do not save again.
	      interval=0;
	    }
	  }
	}
	// We just tried to save. Reset the clock.
	reset();
      }
      if (interval>0) {
	// Another save is desired.
	Scheduler.instance().executeAt(this,lastSaved+interval);
      }
    }
  }
}
