package CH.ifa.draw.util;

import java.awt.*;
import java.util.*;

public interface ExtendedGraphics  {
 /** Additional Graphics methods, e.g. for drawing dashed lines. */

  /**
   * Draws a dashed line starting at (x1,y1) and ending at (x2,y2)
   * with the interval size strokeSize. That means a stroke of length
   * strokeSize is followed by a gap of length strokeSize and so on.
   *
   * @param strokeSize the length of dashes and gaps (as described
   *                   above).
   * @param ...        all other parameters are unchanged to the
   *                   method <code>Graphics.drawLine()</code>.
   *
   * @see Graphics#drawLine
   */
  public void drawDashedLine(int x1, int y1, int x2, int y2, int strokeSize);

  /**
   * Draws a dashed line starting at (x1,y1) and ending at (x2,y2)
   * using the stroke pattern defined in <code>strokeSequence</code>.
   * <p>
   * The <code>strokeSequence</code> is a string of integers separated
   * by spaces. The values alternating determine dash and gap lengths,
   * beginning with a dash length. The string has to contain an even
   * number of such values and may not be empty.
   * As a special case it is allowed to supply one single value, this
   * value will then be used both for dashes and gaps.
   * <pre>
   * Examples:
   *    "1 7"       dotted line           . . . . . . .
   *
   *    "5 4"       dashed line           _ _ _ _ _ _ _
   *
   *    "5 4 1 4"   dash-dotted line      _ . _ . _ . _
   *
   *    "4"         dashed line           - - - - - - -
   * </pre>
   * If you want to check a stroke pattern for validity, use the
   * static method {@link ExtendedGraphicsImpl#isValidStrokePattern}.
   * </p>
   *
   * @param strokeSequence the pattern for dashes and gaps (as
   *                   described above).
   * @param ...        all other parameters are unchanged to the
   *                   method <code>Graphics.drawLine()</code>.
   *
   * @see Graphics#drawLine
   */
  public void drawDashedLine(int x1, int y1, int x2, int y2, String strokeSequence);

  /**
   * Draws a dashed Polyline.
   *
   * @param strokeSize the length of dashes and gaps
   *                   (see <code>drawDashedLine()</code>).
   * @param ...        all other parameters are unchanged to the
   *                   method <code>Graphics.drawPolyline()</code>.
   *
   * @see #drawDashedLine(int,int,int,int,int)
   * @see Graphics#drawPolyline
   */
  public void drawDashedPolyline(int xPoints[], int yPoints[], int nPoints,
				 int strokeSize);

  /**
   * Draws a dashed Polyline.
   *
   * @param strokeSequence the pattern for dashes and gaps
   *                   (see <code>drawDashedLine()</code>).
   * @param ...        all other parameters are unchanged to the
   *                   method <code>Graphics.drawPolyline()</code>.
   *
   * @see #drawDashedLine(int,int,int,int,String)
   * @see Graphics#drawPolyline
   */
  public void drawDashedPolyline(int xPoints[], int yPoints[], int nPoints,
				 String strokeSequence);

  /**
   * Draws a dashed Polyline.
   *
   * @param strokeSize the length of dashes and gaps
   *                   (see <code>drawDashedLine()</code>).
   * @param points     a <code>Vector</code> of {@link Point}
   *                   objects that describe the polyline.
   *
   * @see #drawDashedLine(int,int,int,int,int)
   * @see Graphics#drawPolyline
   */
  public void drawDashedPolyline(Vector points, int strokeSize);

  /**
   * Draws a dashed Polyline.
   *
   * @param strokeSequence the pattern for dashes and gaps
   *                   (see <code>drawDashedLine()</code>).
   * @param points     a <code>Vector</code> of {@link Point}
   *                   objects that describe the polyline.
   *
   * @see #drawDashedLine(int,int,int,int,String)
   * @see Graphics#drawPolyline
   */
  public void drawDashedPolyline(Vector points, String strokeSequence);

}
