package CH.ifa.draw.util;

import java.awt.*;
import java.util.*;
import de.renew.util.StringUtil;

public class ExtendedGraphicsImpl extends ContextGraphics
				  implements ExtendedGraphics {

    /** Extends ContextGraphics by the ability to draw dashed lines.
     */
  public ExtendedGraphicsImpl(Graphics g, Color background) {
      super(g,background);
  }

  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedLine(int x1, int y1, int x2, int y2, int strokeSize) {
      drawDashedLine(x1,y1,x2,y2,new int[]{strokeSize, strokeSize}, 5);
  }

  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedLine(int x1, int y1, int x2, int y2, String strokeSequence) {
      drawDashedLine(x1,y1,x2,y2,convert(strokeSequence),5);
  }
  

  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedPolyline(int xPoints[], int yPoints[], int nPoints,
				 int strokeSize) {
      drawDashedPolyline(xPoints, yPoints, nPoints, 
			 new int[] {strokeSize, strokeSize});
  }

  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedPolyline(int xPoints[], int yPoints[], int nPoints,
				 String strokeSequence) {
      drawDashedPolyline(xPoints, yPoints, nPoints, 
			 convert(strokeSequence));
  }

  /**
   * The common body of the methods defined by the interface
   * {@link #drawDashedPolyline(int[], int[], int, int)} and
   * {@link #drawDashedPolyline(int[], int[], int, String)}.
   **/
  private void drawDashedPolyline(int xPoints[], int yPoints[], int nPoints,
				  int[] strokeSequence) {
      if (nPoints>=2) {
	  int offset=0;
	  int x=xPoints[0],y=yPoints[0];
	  for (int i=1; i<nPoints; ++i) {
	    int x1=x, y1=y;
	    x=xPoints[i]; y=yPoints[i];
	    offset=drawDashedLine(x1,y1,x,y,strokeSequence,offset);
	  }
      }
  }


  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedPolyline(Vector points, int strokeSize) {
      drawDashedPolyline(points, new int[] {strokeSize, strokeSize});
  }

  /* This method is defined by the interface ExtendedGraphics. */
  public void drawDashedPolyline(Vector points, String strokeSequence) {
      drawDashedPolyline(points, convert(strokeSequence));
  }

  /**
   * The common body of the methods defined by the interface
   * {@link #drawDashedPolyline(Vector, int)} and
   * {@link #drawDashedPolyline(Vector, String)}.
   **/
  private void drawDashedPolyline(Vector points, int[] strokeSequence) {
      int index=points.size();
      if (index>=2) {
	  int offset=0;
	  Point old=(Point)points.elementAt(--index);
	  while (index>0) {
	    Point next=(Point)points.elementAt(--index);
	    offset=drawDashedLine(old.x,old.y,next.x,next.y,strokeSequence,offset);
	    old=next;
	  } 
      }
  }


  /**
   * Draws a dashed line with different dash and gap sizes.
   *
   * @param (x1,y1)    the coordinates of the point at line beginning
   * @param (x2,y2)    the coordinates of the point at line end
   * @param strokeSequence an array of <code>int</code> values as
   *                   generated by {@link #convert}.
   *                   The values are interpreted as alternating 
   *                   dash and gap lengths, beginning with a dash.
   *                   The array must have an even number of elements.
   * @param offset     the size of the beginning gap. By default,
   *                   the line begins with a dash immediately at
   *                   its start point.
   *
   * @return the offset to be used for the next line, if it is a
   *         continuation of the just drawn line
   */
  private int drawDashedLine(int x1, int y1, int x2, int y2, int[] strokeSequence,
			     int offset) {
    double alpha=Math.atan2(x2-x1,y2-y1);
    double dx=Math.sin(alpha),
	   dy=Math.cos(alpha);
    double length=Geom.length(x1,y1, x2,y2);
    int s=offset;
    while (s<length) {
      for (int stroke=0; stroke<strokeSequence.length; stroke++) {
        int xa,ya,xb,yb;
	boolean draw = true;

        // Beginning of dash	
	if (s<=0) {
	    xa=x1;
	    ya=y1;
	} else if (s>=length) {
	    // The whole dash has to be clipped.
	    draw = false;
	    xa=x1;
	    ya=y1;
	} else {
	    xa=x1+(int)Math.round(s*dx);
	    ya=y1+(int)Math.round(s*dy);
	}
	s+=strokeSequence[stroke];
	stroke++;

	// End of dash, beginning of gap
	if (s<=0) {
	    // The whole dash has to be clipped.
	    draw = false;
	    xb=x2;
	    yb=y2;
	} else if (s>=length) {
	    xb=x2;
	    yb=y2;
	} else {
	    xb=x1+(int)Math.round(s*dx);
	    yb=y1+(int)Math.round(s*dy);
	}
	s+=strokeSequence[stroke];
	//stroke++; is done by the for statement

	// Draw the dash, if not clipped.
	if (draw) {
	    drawLine(xa,ya,xb,yb);	     
	}
      }
    }
    s-=length;
    if (s>=strokeSequence[strokeSequence.length - 1]) {
	// If the offset for a continued line is larger than
	// the last dash length, we have to move the offset
	// backwards by one full stroke pattern length.
	// More than one pattern length should not be needed
	// as the line length has been subtracted before.
	for (int i=0; i < strokeSequence.length; i++) {
	    s-=strokeSequence[i];
	} 	
    }
    return s;
  }

  /** Additionally supports underlined text. */
  public void drawString(String text, int x, int y) {
    if (text.length()>0) {
      g.drawString(text,x,y);
      // for special fonts like ZapfDingbats, better use:
      //g.drawBytes(text.getBytes(),0,line.length(),x,y);
      Font font=g.getFont();
      if (font instanceof ExtendedFont &&
	((ExtendedFont)font).isUnderlined()) {
	int pixel=g.getFont().getSize()/12;
	if (pixel==0) pixel=1;
	y+=pixel;
	g.fillRect(x, y, g.getFontMetrics().stringWidth(text), pixel);
      }
    }
  }

  /**
   * Checks if the given string can be interpreted as a valid
   * stroke sequence for the <code>drawDashed</code>... methods
   * of the interface <code>ExtendedGraphics</code>.
   *
   * @param strokeSequence the string to check.
   *
   * @return <code>true</code> if the string contains a valid
   *         stroke sequence.
   **/
  public static boolean isValidStrokePattern(String strokeSequence) {
    try {
      convert(strokeSequence);
    } catch (IllegalArgumentException e) {
      return false;
    } catch (NullPointerException e) {
      return false;
    }
    return true;
  }

  /**
   * Converts the given string into an array of <code>int</code>
   * values to be used by the internal draw methods of this class.
   *
   * @throws NullPointerException
   *         if the string is <code>null</code>.
   *
   * @throws IllegalArgumentException
   *         if the string contains a wrong number of values
   *         or no values at all.
   *
   * @throws NumberFormatException
   *         if one of the values cannot be converted into an
   *         <code>int</code>.
   **/
  private static int[] convert(String strokeSequence) {
    // Eliminate all unneeded spaces.
    String sequence = StringUtil.unspace(strokeSequence);

    // If the string doesn't contain any value, report an error.
    if (sequence.equals("")) {
      throw new IllegalArgumentException(
                "Empty stroke sequence \""+strokeSequence+"\".");
    } 

    // Separate the string and count the number of values.
    String[] parts = StringUtil.split(sequence, " ");
    int count = parts.length;

    // A single value gets duplicated to be used for dashes and gaps.
    // If the string contains an odd number of values, report an error.
    if (count == 1) {
      parts = new String[] {parts[0], parts[0]};
      count = 2;
    } else if ((count & 1) == 1) {
      throw new IllegalArgumentException(
                "Odd number of values in stroke sequence \""
		+sequence+"\".");
    } 

    // Convert all values into integers and put them into the result
    // array. The conversion may throw a NumberFormatException.
    int[] result = new int[count];
    for (int i=0; i < count; i++) {
      result[i] = Integer.parseInt(parts[i]);
      if (result[i] <= 0) {
	throw new IllegalArgumentException(
		  "Stroke lengths must be positive (\""+result[i]
		  +"\"is not allowed).");
      }       
    } 

    return result;
  }

}
