package CH.ifa.draw.util;

/**
 * This helper class contains static functions which
 * can be queried for a default behaviour of some GUI
 * elements.
 * <p></p>
 * GUIProperties.java
 * Created: Wed Mar  7  2001
 * @author Michael Duvigneau
 */

public class GUIProperties {

  /** This class cannot be instantiated. **/
  private GUIProperties() {}

  /* The following fields act as a cache for once 
   * evaluated properties.
   */
  private static Boolean _avoidFrameReshape = null;
  private static Integer _loadServerPort =    null;
  private static Integer _menuFontSize =      null;
  private static Boolean _windowResizable =   null;
  private static Integer _defaultFontSize =   null;
  private static Boolean _specialUpdate =     null; 
  private static Boolean _noGraphics =        null;

  /**
   * Interprets the value of a system property to determine
   * if frames may be positioned or sized. This function should
   * be queried before each call to <code>setSize()</code>,
   * <code>setLocation()</code> or <code>setBounds()</code> of an
   * instance of <code>java.awt.Window</code>. Subsequent calls to
   * <code>pack()</code> or <code>setVisible(true)</code> should
   * also be avoided, if this function returns <code>true</code>.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.avoidFrameReshape</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Program-controlled positioning and sizing of AWT windows.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT><code>false</code> (default):</DT>
   *     <DD>All windows will be positioned and sized at our needs.
   *     </DD>
   *     <DT><code>true</code>:</DT>
   *     <DD>Sacrifices the comfort of perfectly positioned windows
   *         to get windows that display their contents correctly.
   *         <br>
   *         This may be needed if under some circumstances (JDK1.3
   *         and twm, for example) the default behaviour will lead to
   *         "jumping" windows or windows with hidden areas.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static boolean avoidFrameReshape() {
    if (_avoidFrameReshape == null) {
      String prop = System.getProperty("de.renew.avoidFrameReshape");
      if (prop == null) {
	prop = "false";
      } else if ("".equals(prop)) {
	prop = "true";
      } 
      _avoidFrameReshape = Boolean.valueOf(prop);;
      if (_avoidFrameReshape.booleanValue()) {
	System.out.println("Frame reshaping deactivated.");	 
      }      
    }
    return _avoidFrameReshape.booleanValue();
  }


  /**
   * Interprets the value of a system property to determine
   * if a drawing load server should be started.
   * This function returns a positive integer value that
   * specifies the port, if the property is set. 
   * It returns <code>-1</code>, if no load server should
   * be started.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.loadServerPort</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Setup of a drawing load server.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT>not set (default):</DT>
   *     <DD>No drawing load server will be started.
   *     </DD>
   *     <DT>positive integer:</DT>
   *     <DD>A drawing load server will be started.
   *         The integer value is interpreted as the port number
   *         where the server will listen for requests.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static int loadServerPort() {
    if (_loadServerPort == null) {
      _loadServerPort = new Integer(-1);
      String serverPortStr=System.getProperty("de.renew.loadServerPort");
      if (serverPortStr!=null) {
	try {
	  _loadServerPort=Integer.valueOf(serverPortStr);
	} catch (NumberFormatException e) {
	  System.err.println("Error reading property de.renew.loadServerPort="
			     +serverPortStr+": "+e);
	}
      }
    }
    return _loadServerPort.intValue();
  }


  /**
   * Interprets the value of a system property to determine
   * the font size for the application menus.
   * This function returns a positive integer value that
   * specifies the font size, if the property is set. 
   * It returns <code>-1</code>, if the default size should
   * be used.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.menuFontSize</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Application menu font size.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT>not set (default):</DT>
   *     <DD>The AWT-default font size will be used..
   *     </DD>
   *     <DT>positive integer:</DT>
   *     <DD>The integer value is interpreted as a font size
   *         to be used for menus.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static int menuFontSize() {
    if (_menuFontSize == null) {
      _menuFontSize = new Integer(-1);
      String fontSizeStr=System.getProperty("de.renew.menuFontSize");       
      if (fontSizeStr!=null) {
	try {
	  _menuFontSize=Integer.valueOf(fontSizeStr);
	} catch (NumberFormatException e) {
	  System.err.println("Error reading property de.renew.menuFontSize="
			     +fontSizeStr+": "+e);
	}
      }
    } 
    return _menuFontSize.intValue();
  }

  /**
   * Interprets the value of a system property to determine
   * if the main menu frame should be resizable.
   * <p>
   * <b>This property is no longer in use!</b>
   * The menu frame is resizable on all platforms since the
   * toolbars can be rearranged to fit the frame.
   * </p>
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.windowResizable</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Resizability of the menu frame.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT><code>true</code> (default on platform Windows):</DT>
   *     <DD>The menu frame will be resizable.
   *         This is needed on the Windows platform because the size
   *         of the menu frame does not match the size of its contents.
   *     </DD>
   *     <DT><code>false</code> (default on other platforms):</DT>
   *     <DD>The menu frame has a fixed size.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static boolean windowResizable() {
    if (_windowResizable == null) {    
      String windowResizableStr=System.getProperty("de.renew.windowResizable");
      if (windowResizableStr==null) {
	// If no parameter is given, the window is
	// resizable on Windows-Systems only.
	if (System.getProperty("os.name").indexOf("Win")>=0) {
	  windowResizableStr="true";
	}
      } else if ("".equals(windowResizableStr)) {
	windowResizableStr="true";
      }
      _windowResizable = Boolean.valueOf(windowResizableStr);
    }
    return _windowResizable.booleanValue();
  }

  /**
   * Interprets the value of a system property to determine
   * the font size for new text figures.
   * This function returns a positive integer value that
   * specifies the font size, if the property is set. 
   * It returns <code>-1</code>, if the default size should
   * be used.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.defaultFontSize</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Default font size for new text figures.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT>not set (default):</DT>
   *     <DD>The AWT-default font size will be used..
   *     </DD>
   *     <DT>positive integer:</DT>
   *     <DD>The integer value is interpreted as a font size
   *         to be used for new text figures.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static int defaultFontSize() {
    if (_defaultFontSize == null) {
      _defaultFontSize = new Integer(-1);
      String fontSizeStr=System.getProperty("de.renew.defaultFontSize");       
      if (fontSizeStr!=null) {
	try {
	  _defaultFontSize=Integer.valueOf(fontSizeStr);
	} catch (NumberFormatException e) {
	  System.err.println("Error reading property de.renew.defaultFontSize="
			     +fontSizeStr+": "+e);
	}
      }
    } 
    return _defaultFontSize.intValue();
  }

  /**
   * Interprets the value of a system property to determine
   * if the drawing graphics should be updated in a simple,
   * but flickering way.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>CH.ifa.draw.specialUpdate</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Graphics update strategy for drawings.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT><code>false</code> (default for JDK 1.0 and 1.1):</DT>
   *     <DD>The buffered update strategy will be used.
   *         This avoids flickering while the drawing is updated.
   *     </DD>
   *     <DT><code>true</code> (default for other JDKs):</DT>
   *     <DD>The simple update strategy will be used.
   *         This is needed for JDKs since version 1.2 because
   *         the buffered update causes errors.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static boolean specialUpdate() {
    if (_specialUpdate == null) {
      String prop = System.getProperty("CH.ifa.draw.specialUpdate");
      if (prop == null) {	 
	String javaVersion=System.getProperty("java.version");
	if (javaVersion==null) {
	  // This should not happen.
	  javaVersion="0.0";
	}
	String majorRelease=javaVersion.substring(0,3);
	_specialUpdate = new Boolean( !majorRelease.equals("1.0")
				   && !majorRelease.equals("1.1") );
      } else if ("".equals(prop)) {
	_specialUpdate = Boolean.TRUE;
      } else {
	_specialUpdate=Boolean.valueOf(prop);
      }
    }
    return _specialUpdate.booleanValue();
  }

  /**
   * Interprets the value of a system property to determine
   * whether graphics will be in use when loading a drawing.
   * <DL>
   * <DT><b>System property:</b></DT>
   * <DD><code>de.renew.noGraphics</code>
   * </DD>
   * <DT><b>Influences:</b></DT>
   * <DD>Graphics update when a drawing is loaded.
   * </DD>
   * <DT><b>Valid values:</b></DT>
   * <DD><DL><DT>not set (default):</DT>
   *     <DD>The display of figures will be updated on load.
   *     </DD>
   *     <DT>set:</DT>
   *     <DD>The display won't be actualized. This allows
   *         a programmer to load a drawing without starting
   *         the GUI. But it will cause exceptions if you
   *         try to draw the figures in this mode.
   *     </DD></DL>
   * </DD></DL>
   **/
  public static boolean noGraphics() {
    if (_noGraphics == null) {
      if (System.getProperty("de.renew.noGraphics") != null) {
	_noGraphics = Boolean.TRUE;
      } else {
	_noGraphics = Boolean.FALSE;
      }
    } 
    return _noGraphics.booleanValue();
  }
}

