/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package de.renew.application;

import de.renew.database.SetupHelper;
import de.renew.simulator.*;
import de.renew.shadow.*;
import de.renew.util.ParameteredCommandLine;
import java.io.*;


/**
 * Class documentation.
 */
public class ShadowSimulator
{
	/**
	 * Runs a simulator for a given shadow net system and a primary net to create an instance of.
	 * If the database backing VM properties are set, the net instances are loaded from
	 * the database.
	 * @param simulator The simulator to be run.
	 * @param netSystem The shadow net system, as exported from the CPNApplication GUI.
	 * @param primaryNetName The name of the net to create the primary instance of.
	 * The primary net instance is only created if there is no initialized simulation
	 * in the database or if there is no database at all.
	 * @param tracerFactory The tracer factory for the messages.
	 * @param run Whether to run the simulation (otherwise init it only).
	 * @exception SyntaxException The shadow net system cannot be compiled.
	 */
	public static void simulate(Simulator simulator, ShadowNetSystem netSystem,
		String primaryNetName, TracerFactory tracerFactory, boolean run)
		throws SyntaxException
	{
		ShadowLookup lookup = netSystem.compile();
		if (lookup == null)
		{
			throw new RuntimeException("Could not compile net system");
		}

		SetupHelper.SimulationState state = SetupHelper.setup(null, false);

		// We are starting or stopping the simulator before the primary net instance is created,
		// because the transaction strategy may clear its datasets while the
		// simulation becomes initialized. This would kill the primary net instance.
		if (run)
		{
			// Starts the simulator and updates the database simulation state.
			simulator.startRun();
		}
		else
		{
			// Stops the simulator and updates the database simulation state.
			simulator.stopRun();
		}

		// If the simulation state was not initialized in the database,
		// then we create the primary net instance.
		if (!state.wasSimulationInited())
		{
			Net primaryNet = lookup.getNet(primaryNetName);
			if (primaryNet == null)
			{
				throw new IllegalArgumentException("Net system does not contain the net " + primaryNetName);
			}

			NetInstance primaryInstance = SimulatorHelper.fireInitially(primaryNet,
				tracerFactory.getTracer("Initially", "(0) "), true);

			if (primaryInstance == null)
			{
				throw new RuntimeException("The net did not accept the init request.");
			}
		}
	}

	/**
	 * Runs a simulator for a given shadow net system file name and a primary
	 * net name to create an instance of.
	 * If the database backing VM properties are set, the net instances are loaded from
	 * the database.
	 * @param simulatorClass The name of the simulator class to be run.
	 * @param netSystemFileName The file name of the shadow net system,
	 * as exported from the CPNApplication GUI.
	 * @param primaryNetName The name of the net to create the primary instance of.
	 * The primary net instance is only created if there is no initialized simulation
	 * in the database or if there is no database at all.
	 * @param tracerFactory The tracer factory for the messages.
	 * @param run Whether to run the simulation (otherwise init it only).
	 * @return The regular simulator now running.
	 * @exception IllegalArgumentException A parameter is invalid.
	 * @exception IOException A read error occurred.
	 * @exception SyntaxException The shadow net system cannot be compiled.
	 */
	public static Simulator simulate(String simulatorClassName, String netSystemFileName,
		String primaryNetName, TracerFactory tracerFactory, boolean run)
		throws IllegalArgumentException, IOException, SyntaxException
	{
		Simulator simulator;
		try
		{
			Class simulatorClass = Class.forName(simulatorClassName);
			simulator = (Simulator) simulatorClass.getConstructor(new Class[] {TracerFactory.class,Boolean.TYPE})
				. newInstance(new Object[] {tracerFactory,Boolean.TRUE});
		}
		catch (Exception e)
		{
			throw new IllegalArgumentException("Invalid simulator class name "
				+ simulatorClassName + ", because " + e);
		}

		FileInputStream stream = new FileInputStream(netSystemFileName);
		ObjectInput input = new ObjectInputStream(stream);
		ShadowNetSystem netSystem;
		try
		{
			netSystem = (ShadowNetSystem) input.readObject();
		}
		catch (ClassNotFoundException e)
		{
			throw new IllegalArgumentException("Invalid shadow net system file name "
				+ netSystemFileName + ", because " + e);
		}

		// Set up net loading from .sns-files in the net path.
		netSystem.setNetLoader(new DefaultShadowNetLoader());
		Net.setNetLoader(new DefaultCompiledNetLoader(netSystem));

		simulate(simulator, netSystem, primaryNetName, tracerFactory, run);

		return simulator;
	}

	/**
	 * The command line shadow simulator entry point.
	 * @param args The command line arguments, as described in the help text.
	 */
	public static void main(String[] args)
		throws Exception
	{
		System.out.println("The Renew stand-alone shadow simulator");
		System.out.println("");

		ParameteredCommandLine line = new ParameteredCommandLine(args,
			new String[] {"-h", "-i"}, new int[] {0, 0});
		args = line.getRemainingArgs();

		if (args.length != 3 || line.hasParameter("-h"))
		{
			System.out.println("Parameters: <simulator> <net system> <primary net> [-i]");
			System.out.println("  <simulator>   : The java class of the simulator to be used.");
			System.out.println("                  (e.g. de.renew.simulator.ConcurrentSimulator)");
			System.out.println("  <net system>  : The exported shadow net system file.");
			System.out.println("  <primary net> : The name of the primary net to create an instance of.");
			System.out.println("  -i            : Initialize the simulation only, don't run it.");

			System.exit(0);
		}

		simulate(args[0], args[1], args[2], new ConsoleTracerFactory(), !line.hasParameter("-i"));
	}
}
