package de.renew.database;

import de.renew.database.*;
import de.renew.database.entitylayer.*;
import de.renew.gui.CPNSimulation;
import de.renew.remote.*;
import de.renew.shadow.*;
import de.renew.simulator.*;
import de.renew.util.*;
import java.io.*;
import java.lang.reflect.*;
import java.rmi.*;
import java.sql.*;
import java.util.*;


/**
 * This class contains some static utility methods that allow
 * the caller to reestablish a simulation state.
 */

public class SetupHelper
{
	/**
	 * This class is totally static. One must not create instances
	 * of it.
	 */
	private SetupHelper()
	{
	}

	/**
	 * Load a set of nets from disk in the form of a shadow net system.
	 * Then compile the net system to create the semantic net level.
	 *
	 * @param filename a string that denotes that file to be loaded
	 * @exception java.io.IOException if the file access failed
	 * @exception java.lang.ClassNotFoundException if the deserialization
	 *   failed
	 * @exception de.renew.shadow.SyntaxException if the loaded net did not
	 *   compile correctly.
	 */
	public static void setupNets(String filename)
		throws IOException, ClassNotFoundException, SyntaxException
	{
		// Read the shadow net system from disk.
		FileInputStream stream = new FileInputStream(filename);
		ObjectInput input = new ObjectInputStream(stream);
		ShadowNetSystem netSystem = (ShadowNetSystem) input.readObject();

		// Compile the net system.
		netSystem.compile();

		// The nets have automatically been registered and can be queried by
		// Net.forName(String).
	}

	/**
	 * This method restores a set of nets from a database source.
	 * For the correct usage of this method it is required
	 * that no simulator is currently running. Additionally,
	 * a transaction source must not yet have been established,
	 * otherwise the database might get corrupted.
	 *
	 * @param source the database source
	 *
	 * @return A NetInstanceMap of all restored NetInstances.
	 *
	 * @exception Exception If any exception occurs.
	 * These exceptions are thrown by the RestoreSource
	 * interface implementations.
	 */
	public static NetInstanceMap restoreNetInstances(RestoreSource source)
		throws Exception
	{
		// Create all net instances with their initial markings.
		String[] netIDs = source.getAllNetIDs();

		// Create arrays for the reestablished nets.
		int n = netIDs.length;
		Net[] nets = new Net[n];
		NetInstance[] instances = new NetInstance[n];
		NetInstanceMap map = new NetInstanceMap();

		// Create all net instances.
		for (int i = 0; i < n; i++)
		{
		        // Create using the initial marking.
			String id = netIDs[i];
			Net net=Net.forName(source.getNetName(id));
			nets[i] = net;
			NetInstance instance = net.buildInstance();
			instance.setID(id);
			instances[i] = instance;
			map.put(id,instance);

			// Remove old tokens and discard them.
			Enumeration places = net.places();
			while (places.hasMoreElements())
			{
				Place place = (Place) places.nextElement();
				PlaceInstance pi = instance.getInstance(place);

				pi.lock.lock();
				pi.extractAllTokens(new Vector(), new Vector());
				pi.lock.unlock();
			}
		}

		Hashtable idToToken=source.getTokens(map);

		// Establish all token IDs.
		Enumeration enum=idToToken.keys();
		while (enum.hasMoreElements()) {
		    String id=(String)enum.nextElement();
		    Object token=idToToken.get(id);
		    IDRegistry.getInstance().setAndReserveID(token,id);
		}

		// Change to the saved marking.
		for (int i = 0; i < n; i++)
		{
			NetInstance instance = instances[i];

			Enumeration places = nets[i].places();
			while (places.hasMoreElements())
			{
				Place place = (Place) places.nextElement();
				PlaceInstance pi = instance.getInstance(place);

				pi.lock.lock();
				// Add new tokens.
				Vector ids = new Vector();
				source.fillinAllTokens(pi, ids);
				int num = ids.size();
				for (int j = 0; j < num; j++)
				{
				        String id=(String)ids.elementAt(j);
				        Object token=idToToken.get(id);

				        // The time does not matter: time=0.
				        pi.insertTokenWithID(token,id,0);
				}
				pi.lock.unlock();
			}
		}

		// Make sure to make token IDs discardable.
		// No token IDs should be discarded right now,
		// but we have to undo one reservation.
		enum=idToToken.keys();
		while (enum.hasMoreElements()) {
		    String id=(String)enum.nextElement();
		    Object token=idToToken.get(id);
		    IDRegistry.getInstance().unreserve(token);
		}

		return map;
	}

	/**
	 * This method is one of the main entry points of this class.
	 * It simplifies the setup of a simulation
	 * from a database with certain net instances and tokens.
	 * Despite of the parametered overload, it takes all information
	 * from the system properties de.renew.simdb.driver,
	 * de.renew.simdb.dialect, de.renew.simdb.url, de.renew.simdb.user and
	 * de.renew.password. If they do not exist, no simulation is restored, so
	 * you can savely call this method in any case. It also reopens all open
	 * net instance drawings, so this method can be used for the CPNApplication
	 * database restore.
	 * @param simulation The simulation to init or run if it was inited or
	 * running. If this parameter is null, no simulation is reinited or rerun.
	 * @param verbose Whether to show a message for each database access.
	 * @return The previous simulation state object grouping the net instances
	 * with an open instance drawing and the last run state of the simulation.
	 */
	public static SimulationState setup(CPNSimulation simulation, boolean verbose)
	{
		if (verbose)
		{
			Entity.setMessageStream(System.err);
		}

		String dbDriver = System.getProperty("de.renew.simdb.driver");
		String dbDialect = System.getProperty("de.renew.simdb.dialect");
		String dbUrl = System.getProperty("de.renew.simdb.url");
		String dbUser = System.getProperty("de.renew.simdb.user");
		String dbPassword = System.getProperty("de.renew.simdb.password");

		if (dbUrl == null && dbDriver == null)
		{
			return SimulationState.TERMINATED_STATE;
		}

		if (dbUrl == null || dbDriver == null)
		{
			System.err.println("Error: Specify de.renew.simdb.url and"
				+ " de.renew.simdb.driver system properties together"
				+ " to use a simulation database.");
			return SimulationState.TERMINATED_STATE;
		}

		if (dbPassword == null)
		{
			dbPassword = "";
		}

		if (dbDialect == null)
		{
			dbDialect = "de.renew.database.entitylayer.SQLDialect";
			System.err.println("Warning: The system property"
				+ " de.renew.simdb.dialect has not been specified. Using"
				+ " default dialect " + dbDialect);
		}

		try
		{
			Class driverClass = Class.forName(dbDriver);
			Constructor constructor = driverClass.getConstructor(new Class[0]);
			constructor.newInstance(new Object[0]);
		}
		catch (ClassNotFoundException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class " + dbDriver + " not found.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (IllegalAccessException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class's default constructor"
				+ " is not accessable.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (IllegalArgumentException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class's default constructor"
				+ " cannot be invoked.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (InstantiationException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class is abstract or an interface.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (InvocationTargetException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class's default constructor"
				+ " threw an exception:\n" + e.getTargetException());
			return SimulationState.TERMINATED_STATE;
		}
		catch (NoSuchMethodException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class has no default constructor.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (SecurityException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "Database driver class's default constructor"
				+ " is not accessable.");
			return SimulationState.TERMINATED_STATE;
		}

		SQLDialect dialect = null;
		try
		{
			Class dialectClass = Class.forName(dbDialect);
			Constructor constructor = dialectClass.getConstructor(new Class[0]);
			dialect = (SQLDialect) constructor.newInstance(new Object[0]);
		}
		catch (ClassNotFoundException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class " + dbDialect + " not found.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (IllegalAccessException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class's default constructor"
				+ " is not accessable.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (IllegalArgumentException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class's default constructor"
				+ " cannot be invoked.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (InstantiationException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect is an abstract class or an interface.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (InvocationTargetException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class's default constructor"
				+ " threw an exception:\n" + e.getTargetException());
			return SimulationState.TERMINATED_STATE;
		}
		catch (NoSuchMethodException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class has no default constructor.");
			return SimulationState.TERMINATED_STATE;
		}
		catch (SecurityException e)
		{
			System.err.println("Cannot set simulation database.\n"
				+ "SQL dialect class's default constructor"
				+ " is not accessable.");
			return SimulationState.TERMINATED_STATE;
		}

		DatabaseRestoreSource restoreSource = null;
		List netInstances = new ArrayList();
		try
		{
			System.err.println("Connecting to simulation database " + dbUrl + "...");
			if (dbUser != null)
			{
				restoreSource = new DatabaseRestoreSource(dbUrl,
					dbUser, dbPassword, dialect);
			}
			else
			{
				restoreSource = new DatabaseRestoreSource(dbUrl, dialect);
			}
			System.err.println("Connected.");

			// We are using the database.
			// Make sure that all nets use a global IDRegistry,
			// so that we can keep a global list of tokens.
			de.renew.simulator.NetInstanceImpl.useGlobalIDRegistry=true;

			if (restoreSource.wasSimulationInited())
			{
				System.err.println("Loading simulation state...");

				// Make sure that all nets get compiled,
				// but do not create any instances.
				if (simulation != null)
				{
					simulation.initSimulation(null);
				}

				NetInstanceMap map = SetupHelper.restoreNetInstances(restoreSource);
				IDSource.setFactory(new TrivialIDFactory(restoreSource
					. getLastId() + 1));

				String[] viewedNetInstanceIds = restoreSource.getViewedNetIDs();
				for (int i=0; i<viewedNetInstanceIds.length; i++)
				{
					netInstances.add(map.get(viewedNetInstanceIds[i]));
				}

				System.err.println("Simulation state loaded.");
			}
		}
		catch (Exception e)
		{
			System.err.println("Cannot load simulation state.\n"
				+ "There was an exception:");
			e.printStackTrace();
			System.err.println("Simulation database abandonned.");
			if (simulation != null)
			{
				simulation.simulationTerminate();
			}
			return SimulationState.TERMINATED_STATE;
		}

		try
		{
			if (dbUser != null)
			{
				TransactionSource.setStrategy(new DatabaseTransactionStrategy(
					dbUrl, dbUser, dbPassword, dialect));
			}
			else
			{
				TransactionSource.setStrategy(new DatabaseTransactionStrategy(
					dbUrl, dialect));
			}
		}
		catch (Exception e)
		{
			System.err.println("Cannot assign simulation database.\n"
				+ "There was an exception:");
			e.printStackTrace();
			System.err.println("Simulation database abandonned.");
			if (simulation != null)
			{
				simulation.simulationTerminate();
			}
			return SimulationState.TERMINATED_STATE;
		}

		try
		{
			if (restoreSource.wasSimulationRunning() && simulation != null)
			{
				System.err.println("Rerunning simulation...");
				simulation.simulationRun();
			}
		}
		catch (Exception e)
		{
			System.err.println("Cannot rerun simulation.\n"
				+ "There was an exception:");
			e.printStackTrace();
			System.err.println("Simulation database abandonned.");
			if (simulation != null)
			{
				simulation.simulationTerminate();
			}
			return SimulationState.TERMINATED_STATE;
		}

		return new SimulationState(netInstances, restoreSource);
	}
	
	/**
	 * This class groups the net instances with an open instance drawing
	 * and the last run state of the simulation.
	 * Objects of this class are returned by the setup method.
	 */
	public static class SimulationState
	{
		/**
		 * The terminated simulation state. This is the initial
		 * state with no open net instances and no initialized simulation.
		 */
		private static final SimulationState TERMINATED_STATE
			= new SimulationState(Collections.EMPTY_LIST, null);
	
		/**
		 * A list of net instances that had an open net instance drawing.
		 */
		private List netInstances;

		/**
		 * Whether the simulation run state was inited.
		 */
		private boolean wasSimulationInitiedFlag;

		/**
		 * Whether the simulation run state was running.
		 */
		private boolean wasSimulationRunningFlag;

		/**
		 * Creates a new simulation state object.
		 * @param netInstances A list of net instances that had an
		 * open net instance drawing.
		 * @param source The restore source used to read the simulation state.
		 */
		private SimulationState(List netInstances, RestoreSource source)
		{
			this.netInstances = netInstances;
			try
			{
				wasSimulationInitiedFlag = source != null && source.wasSimulationInited();
				wasSimulationRunningFlag = source != null && source.wasSimulationRunning();
			}
			catch (Exception e)
			{
				System.err.println("Cannot determine simulation run state:");
				e.printStackTrace(System.err);
				wasSimulationInitiedFlag = false;
				wasSimulationRunningFlag = false;
			}
		}
		
		/**
		 * A list of net instances that had an open net instance drawing.
		 */
		public List getNetInstances()
		{
			return netInstances;
		}

		/**
		 * Returns whether the simulation run state was inited (not terminated).
		 * @return Whether the simulation run state was inited.
		 */
		public boolean wasSimulationInited()
		{
			return wasSimulationInitiedFlag;
		}
	
		/**
		 * Returns whether the simulation run state was running.
		 * @return Whether the simulation run state was running.
		 */
		public boolean wasSimulationRunning()
		{
			return wasSimulationRunningFlag;
		}
	}
}
