package de.renew.event;

import collections.*;
import de.renew.util.RunQueue;
import de.renew.util.Semaphor;

/**
 * A <code>ListenerSet</code> manages a collection
 * of event listeners and is able to distribute events
 * to all of its listeners.
 *
 * Synchronization is done on this object.
 *
 * When an event is delivered synchronously, it is ensured that
 * the event set is locked, but that changes to the set of listeners
 * do not effect the current notifications.
 *
 * This is a utility class that is not type safe. It should be
 * subclassed to get adequate functionality.
 *
 * @author <a href="mailto:kummer@informatik.uni-hamburg.de">Olaf Kummer</a>
 */
public class ListenerSet {
  /**
   * The <code>queue</code> is a global object that dispatches
   * events to those listener that do not wnat to be notifed
   * synchronously.
   **/
  private static RunQueue queue=new RunQueue();

	/**
	 * Delay the execution of the calling thread until
	 * all currently pending messages are delivered.
	 **/
	public static void awaitEmptyQueue() {
		final Semaphor s=new Semaphor();
		queue.add(new Runnable() {
				public void run() {
					s.V();
				}
			});
		s.P();
	}
		
  /**
   * The listeners to notify if events occur.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  synchronized void include(NetEventListener listener) {
    listeners.include(listener);
  }
  
  synchronized void exclude(NetEventListener listener) {
    listeners.exclude(listener);
  }

  /**
   * Call the dispatcher once for each registered listener.
   **/
  synchronized void dispatch(final ListenerSetDispatcher dispatcher) {
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      final NetEventListener listener=(NetEventListener)enum.nextElement();
      if (listener.wantSynchronousNotification()) {
				dispatcher.dispatchTo(listener);
      } else {
				queue.add(new Runnable(){
						public void run() {
							dispatcher.dispatchTo(listener);
						}
					});
      }
    }    
  }
}
