package de.renew.gui;

import CH.ifa.draw.framework.DrawingEditor;
import de.renew.util.StringUtil;
import java.io.File;
import java.util.Vector;
import java.util.Iterator;
import CH.ifa.draw.framework.Drawing;


/**
 * Keeps track of open <code>CPNDrawing</code>s and allows to get
 * a drawing for a known net.
 * <p>
 * </p>
 * CPNDrawingLoader.java
 * Created: Mon Dec  3  2001
 * @author Michael Duvigneau
 **/
class CPNDrawingLoader {

	/**
	 * The application to call when a new drawing has to be opened.
	 **/
	private CPNApplication app;

	/**
	 * An array of path strings denoting the directories where to
	 * look for drawing files.
	 **/
	private String[] netSource;

	/**
	 * The list of all known drawings. It is kept up to date by
	 * calls from the application. All actions of the loader are
	 * synchronized against this object.
	 **/
	private Vector drawings = new Vector();

	/**
	 * This field transports the information about a new drawing to
	 * the {@link #findDrawingFile} method. 
	 * <p>
	 * This field may be accessed only when synchronized by the
	 * {@link #drawings} monitor.
	 * </p>
	 **/
	private CPNDrawing newDrawing = null;

	private static boolean DEBUG = false;

	/**
	 * Creates a new drawing loader connected to the given
	 * application. The net load path is extracted from the system
	 * property <code>de.renew.netPath</code>.
	 **/
	CPNDrawingLoader(CPNApplication app) {
		this.app = app;
		setNetPath(System.getProperty("de.renew.netPath", System.getProperty("user.dir")));
		DEBUG = !("false".equalsIgnoreCase(System.getProperty("debug","false")));
	}

	/**
	 * Sets a search path (like the CLASSPATH) to look for net
	 * drawing files when a drawing is missing. The directories in
	 * the path are separated by {@link File#pathSeparatorChar}.
	 **/
	void setNetPath(String path) {
    setNetPath(StringUtil.splitPaths(path));
	}

	/**
	 * Sets search paths (like the CLASSPATH) to look for net
	 * drawing files when a drawing is missing. Each String in the
	 * array denotes exactly one directory to search.
	 **/
	void setNetPath(String[] paths) {
		if (paths == null) {
			this.netSource = new String[0];
			return;
		}
		this.netSource = new String[paths.length];
    for (int i=0;i<paths.length;++i) {
			netSource[i]=StringUtil.makeCanonical(paths[i]);
      if (DEBUG) System.out.println("Drawing loader source: "+netSource[i]);
    }
	}

	/**
	 * Returns the search paths in their canonical representation.
	 **/
	String[] getNetPath() {
		return netSource;
	}

	/**
	 * Informs the drawing loader that the given drawing has been
	 * loaded by the application.
	 *
	 * Only called from CPNApplication!
	 *
	 * @param drawing  the drawing to be registered.
	 **/
	void addDrawing(CPNDrawing drawing) {
		synchronized (drawings) {
			drawings.addElement(drawing);
			newDrawing = drawing;
		}
	}

	/**
	 * Informs the drawing loader that the given drawing has been
	 * closed in the application.
	 *
	 * Only called from CPNApplication!
	 *
	 * @param drawing  the drawing to be unregistered.
	 */
	void releaseDrawing(CPNDrawing drawing) {
		synchronized (drawings) {
			drawings.removeElement(drawing);
		}
		drawing.discardShadow();
	}

	/**
	 * Returns an iterator over all drawings currently loaded and
	 * known to the loader.
	 **/
	Iterator loadedDrawings() {
		return drawings.iterator();
	}

	/**
	 * Returns a drawing for the given net name. If no matching
	 * drawing is known, tries to load the drawing from a matching
	 * file (see {@link #findDrawingFile}).
	 *
	 * @return
	 *   the drawing with the given name - 
	 *   <code>null</code>, if it no matching drawing or file could
	 *   be found.
	 **/	
	CPNDrawing getDrawing(String name)
	{
		if (name == null) {
			return null;
		}
		synchronized (drawings) {
			for (int i = 0; i < drawings.size(); i++) {
				CPNDrawing currentDrawing = (CPNDrawing) drawings.elementAt(i);
				String currentName = currentDrawing.getName();

				if (name.equals(currentName)) {
					return currentDrawing;
				}
			}
			CPNDrawing loaded = findDrawingFile(name);
			if (loaded != null) {
				return loaded;
			}
		}
		return null;
	}

	/**
	 * Tries to find and load a drawing for the given net name.
	 * The drawing is loaded and opened in the application.
	 * It is not checked whether a net drawing with the given name
	 * is already loaded.
	 *
	 * @return
	 *   <code>true</code>, if a drawing was loaded -
	 *   <code>false</code>, if it no matching file could be found.
	 **/	
	CPNDrawing findDrawingFile(String name) {
		String path;
    File file;
		boolean error;
		for (int i=0; i<netSource.length; ++i) {
			error = false;
			path = netSource[i] + File.separator + name + ".rnw";
			try {
				file = new File(path);
				error = !file.canRead();
			} catch (Exception e) {
				System.err.println("CPN drawing loader: " + path
					+ " caused a " + e);
				error = true;
			}
			if (! error) {
				synchronized(drawings) {
					newDrawing = null;
					Drawing frontDrawing = app.drawing();
					app.loadAndOpenDrawing(path);
					app.showDrawingViewFrame(frontDrawing);
					// addDrawing() will set newDrawing when called by app.
					// No other thread is able to do that because we are
					// already synchronized with the drawings vector.
					if (newDrawing != null) {
						return newDrawing;
					}
				}
			}
		}
		return null;
	}

}
