/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package de.renew.gui;

import CH.ifa.draw.framework.DrawingEditor;
import CH.ifa.draw.framework.FigureWithID;
import de.renew.shadow.*;
import de.renew.simulator.*;
import de.renew.util.SelectiveClassLoader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Iterator;
import java.util.ConcurrentModificationException;


/**
 * Class documentation.
 */
public class CPNSimulation
{
	private ShadowNetSystem netSystem;
	private Simulator simulator = null;
	private boolean onlySequential;
	private BreakpointManager breakpointManager = null;
	private CPNDrawingLoader drawingLoader;

	/**
	 * Constructor documentation.
	 * @param compiler
	 * @param onlySequential
	 */
	public CPNSimulation(ShadowCompiler compiler, boolean onlySequential, CPNDrawingLoader loader)
	{
		if (classReinit)
		{

			// The classes are supposed to be reinitialized each time
			// the compiler is accessed. This is realized by wrapping the compiler
			// in an object that forwards all calls, but accesses
			// the class source before doing so.
			compiler = new ClassSourceResetCompiler(compiler);
		}
		netSystem = new ShadowNetSystem(compiler, new GuiShadowNetLoader(loader));
		Net.setNetLoader(new DefaultCompiledNetLoader(netSystem));
		this.onlySequential = onlySequential;
		this.drawingLoader = loader;
	}

	/**
	 * Method documentation.
	 * @return
	 */
	public boolean isStrictlySequential()
	{
		return onlySequential;
	}



	/**
	 * Method documentation.
	 */
	public void buildAllShadows()
	{
		boolean again = true;
		while (again) {
			again = false;
			Iterator drawings = drawingLoader.loadedDrawings();

			try {
				// Build new shadows of all nets and all elements.
				// Do everything in a fine order:
				while (drawings.hasNext()) {
					CPNDrawing currentDrawing = (CPNDrawing) drawings.next();
					currentDrawing.buildShadow(netSystem);
				}
			} catch (ConcurrentModificationException e) {
				System.err.println("CPNSimulation.buildAllShadows(): Concurrent modification. Redo from start...");
				again = true;
			}
		}
	}

	/**
	 * Method documentation.
	 * @return
	 */
	public ShadowNetSystem getNetSystem()
	{

		// Make sure that all shadows are correctly set up.
		// buildAllShadows();
		// Return the shadow net system.
		return netSystem;
	}

	CPNDrawingLoader getDrawingLoader() {
		return drawingLoader;
	}
		
	/**
	 * Method documentation.
	 * @param newBreakpointManager
	 */
	void setBreakpointManager(BreakpointManager newBreakpointManager)
	{
		breakpointManager = newBreakpointManager;
	}

	/**
	 * Returns the actual breakpoint manager.
	 * Attention: May be <code>null</code>!
	 */
	public BreakpointManager getBreakpointManager()
	{
		return breakpointManager;
	}

	/**
	 * Method documentation.
	 * @return
	 * @throws SyntaxException
	 */
	private ShadowLookup compile() throws SyntaxException
	{

		// Make sure to recreate all shadows, so that changes of
		// the drawings are incorporated.
		buildAllShadows();

		// Make sure all old Net entries in the name lookup
		// are discarded.
		de.renew.simulator.Net.forgetAllNets();

		// Compile and flatten the lookup table.
		ShadowLookup lookup = netSystem.compile();

		return lookup;
	}

	/**
	 * Method documentation.
	 * @throws SyntaxException
	 */
	public void syntaxCheckOnly() throws SyntaxException
	{
		compile();
	}

	/**
	 * Start a new simulation.
	 * Return the initially created NetInstance,
	 * if the creation of the initial instance
	 * was successful, or null, if no initial creation took
	 * place.
	 */
	public NetInstance initSimulation(String mainNet) throws SyntaxException
	{
		simulationTerminate();

		// Compile net.
		ShadowLookup shadowLookup = compile();

		// Rig compiled net with breakpoints.
		if (breakpointManager != null)
		{
			breakpointManager.addPresetBreakpoints(shadowLookup);
			breakpointManager.clearLog();
		}

		simulator = null;
		newSimulator();

		if (mainNet != null)
		{
			Net net = shadowLookup.getNet(mainNet);
			if (net != null) {

				// Create an initial net instance.
				NetInstance primaryInstance = de.renew.simulator.SimulatorHelper.fireInitially(net,
						new ConsoleTracer("Initially", "(0) "), true);

				if (simulator != null)
				{
					if (primaryInstance == null)
					{

						// A mainNet was requested, but couldn't
						// be created by firing initially.
						// So stop the simulation.
						simulator.stopRun();
					}
					else
					{

						// A mainNet was requested and created.
						// Make sure the simulator recognizes any changes
						// that might have happened during the firing.
						simulator.refresh();
					}
				}
				return primaryInstance;
			}
		}

		return null;
	}

	/**
	 * Create a new simulator.
	 * The actual object that is created depends
	 * on the desired multiplicity. A zero multiplicity
	 * denotes a sequential simulator, which is minimally
	 * concurrent.
	 */
	private void newSimulator()
	{
		if (simulatorMultiplicity == 0 || onlySequential)
		{
			simulator = new SequentialSimulator(new ConsoleTracerFactory(""),!eagerSimulation);
		}
		else if (simulatorMultiplicity == 1)
		{
			simulator = new ConcurrentSimulator(new ConsoleTracerFactory(""),!eagerSimulation);
		}
		else
		{
			simulator = new ParallelSimulator(simulatorMultiplicity,!eagerSimulation);
		}
	}


	/**
	 * Version history:
	 * 1  no header, included NetInstances and SearchQueue
	 * 2  header(label, version, simulator type), NetInstances,
	 *    Nets, SearchQueue - since Renew 1.2 beta 11
	 * 3  small changes to ReflectionSerializer (Class[], nulls)
	 *    can still read streams of version 2
	 * 4  marking and search queue are now saved with time stamps
	 *    incompatible change
	 * 5  PlaceInstance is now an abstract class with subclasses
	 *    incompatible change
	 *(6) used in branch agent_serialization
	 * 7  Different ID handling for net elements, introduction of
	 *    remote layer - incompatible change
	 */
	private static final int STATE_STREAM_VERSION = 7;
	private static final String STATE_STREAM_LABEL = "RenewState";

	/**
	 * Writes all currently known <code>NetInstance</code>s,
	 * <code>Net</code>s, the <code>SearchQueue</code> contents
	 * and some additional information to the stream. The written
	 * information is sufficient to continue the simulation from
	 * the same state after deserialization.
	 * <p>
	 * If the given <code>ObjectOutput</code> is a <code><b>
	 * de.renew.util.RenewObjectOutputStream</code></b>, its
	 * feature of cutting down the recursion depth by delaying
	 * the serialization of some fields will be used.
	 * </p><p>
	 * <b>Side effect:</b> Any open <code>BindingSelectionFrame</code>
	 * will be closed on execution of this method!
	 * </p>
	 *
	 * Added Feb 29 2000  Michael Duvigneau
	 *
	 * @param output target stream (see note about RenewObjectOutputStream above)
	 */
	public void saveState(java.io.ObjectOutput output) throws java.io.IOException
	{

		// I don't want concurrent changes to the simulation
		// state, so I have to synchronize...
		synchronized (this)
		{

			if (simulator != null)
			{
				simulationStop();

				// Header, contains:
				// - label
				// - file format version number
				// - type of simulator used
				output.writeObject(STATE_STREAM_LABEL);
				output.writeInt(STATE_STREAM_VERSION);
				output.writeInt(simulatorMultiplicity);

				// First part: save all NetInstances currently shown
				// in drawings. Although they are the only visible
				// part of the simulation, they are not necessarily
				// sufficient to describe the simulation state
				// completely.
				de.renew.gui.CPNInstanceDrawing.saveAllInstances(output);

				// Second part: save all Nets currently known by
				// the Net.forName() lookup mechanism.
				// This ensures that the static part of all compiled
				// nets will be available on deserialization.
				de.renew.simulator.Net.saveAllNets(output);

				// Third part: add all entries from the SearchQueue.
				// These entries alone are sufficient to describe the
				// current simulation state completely.
				// But this information does not neccessarily include all
				// nets possibly required by future simulation steps.
				de.renew.simulator.SearchQueue.saveQueue(output);

				// Last part: Restore the ID registry with its well-known
				// IDs for every token.
				de.renew.simulator.IDRegistry.save(output);
			}
		}
	}


	/**
	 * Restores a simulation saved by <code>saveState()</code>.
	 * Reads all stored <code>NetInstance</code>s and tries to
	 * reopen corresponding <code>CPNInstanceDrawing</code>s.
	 * The result is a ready-to-run simulation.
	 * <p>
	 * If the given <code>ObjectInput</code> is a <code><b>
	 * de.renew.util.RenewObjectInputStream</code></b>, the
	 * neccessary steps to cover delayed serialization will
	 * be made.
	 * </p><p>
	 * <b>Attention:</b> Any running simulation will be terminated!
	 * </p><p>
	 * The ObjectInputStream will be read using
	 * <code>de.renew.util.ClassSource</code> to provide
	 * its ability of reloading all user defined classes.
	 * </p>
	 *
	 * Added Apr 11 2000  Michael Duvigneau
	 *
	 * @param input source stream (see note about RenewObjectInputStream above)
	 * @param editor destination to open instance drawings within.
	 * Currently the only known editor is CPNApplication.
	 */
	public void loadState(java.io.ObjectInput input,
						  CPNApplication editor) throws java.io.IOException, ClassNotFoundException
	{

		// I don't want concurrent changes to the simulation
		// state, so I have to synchronize...
		synchronized (this)
		{

			simulationTerminate();

			// As we do not use any compiler, we have to install
			// a new ClassLoader by our own, if this behaviour is
			// requested.
			if (classReinit)
			{
				ClassSourceResetCompiler.setupClassSource();

				// Also we have to make sure all old Net entries in
				// the name lookup are discarded.
			}
			de.renew.simulator.Net.forgetAllNets();

			System.out.println("Loading simulation state...");

			Enumeration instances;

			try
			{

				// Check for valid header, which includes:
				// - label
				// - file format version number
				// - type of simulator used
				String streamLabel = (String) input.readObject();

				if (!streamLabel.equals(STATE_STREAM_LABEL))
				{
					throw new java.io.StreamCorruptedException("Stream does not seem to contain renew state data.");
				}
				int streamVersion = input.readInt();

				if (streamVersion == STATE_STREAM_VERSION)
				{

					// that's ok.
				}
				else if ((streamVersion == 2) && (STATE_STREAM_VERSION == 3))
				{

					// that's ok, too.
				}
				else
				{

					// that's most probably not ok.
					System.err.println("Warning: State data is of different version " + "("
									   + streamVersion + ") than the current ("
									   + STATE_STREAM_VERSION + ").");
					System.err.println("Trying to read it anyway, expecting errors.");
				}
				int streamSimulatorMultiplicity = input.readInt();

				if (streamSimulatorMultiplicity != simulatorMultiplicity)
				{
					System.err.println("Warning: Simulation state was saved "
									   + "using a different simulator multiplicity " + "("
									   + streamSimulatorMultiplicity + ") "
									   + "than currently selected (" + simulatorMultiplicity
									   + ").");
				}

				// First part: read all NetInstances to be shown
				// in drawings.
				instances = de.renew.gui.CPNInstanceDrawing.loadInstances(input);

				// Second part: read all compiled Nets
				de.renew.simulator.Net.loadNets(input);

				// Third part: add all neccessary entries to the
				// SearchQueue.
				de.renew.simulator.SearchQueue.loadQueue(input);

				// Last part: reestablish the global Id registry for
				// token IDs.
				de.renew.simulator.IDRegistry.load(input);
			}
			catch (java.io.IOException e)
			{

				// If an exception occurs, pass it to the caller.
				// But restore a known state first - the 'no
				// simulation active' state.
				simulationTerminate();
				throw e;
			}
			catch (ClassNotFoundException e)
			{

				// The same as above, but for another exception type...
				simulationTerminate();
				throw e;
			}
			catch (StackOverflowError e)
			{

				// The same as above, but for another exception type...
				simulationTerminate();
				throw e;
			}

			// Create a new simulator.
			newSimulator();

			// Now try to open CPNInstanceDrawings
			// for all net instances which were loaded.
			NetInstance currentInstance;

			while (instances.hasMoreElements())
			{
				currentInstance = (NetInstance) instances.nextElement();
				((CPNApplication) editor).openInstanceDrawing(currentInstance);
			}

		}
	}

	/**
	 * Method documentation.
	 * @return
	 */
	public boolean isSimulationActive()
	{
		return simulator != null && simulator.isActive();
	}

	/**
	 * Method documentation.
	 */
	public void simulationTerminate()
	{

		// To avoid simulation termination concurrent to serialization
		synchronized (this)
		{
			if (simulator != null)
			{
				Simulator oldSimulator = simulator;

				simulator = null;
				oldSimulator.terminateRun();
			}

			// Go back to simulation time 0 and
			// clear all outstanding search requests.
			de.renew.simulator.SearchQueue.reset(0);

			// Clear all channel watchers.
			de.renew.watch.ChannelSupervisor.reset();

			// Clear all token IDs that might be
			// still registered.
			de.renew.simulator.IDRegistry.reset();

			// Clear all old breakpoints - they won't work in any
			// new simulation and their existence prevents garbage
			// collection of old net instances.
			if (breakpointManager != null)
			{
				breakpointManager.deleteAllBreakpoints();
				breakpointManager.clearLog();
			}
		}
	}

	// To be called after an external firing of a transition was initiated.
	/**
	 * Method documentation.
	 */
	public void simulationRefresh()
	{
		if (simulator != null)
		{
			simulator.refresh();
		}
	}

	/**
	 * Method documentation.
	 */
	public void simulationStop()
	{
		if (simulator != null)
		{
			simulator.stopRun();
		}
	}

	/**
	 * Method documentation.
	 */
	public void simulationRun()
	{

		// To avoid simulation concurrent to serialization
		synchronized (this)
		{
			if (breakpointManager != null)
			{
				breakpointManager.clearLog();
			}
			if (simulator != null)
			{
				simulator.startRun();
			}
		}
	}

	/**
	 * Method documentation.
	 * @return
	 */
	public int simulationStep()
	{

		// To avoid simulation concurrent to serialization
		synchronized (this)
		{
			if (breakpointManager != null)
			{
				breakpointManager.clearLog();
			}
			if (simulator != null)
			{
				return simulator.step();
			}
			else
			{
				return Simulator.statusDisabled;
			}
		}
	}

	private static boolean classReinit;
	private static boolean eagerSimulation;
	private static int simulatorMultiplicity;

	static
	{
		classReinit = Boolean.getBoolean("de.renew.classReinit");
		if (classReinit)
		{
			System.err.println("Using class reinit mode.");
		}

		eagerSimulation = Boolean.getBoolean("de.renew.eagerSimulation");
		if (eagerSimulation)
		{
			System.err.println("Using eager simulation mode.");
		}

		simulatorMultiplicity = Integer.getInteger("de.renew.simulatorMode", 1).intValue();

		if (simulatorMultiplicity == 0 || simulatorMultiplicity == -1)
		{
			simulatorMultiplicity = 0;
			System.err.println("Using sequential simulator ...");
		}
		else if (simulatorMultiplicity < -1)
		{
			System.err.println("Using " + (-simulatorMultiplicity) + " sequential simulators ...");
			System.err.println("Caution! This is an experimental feature!");
		}
		else if (simulatorMultiplicity > 1)
		{
			System.err.println("Using " + simulatorMultiplicity + " concurrent simulators ...");
			System.err.println("Caution! This is an experimental feature!");
		}
	}
}
