package de.renew.remote;

import java.rmi.*;
import de.renew.simulator.NetInstance;
import de.renew.gui.RenewMode;
import CH.ifa.draw.framework.Figure;


/**
 * An object accessor allows to view some attributes
 * of its associated object.
 *
 * The associated object is not accessible directly but the
 * accessor provides String descriptions of the object as a
 * whole and its fields. This is somewhat like but less
 * powerful than what the java reflection api provides.
 *
 * <p>
 * This is a remote interface that allows an implementation
 * that can be used via RMI.
 * The intention is to keep the associated object on the
 * server and transmit only its description as strings.
 * </p>
 *
 * @author Olaf Kummer
 */
public interface ObjectAccessor extends Remote
{
	/**
	 * Returns a net instance accessor if the the object the accessor
	 * encapsulates represents a net instance.
	 * Otherwise, a ClassCastException is thrown.
	 * @return The net instance accessor.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 * @exception java.lang.ClassCastException If the object is no net instance.
	 */
	public NetInstanceAccessor asNetInstance()
		throws RemoteException, ClassCastException;

	/**
	 * Return the string representation of the accessed object as
	 * generated by its <code>toString()</code> method.
	 *
	 * @return the object's string representation
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public String asString()
		throws RemoteException;

  /**
   * Return an object accessor for the value of the
   * field specified by the given index.
   * Returns <code>null</code> if the associated object of
   * this accessor is a primitive value or null reference.
   *
   * @param i the index of the field to be returned
   * @return an object accessor for the specified field's value
   * @exception java.lang.IllegalAccessException The field is not accessable
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public ObjectAccessor getField(int i)
		throws RemoteException, IllegalAccessException;

  /**
   * Return the number of public fields of the associated object.
   * Always zero for primitive values and null references.
   *
   * @return the number of public fields
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public int getFieldCount() throws RemoteException;

  /**
   * Return the names of all fields of the associated object.
   * Always an empty array for primitive values and null references.
   *
   * @return an array of field names
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public String[] getFieldNames() throws RemoteException;

  /**
   * Return the values of all fields of the associated
   * object as a string representation.
   * This produces the same result as getting all fields one
   * by one via the <code>getField</code> method and putting
   * their <code>asString</code> results into an array.
   *
   * Returns always an empty array for primitive values and
   * null references.
   *
   * @return an array of field value string descriptions
   * @exception java.lang.IllegalAccessException The field is not accessable
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public String[] getFieldValues()
		throws RemoteException, IllegalAccessException;

	/**
	 * Creates a new figure for the wrapped object,
	 * by passing it to the given mode.
	 * Since the object is only known by the server,
	 * and the figure is needed by the client, this is
	 * the solution to get a figure for an object.
	 * Both the RenewMode and the Figure are serializable,
	 * so the mode is serialized, the figure is created
	 * using the RenewMode.getTokenFigure() on the server,
	 * and then the figure is also serialized.
	 * @param mode The mode to be used.
	 * @param expanded Whether to return the expanded form.
	 * @return The new figure.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public Figure getFigureFromMode(RenewMode mode, boolean expanded)
		throws RemoteException;

	/**
	 * Returns whether the object the accessor encapsulates
	 * is an instance of a given class.
	 * @param testClass The class.
	 * @return Whether the object represents such an instance.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public boolean isInstanceOf(Class testClass)
		throws RemoteException;
}
