package de.renew.remote;

import java.io.IOException;
import java.io.ObjectOutput;
import java.lang.reflect.*;
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import de.renew.simulator.NetInstance;
import de.renew.gui.RenewMode;
import CH.ifa.draw.framework.Figure;


/**
 * Implements the MarkingAccessor interface,
 * representing the marking of a place instance as a snapshot.
 * Its contents will not change even if transitions fire.
 *
 * @author Thomas Jacob
 */
public class ObjectAccessorImpl extends UnicastRemoteObject implements ObjectAccessor
{
	/**
	 * The object for the accessor.
	 */
	protected final Object object;

	/**
	 * Creates a new ObjectAccessor.
	 * @param object The object for the accessor.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public ObjectAccessorImpl(Object object)
		throws RemoteException
	{
		super(0, SocketFactoryDeterminer.getInstance(), SocketFactoryDeterminer.getInstance());
		this.object = object;
	}

	/**
	 * Returns a net instance accessor if the the object the accessor
	 * encapsulates represents a net instance.
	 * Otherwise, a ClassCastException is thrown.
	 * @return The net instance accessor.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 * @exception java.lang.ClassCastException If the object is no net instance.
	 */
	public NetInstanceAccessor asNetInstance()
		throws RemoteException, ClassCastException
	{
		return new NetInstanceAccessorImpl((NetInstance) object);
	}

	/**
	 * Return the string representation of the accessed object as
	 * generated by its <code>toString()</code> method.
	 *
	 * @return the object's string representation
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public String asString()
		throws RemoteException
	{
		return object.toString();
	}

	/**
	 * Return the number of public fields of the associated object.
	 * Always zero for primitive values and null references.
	 *
	 * @return the number of public fields
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public int getFieldCount()
		throws RemoteException
	{
		return object.getClass().getFields().length;
	}

	/**
	 * Return the names of all fields of the associated object.
	 * Always an empty array for primitive values and null references.
	 *
	 * @return an array of field names
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public String[] getFieldNames()
		throws RemoteException
	{
		Field[] fields = object.getClass().getFields();
		String[] names = new String[fields.length];

		for (int fieldNr=0; fieldNr < fields.length; fieldNr++)
			names[fieldNr] = fields[fieldNr].getName();

		return names;
	}

	/**
	 * Return the values of all fields of the associated
	 * object as a string representation.
	 * This produces the same result as getting all fields one
	 * by one via the <code>getField</code> method and putting
	 * their <code>asString</code> results into an array.
	 *
	 * Returns always an empty array for primitive values and
	 * null references.
	 *
	 * @return an array of field value string descriptions
	 * @exception java.rmi.RemoteException if an RMI failure occured.
 	 * @exception java.lang.IllegalAccessException The field is not accessable
	 */
	public String[] getFieldValues()
		throws RemoteException, IllegalAccessException
	{
		Field[] fields = object.getClass().getFields();
		String[] values = new String[fields.length];

		for (int fieldNr=0; fieldNr < fields.length; fieldNr++)
			values[fieldNr] = fields[fieldNr].get(object).toString();

		return values;
	}

	/**
	 * Return an object accessor for the value of the
	 * field specified by the given index.
	 * Returns <code>null</code> if the associated object of
	 * this accessor is a primitive value or null reference.
	 *
	 * @param i the index of the field to be returned
	 * @return an object accessor for the specified field's value
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 * @exception java.lang.IllegalAccessException The field is not accessable
	 */
	public ObjectAccessor getField(int i)
		throws RemoteException, IllegalAccessException
	{
		return new ObjectAccessorImpl(object.getClass().getFields()[i].get(object));
	}

	/**
	 * Creates a new figure for the wrapped object,
	 * by passing it to the given mode.
	 * Since the object is only known by the server,
	 * and the figure is needed by the client, this is
	 * the solution to get a figure for an object.
	 * Both the RenewMode and the Figure are serializable,
	 * so the mode is serialized, the figure is created
	 * using the RenewMode.getTokenFigure() on the server,
	 * and then the figure is also serialized.
	 * @param mode The mode to be used.
	 * @param expanded Whether to return the expanded form.
	 * @return The new figure.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public Figure getFigureFromMode(RenewMode mode, boolean expanded)
		throws RemoteException
	{
		return mode.getTokenFigure(object, expanded);
	}

	/**
	 * Returns whether the object the accessor encapsulates
	 * is an instance of a given class.
	 * @param testClass The class.
	 * @return Whether the object represents such an instance.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public boolean isInstanceOf(Class testClass)
		throws RemoteException
	{
		return testClass.isInstance(object);
	}

	/**
	 * Serializes the object contained in this accessor
	 * to a given object output stream.
	 *
	 * @param out    The stream to write to.
	 *
	 * @exception IOException 
	 *   if an I/O problem occurs.
	 */
	public void writeTo(ObjectOutput out)
		throws IOException
	{
		out.writeObject(object);
	}
}
