/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package de.renew.remote;

import collections.CollectionEnumeration;
import collections.HashedSet;
import collections.UpdatableSet;
import de.renew.event.PlaceEvent;
import de.renew.event.PlaceEventListener;
import de.renew.event.TokenEvent;
import de.renew.simulator.*;
import java.io.Serializable;
import java.rmi.*;
import java.rmi.server.*;


/**
 * This class implements the <code>PlaceInstanceAccessor</code>
 * interface and nothing more than needed to implement it.
 * <p>
 *
 * </p>
 * PlaceInstanceAccessorImpl.java
 * Created: Sun Jul 16  2000
 * @author Michael Duvigneau
 */

public class PlaceInstanceAccessorImpl extends ObjectAccessorImpl implements PlaceEventListener,
		PlaceInstanceAccessor
{
	/**
	 * Creates a new place instance accessor for the given place
	 * instance.
	 *
	 * @param placeInstance the place instance to access
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public PlaceInstanceAccessorImpl(PlaceInstance place) throws RemoteException
	{
		super(place);
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public NetElementID getID() throws RemoteException
	{
		return ((PlaceInstance) object).getPlace().getID();
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public PlaceAccessor getPlace() throws RemoteException
	{
		return new PlaceAccessorImpl(((PlaceInstance) object).getPlace());
	}

	/**
	 * Returns the place instance, if the caller knows that
	 * this is the local representation. This is required for setting breakpoints.
	 * @return The place instance.
	 */
	public PlaceInstance getPlaceInstance()
	{
		return ((PlaceInstance) object);
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public NetInstanceAccessor getNetInstance() throws RemoteException
	{
		return new NetInstanceAccessorImpl(((PlaceInstance) object).getNetInstance());
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public MarkingAccessor getMarking() throws RemoteException
	{
		return new MarkingAccessorImpl(((PlaceInstance) object));
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public TokenCountsAccessor getTokenCounts() throws RemoteException
	{
		return new TokenCountsAccessorImpl(((PlaceInstance) object));
	}

	/* This method is specified by the PlaceInstanceAccessor interface. */
	/**
	 * Method documentation.
	 * @param token
	 * @throws RemoteException
	 */
	public void addSerializableToken(Serializable token) throws RemoteException
	{

		// TODO: implement this de.renew.remote.PlaceInstanceAccessor method
	}

	// ------------------------------------ Place event forwarding ----

	/**
	 * The set of registered remote event listeners that want
	 * to be informed about place events.
	 */
	private UpdatableSet listeners = new HashedSet();

	/**
	 * Registers the given remote event listener so that it will
	 * receive future remote event messages concerning the accessed
	 * place instance.
	 * <p>
	 * If the listener is registered already, the additional
	 * registration try is ignored. The <code>equals()</code>
	 * method serves as indicator to allow the mechanism to
	 * work also for remote objects.
	 * </p>
	 * <p>
	 * This method is specified by the <code>RemoteEventProducer</code>
	 * interface which is required by the <code>PlaceInstanceAccessor</code>
	 * interface.
	 * </p>
	 *
	 * @param listener the listener to register
	 * @exception java.rmi.RemoteException if a RMI failure occured.
	 */
	public void addRemoteEventListener(RemoteEventListener listener) throws RemoteException
	{
		// We register this accessor as PlaceEventListener
		// when adding the first remote event listener.
		if (listeners.isEmpty())
			((PlaceInstance) object).addPlaceEventListener(this);

		listeners.include(listener);
	}

	/**
	 * Unregisters the given remote event listener so that it will
	 * not receive future remote event messages from this place
	 * instance accessor.
	 * <p>
	 * All listeners that equal the specified one will be
	 * unregistered. If the listener was not registered, the
	 * unregistration try is ignored.
	 * </p>
	 * <p>
	 * This method is specified by the <code>RemoteEventProducer</code>
	 * interface which is required by the <code>PlaceInstanceAccessor</code>
	 * interface.
	 * </p>
	 *
	 * @param listener the listener to unregister
	 * @exception java.rmi.RemoteException if a RMI failure occured.
	 */
	public void removeRemoteEventListener(RemoteEventListener listener) throws RemoteException
	{
		listeners.exclude(listener);

		// We unregister this accessor as PlaceEventListener
		// when removing the last remote event listener.
		if (listeners.isEmpty())
			((PlaceInstance) object).removePlaceEventListener(this);
	}

	/**
	 * This class does not use synchronous notifications, because
	 * it might lead to deadlocks for some applications. Asynchronous
	 * notifications are always safe, although they might arrive slightly
	 * later.
	 *
	 * @return a <code>boolean</code> value
	 */
	public boolean wantSynchronousNotification()
	{
		return false;
	}

	/**
	 * Converts the local place event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void markingChanged(PlaceEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * Converts the local place event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void tokenAdded(TokenEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * Converts the local place event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void tokenRemoved(TokenEvent event)
	{
		fireRemoteEvent();
	}


	/**
	 * Converts the local place event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void tokenTested(TokenEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * Converts the local place event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void tokenUntested(TokenEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * The common reaction to all place events is to send a
	 * remote event to all currently registered remote event
	 * listeners.
	 */
	private void fireRemoteEvent()
	{
		CollectionEnumeration listenerEnum;
		RemoteEventListener listener;

		synchronized (listeners)
		{
			listenerEnum = listeners.elements();
			while (listenerEnum.hasMoreElements())
			{
				listener = (RemoteEventListener) listenerEnum.nextElement();
				try
				{
					listener.update();
				}
				catch (RemoteException e)
				{
					System.err.println("PlaceInstanceAccessor: Remote event to " + listener
									   + " probably got lost due to " + e);
				}
			}
		}
	}
}
