package de.renew.remote;

import de.renew.application.ShadowSimulator;
import de.renew.shadow.*;
import de.renew.simulator.*;
import de.renew.util.ParameteredCommandLine;
import java.io.IOException;
import java.net.MalformedURLException;
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;


/**
 * An implementation of the server interface.
 * @see de.renew.remote.Server
 */
public class ServerImpl extends UnicastRemoteObject implements Server
{
	/**
	 * The current simulator of this server.
	 */
	private Simulator simulator;

	/**
	 * Creates a new server.
	 * @param simulatorClass The name of the simulator class to be run.
	 * @param netSystemFileName The file name of the shadow net system,
	 * as exported from the CPNApplication GUI.
	 * @param primaryNetName The name of the net to create the primary instance of.
	 * @param run Whether to run the simulation (otherwise init it only).
	 * @exception RemoteException An RMI problem occurred.
	 * @exception IllegalArgumentException A parameter is invalid.
	 * @exception IOException A read error occurred.
	 * @exception SyntaxException The shadow net system cannot be compiled.
	 */
	protected ServerImpl(String simulatorClassName, String netSystemFileName,
		String primaryNetName, boolean run)
		throws RemoteException, IllegalArgumentException, IOException, SyntaxException
	{
		super(0, SocketFactoryDeterminer.getInstance(), SocketFactoryDeterminer.getInstance());
		initialize(simulatorClassName, netSystemFileName, primaryNetName, run);
	}

	/**
	 * Creates, binds and returns a new server.
	 * @param line The parametered command line parsed by the main method.
	 * @return The new server.
	 * @exception If a problem occurs during creation or binding.
	 */
	protected static ServerImpl bindNewServer(ParameteredCommandLine line)
		throws Exception
	{
		String[] args = line.getRemainingArgs();

		ServerImpl server = new ServerImpl(args[0], args[1], args[2], !line.hasParameter("-i"));
		String serverUrl = getServerUrl(line);
		Naming.rebind(serverUrl, server);

		return server;
	}

	/**
	 * Searches for a server in the RMI registry
	 * of a given host and returns its.
	 * Use this method for the default server name.
	 * @param hostName The host of the RMI registry.
	 * @return The server.
	 * @exception RemoteException An RMI problem occurred.
	 * @exception NotBoundException The server cannot be found.
	 */
	public static synchronized Server getInstance(String hostName)
		throws RemoteException, NotBoundException
	{
		return getInstance(hostName, Server.DEFAULT_SERVER_NAME);
	}

	/**
	 * Searches for a server in the RMI registry
	 * of a given host and given name and returns its.
	 * Use this method if multiple servers exist.
	 * @param hostName The host of the RMI registry.
	 * @param serverName The name of the server.
	 * @return The server.
	 * @exception RemoteException An RMI problem occurred.
	 * @exception NotBoundException The server cannot be found.
	 */
	public static synchronized Server getInstance(String hostName, String serverName)
		throws RemoteException, NotBoundException
	{
		try
		{
			return (Server) Naming.lookup(makeServerUrl(hostName, serverName));
		}
		catch (MalformedURLException e)
		{
			throw new IllegalStateException("An internal error occurred: " + e);
		}
		catch (ClassCastException e)
		{
			throw new IllegalArgumentException(hostName + ":" + serverName + " is not a renew server");
		}
	}

	/**
	 * Returns the server URL determined from a given command line.
	 * @param line The parametered command line parsed by the main method.
	 * @return The server URL.
	 */
	protected static String getServerUrl(ParameteredCommandLine line)
	{
		return line.hasParameter("-s")
			? makeServerUrl("localhost", line.getParameterValues("-s")[0])
			: makeServerUrl("localhost");
	}

	/**
	 * Returns the current simulator of this server as an accessor.
	 * @return The simulator as an accessor.
	 * @exception RemoteException An RMI problem occurred.
	 */
	public SimulatorAccessor getSimulator()
		throws RemoteException
	{
		return new SimulatorAccessorImpl(simulator);
	}

	/**
	 * Initializes the server.
	 * @param simulatorClass The name of the simulator class to be run.
	 * @param netSystemFileName The file name of the shadow net system,
	 * as exported from the CPNApplication GUI.
	 * @param primaryNetName The name of the net to create the primary instance of.
	 * @param run Whether to run the simulation (otherwise init it only).
	 * @exception IllegalArgumentException A parameter is invalid.
	 * @exception IOException A read error occurred.
	 * @exception SyntaxException The shadow net system cannot be compiled.
	 */
	protected void initialize(String simulatorClassName, String netSystemFileName,
		String primaryNetName, boolean run)
		throws IllegalArgumentException, IOException, SyntaxException
	{
		simulator = ShadowSimulator.simulate(simulatorClassName, netSystemFileName,
			primaryNetName, SilentTracerFactory.INSTANCE, run);
	}

	/**
	 * Runs the server.
	 * @param args The runtime arguments.
	 * @exception Exception Any exception occurred while creating the server.
	 */
	public static void main(String[] args)
		throws Exception
	{
		showTitle();

		ParameteredCommandLine line = new ParameteredCommandLine(args,
			new String[] {"-h", "-s", "-i"}, new int[] {0, 1, 0});
		args = line.getRemainingArgs();

		if (args.length != 3 || line.hasParameter("-h"))
		{
			showSyntax();
			System.exit(0);
		}

		ServerImpl server = bindNewServer(line);
		server.runConsole();
		shutdown(server, getServerUrl(line));
	}

	/**
	 * Builds a default RMI server url.
	 * @param hostName The host name.
	 * @return The url.
	 */
	protected static String makeServerUrl(String hostName)
	{
		return "//" + hostName + "/" + Server.class.getName() + "." + Server.DEFAULT_SERVER_NAME;
	}

	/**
	 * Builds the a RMI server url out of a server name.
	 * @param hostName The host name.
	 * @param serverName The server name.
	 * @return The url.
	 */
	protected static String makeServerUrl(String hostName, String serverName)
	{
		return "//" + hostName + "/" + Server.class.getName() + "." + serverName;
	}

	/**
	 * Starts the server console.
	 */
	protected void runConsole()
	{
		System.out.println("Server now running...");
		System.out.println("Enter 'help' for a command help.");

		while (true)
		{
			System.out.print("\n> ");
			StringBuffer buffer = new StringBuffer();

			try
			{
				int ch;
				while ((ch = System.in.read()) >= 0 && ch != '\n')
				{
					if (ch >= ' ')
					{
						buffer.append((char) ch);
					}
				}
			}
			catch (IOException e)
			{
				e.printStackTrace(System.err);
				break;
			}

			String command = buffer.toString().trim().toLowerCase();

			if (command.startsWith("help"))
			{
				System.out.println("Server commands:");
				System.out.println("help             Displays this help.");
				System.out.println("run              Runs the simulation continuously.");
				System.out.println("step or <enter>  Executes only one simulation step.");
				System.out.println("stop             Stops the simulation, but it remains initialized.");
				System.out.println("term             Terminates the simulation.");
				System.out.println("exit or quit     Ends the server and exists.");
			}
			else if (command.startsWith("run"))
			{
				simulator.startRun();
			}
			else if (command.startsWith("step") || command.equals(""))
			{
				simulator.step();
			}
			else if (command.startsWith("stop"))
			{
				simulator.stopRun();
			}
			else if (command.startsWith("term"))
			{
				simulator.terminateRun();
			}
			else if (command.startsWith("exit") || command.startsWith("quit"))
			{
				break;
			}
			else
			{
				System.out.println("Error: Unknown command. Enter 'help' for a command help.");
			}
		}
	}

	/**
	 * Shows the command line syntax in the server console window.
	 */
	protected static void showSyntax()
	{
		System.out.println("Parameters: <simulator> <net system> <primary net> [-s <server name>] [-i]");
		System.out.println("  <simulator>      : The java class of the simulator to be used.");
		System.out.println("                     (e.g. de.renew.simulator.ConcurrentSimulator)");
		System.out.println("  <net system>     : The exported shadow net system file.");
		System.out.println("  <primary net>    : The name of the primary net to create an instance of.");
		System.out.println("  -s <server name> : The name of the server to be created.");
		System.out.println("                     If omitted, a default server name is used.");
		System.out.println("  -i               : Initialize the simulation only, don't run it.");
	}

	/**
	 * Shows the head line (title) in the server console window.
	 */
	protected static void showTitle()
	{
		System.out.println("The Renew RMI simulation server");
		System.out.println("");
	}

	/**
	 * Shuts the server down by terminating the simulation
	 * and unbinding the server from the RMI registry.
	 * @param server The server to be shutdown.
	 * @param serverUrl The url of the server.
	 */
	protected static void shutdown(ServerImpl server, String serverUrl)
	{
		System.out.println("\nServer shutting down...");
		server.terminateRun();
		try
		{
			Naming.unbind(serverUrl);
		}
		catch (Exception e)
		{
			e.printStackTrace(System.err);
		}
		System.out.println("Server shut down.");

		System.exit(0);
	}

	/**
	 * Terminates the simulator's run.
	 */
	protected void terminateRun()
	{
		simulator.terminateRun();
	}
}
