/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package de.renew.remote;

import java.rmi.*;
import java.rmi.server.*;
import de.renew.simulator.*;

/**
 * A simulator accessor allows an application to control
 * a remote simulator by starting and stopping the search process,
 * monitoring firing transitions, listing well-known net instances,
 * etc.
 *
 * A simulator accessor can be registered at the RMI registry as a
 * starting point for remote accesses.
 */

public class SimulatorAccessorImpl extends UnicastRemoteObject implements SimulatorAccessor
{
	final Simulator simulator;

	/**
	 * Constructor documentation.
	 * @param simulator
	 */
	public SimulatorAccessorImpl(Simulator simulator) throws RemoteException
	{
		super(0, SocketFactoryDeterminer.getInstance(), SocketFactoryDeterminer.getInstance());
		this.simulator = simulator;
	}

	/**
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public NetInstanceAccessor[] getNetInstances() throws RemoteException
	{
		NetInstance[] netInstances = NetInstanceList.getAll();
		NetInstanceAccessor[] result = new NetInstanceAccessor[netInstances.length];

		for (int i = 0; i < netInstances.length; i++)
		{
			result[i] = new NetInstanceAccessorImpl(netInstances[i]);
		}
		return result;
	}

	/**
	 * Query whether the simulator is still running.
	 *
	 * return true, if simulator could possibly make another step
	 * in the future.
	 */
	public boolean isActive() throws RemoteException
	{
		return simulator.isActive();
	}

	/**
	 * Start the simulator in the background. Return immediately.
	 */
	public void startRun() throws RemoteException
	{
		simulator.startRun();
	}

	/**
	 * Stop the simulator as soon as possible.
	 * Already firing transitions may continue to fire after
	 * this method has returned, but no new transitions will
	 * start firing.
	 */
	public void stopRun() throws RemoteException
	{
		simulator.stopRun();
	}

	/**
	 * Terminate the simulator once and for all.
	 * Do some final clean-up and exit all threads.
	 *
	 * Already firing transitions may continue to fire after
	 * this method has returned, but no new transitions will
	 * start firing nor is it allowed to restart the simulator
	 * by another call.
	 */
	public void terminateRun() throws RemoteException
	{
		simulator.terminateRun();
	}

	/**
	 * Try to perform one more step, then stop.
	 * Return a status code according
	 * to the five possibilities listed in the Simulator interface.
	 *
	 * @see de.renew.simulator.Simulator#statusStopped
	 * @see de.renew.simulator.Simulator#statusStepComplete
	 * @see de.renew.simulator.Simulator#statusLastComplete
	 * @see de.renew.simulator.Simulator#statusCurrentlyDisabled
	 * @see de.renew.simulator.Simulator#statusDisabled
	 *
	 * @return the status code after the possible step
	 */
	public int step() throws RemoteException
	{
		return simulator.step();
	}

	/**
	 * Make sure to refresh internal data structures after the
	 * firing of a transition outside the control of this simulator.
	 * This is be required by simulators that cache possible bindings.
	 */
	public void refresh() throws RemoteException
	{
		simulator.refresh();
	}

	/**
	 * Stop the simulator and return only after the last transition
	 * has completed firing.
	 */
	public void totallyStopSimulation() throws RemoteException
	{
		simulator.stopRun();
		BindingList.waitUntilEmpty();
	}
}
