package de.renew.remote;

import de.renew.simulator.*;
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.util.Enumeration;
import java.util.Vector;


/**
 * Implements the TokenCountsAccessor interface,
 * representing the token counts of a place instance as a snapshot.
 * Its contents will not change even if transitions fire.
 *
 * @author Thomas Jacob
 */
public class TokenCountsAccessorImpl extends UnicastRemoteObject implements TokenCountsAccessor
{
	/**
	 * The number of free tokens.
	 */
	private int freeTokenCount;

	/**
	 * The number of tested tokens.
	 */
	private int testedTokenCount;

	/**
	 * Creates a new TokenCountsAccessorImpl.
	 * @param placeInstance The place instance for the marking accessor.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public TokenCountsAccessorImpl(PlaceInstance placeInstance)
		throws RemoteException
	{
		super(0, SocketFactoryDeterminer.getInstance(), SocketFactoryDeterminer.getInstance());

		placeInstance.lock.lock();
		try {
			freeTokenCount = placeInstance.getNumberOfTokens();
			testedTokenCount = placeInstance.getNumberOfTestedTokens();
		} finally {
			placeInstance.lock.unlock();
		}
	}

	/**
	 * Return the number of free tokens in this marking.
	 * Even if two tokens are equal, they are both counted individually.
	 * Currently tested tokens are not included in this count.
	 *
	 * @return the number of free tokens
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public int getFreeTokenCount()
		throws RemoteException
	{
		return freeTokenCount;
	}

	/**
	 * Return the number of currently tested tokens in this marking.
	 *
	 * @return the number of tested tokens
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public int getTestedTokenCount()
		throws RemoteException
	{
		return testedTokenCount;
	}

	/**
	 * Returns whether the place instance has no tokens.
	 *
	 * @return Whether the place instance is empty.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public boolean isEmpty()
		throws RemoteException
	{
		return freeTokenCount + testedTokenCount <= 0;
	}
}
