/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package de.renew.remote;

import java.util.Collection;
import java.util.Iterator;
import collections.CollectionEnumeration;
import collections.HashedSet;
import collections.UpdatableSet;
import de.renew.event.FiringEvent;
import de.renew.event.TransitionEventListener;
import de.renew.simulator.*;
import java.rmi.*;
import java.rmi.server.*;


/**
 * This class implements the <code>TransitionInstanceAccessor</code>
 * interface and nothing more than needed to implement it.
 * <p>
 * </p>
 * TransitionInstanceAccessorImpl.java
 * Created: Mon Jul 17  2000
 * @author Michael Duvigneau
 */

public class TransitionInstanceAccessorImpl extends ObjectAccessorImpl
	implements TransitionInstanceAccessor, TransitionEventListener, Triggerable
{
	/**
	 * The set of registered remote event listeners that want
	 * to be informed about transition events.
	 */
	private UpdatableSet firingListeners = new HashedSet();

	/**
	 * The set of registered triggerable forwarders that want
	 * to be informed about search proposals.
	 */
	private UpdatableSet proposalListeners = new HashedSet();

	/**
	 * The triggers of this triggerable.
	 */
	private TriggerCollection triggers = new TriggerCollection(this);

	/**
	 * Creates a new transition instance accessor for the given transition
	 * instance.
	 *
	 * @param transitionInstance the transition instance to access
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public TransitionInstanceAccessorImpl(TransitionInstance transitionInstance)
		throws RemoteException
	{
		super(transitionInstance);
		transitionInstance.addTransitionEventListener(this);
	}

	/**
	 * Registers the given remote event listener so that it will
	 * receive future remote event messages concerning the accessed
	 * transition instance.
	 * <p>
	 * If the listener is registered already, the additional
	 * registration try is ignored. The <code>equals()</code>
	 * method serves as indicator to allow the mechanism to
	 * work also for remote objects.
	 * </p>
	 * <p>
	 * This method is specified by the <code>RemoteEventProducer</code>
	 * interface which is required by the <code>TransitionInstanceAccessor</code>
	 * interface.
	 * </p>
	 *
	 * @param listener the listener to register
	 * @exception java.rmi.RemoteException if a RMI failure occured.
	 */
	public void addRemoteEventListener(RemoteEventListener listener) throws RemoteException
	{
		// We register this accessor as TransitionEventListener
		// when adding the first remote event listener.
		if (firingListeners.isEmpty())
			((TransitionInstance) object).addTransitionEventListener(this);

		firingListeners.include(listener);
	}

	/**
	 * Returns all bindings for the transition instance
	 * and registeres a triggerable forwarder for search proposals.
	 * @param triggerableForwarder The triggerable forwarder.
	 * @return The bindings.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public BindingAccessor[] findAllBindings(TriggerableForwarder triggerableForwarder)
		throws RemoteException
	{
		synchronized (proposalListeners)
		{
			proposalListeners.include(triggerableForwarder);
		}

		BindingAccessor[] bindings;
		try
		{
			triggers.clear();
			Collection allBindings = SimulatorHelper.findAllBindings((TransitionInstance) object, this);
			Iterator iter=allBindings.iterator();

			bindings = new BindingAccessor[allBindings.size()];
			for (int i=0; iter.hasNext(); i++)
			{
				bindings[i] = new BindingAccessorImpl((Binding) iter.next());
			}
		}
		catch (RemoteException e)
		{
			e.printStackTrace(System.err);
			bindings = new BindingAccessor[0];
		}

		return bindings;
	}

	/**
	 * Fires one binding for the accessor's transition.
	 * @param tracer The tracer to be used.
	 * @param asynchronous Whether to perform the firing asynchronous.
	 * @return Whether a binding could be fired.
	 * @exception RemoteException If an RMI problem occurred.
	 */
	public boolean fireOneBinding(TracerAccessor tracer, boolean asynchronous)
		throws RemoteException
	{
		return SimulatorHelper.fireOneBinding(((TransitionInstance) object),
			new RemoteTracer(tracer), asynchronous);
	}

	/**
	 * The common reaction to all transition events is to send a
	 * remote event to all currently registered remote event
	 * listeners.
	 */
	private void fireRemoteEvent()
	{
		CollectionEnumeration listenerEnum;
		RemoteEventListener listener;

		synchronized (firingListeners)
		{
			listenerEnum = firingListeners.elements();
			while (listenerEnum.hasMoreElements())
			{
				listener = (RemoteEventListener) listenerEnum.nextElement();
				try
				{
					listener.update();
				}
				catch (RemoteException e)
				{
					System.err.println("TransitionInstanceAccessor: Remote event to "
						+ listener + " probably got lost due to " + e);
				}
			}
		}
	}

	/**
	 * Converts the local transition event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void firingComplete(FiringEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * Converts the local transition event into a remote event and
	 * forwards it to all remote listeners.
	 */
	public void firingStarted(FiringEvent event)
	{
		fireRemoteEvent();
	}

	/**
	 * This method is specified by the TransitionInstanceAccessor interface.
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public NetElementID getID() throws RemoteException
	{
		return ((TransitionInstance) object).getTransition().getID();
	}

	/**
	 * This method is specified by the TransitionInstanceAccessor interface.
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public TransitionAccessor getTransition() throws RemoteException
	{
		return new TransitionAccessorImpl(((TransitionInstance) object).getTransition());
	}

	/**
	 * Returns the transition instance, if the caller knows that
	 * this is the local representation. This is required for setting breakpoints.
	 * @return The transition instance.
	 */
	public TransitionInstance getTransitionInstance()
	{
		return ((TransitionInstance) object);
	}

	/**
	 * This method is specified by the TransitionInstanceAccessor interface.
	 * Method documentation.
	 * @return
	 * @throws RemoteException
	 */
	public NetInstanceAccessor getNetInstance() throws RemoteException
	{
		return new NetInstanceAccessorImpl(((TransitionInstance) object).getNetInstance());
	}

	/**
	 * Returns whether the transition instance is currently firing
	 * at least one occurrance.
	 * @return Whether it is firing.
	 * @exception java.rmi.RemoteException if an RMI failure occured.
	 */
	public boolean isFiring()
		throws RemoteException
	{
		return ((TransitionInstance) object).pendingFiringEvents().hasMoreElements();
	}

	/**
	 * Called when a trigger proposes a new search.
	 */
	public synchronized void proposeSearch()
	{
		TriggerableForwarder[] triggerableForwarders;
		synchronized (proposalListeners)
		{
			CollectionEnumeration enum = proposalListeners.elements();
			triggerableForwarders = new TriggerableForwarder[enum.numberOfRemainingElements()];
			for (int i=0; enum.hasMoreElements(); i++)
			{
				triggerableForwarders[i] = (TriggerableForwarder) enum.nextElement();
			}

			proposalListeners.clear();
		}

		for (int i=0; i<triggerableForwarders.length; i++)
		{
			try
			{
				triggerableForwarders[i].proposeSearch();
			}
			catch (RemoteException e)
			{
				System.err.println("TransitionInstanceAccessor: Search proposal to "
					+ triggerableForwarders[i] + " probably got lost due to " + e);
			}
		}
	}

	/**
	 * Unregisters the given remote event listener so that it will
	 * not receive future remote event messages from this transition
	 * instance accessor.
	 * <p>
	 * All listeners that equal the specified one will be
	 * unregistered. If the listener was not registered, the
	 * unregistration try is ignored.
	 * </p>
	 * <p>
	 * This method is specified by the <code>RemoteEventProducer</code>
	 * interface which is required by the <code>TransitionInstanceAccessor</code>
	 * interface.
	 * </p>
	 *
	 * @param listener the listener to unregister
	 * @exception java.rmi.RemoteException if a RMI failure occured.
	 */
	public void removeRemoteEventListener(RemoteEventListener listener) throws RemoteException
	{
		firingListeners.exclude(listener);

		// We unregister this accessor as TransitionEventListener
		// when removing the last remote event listener.
		if (firingListeners.isEmpty())
			((TransitionInstance) object).removeTransitionEventListener(this);
	}

	/**
	 * Returns the triggers of this triggerable.
	 * @return The triggers of this triggerable.
	 */
	public TriggerCollection triggers()
	{
		return triggers;
	}

	/**
	 * This class does not use synchronous notifications, because
	 * it might lead to deadlocks for some applications. Asynchronous
	 * notifications are always safe, although they might arrive slightly
	 * later.
	 *
	 * @return a <code>boolean</code> value
	 */
	public boolean wantSynchronousNotification()
	{
		return false;
	}
}
