package de.renew.shadow;

import de.renew.shadow.ShadowNetSystem;
import de.renew.shadow.ShadowNetLoader;
import de.renew.shadow.ShadowNet;
import de.renew.simulator.NetNotFoundException;

import de.renew.util.StringUtil;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.IOException;

import java.util.Enumeration;


/**
 * Tries to load shadow nets from <code>.sns</code> or
 * <code>.sha</code> files. The filename (without extension) must
 * match the net name. The loader looks for these files in the
 * directories specified by the <code>de.renew.netPath</code>
 * system property.
 * <p>
 * Additionally, the shadow net system stored in the file may
 * contain only one shadow net, which must have the same name as
 * the net and the file, too. This restriction exists to avoid
 * unexpected results from loading shadow nets out of files with
 * different names.
 * </p>
 * DefaultShadowNetLoader.java
 * Created: Thu Jan 17  2002
 * @author Michael Duvigneau
 **/
public class DefaultShadowNetLoader implements ShadowNetLoader  {

	private String[] netSource;

	private static boolean DEBUG = false;

	public DefaultShadowNetLoader (){
		setNetPath(System.getProperty("de.renew.netPath", System.getProperty("user.dir")));
		DEBUG = !("false".equalsIgnoreCase(System.getProperty("debug","false")));
	}

	/**
	 * Sets a search path (like the CLASSPATH) to look for shadow
	 * net system files when a net is missing. The directories in
	 * the path are separated by {@link File#pathSeparatorChar}.
	 **/
	void setNetPath(String path) {
    setNetPath(StringUtil.splitPaths(path));
	}

	/**
	 * Sets search paths (like the CLASSPATH) to look for net
	 * drawing files when a drawing is missing. Each String in the
	 * array denotes exactly one directory to search.
	 **/
	void setNetPath(String[] paths) {
		if (paths == null) {
			this.netSource = new String[0];
			return;
		}
		this.netSource = new String[paths.length];
    for (int i=0;i<paths.length;++i) {
			netSource[i]=StringUtil.makeCanonical(paths[i]);
      if (DEBUG) System.out.println("Shadow net loader source: "+netSource[i]);
    }
	}

	public ShadowNet loadShadowNet(String netName, ShadowNetSystem netSystem)
		throws NetNotFoundException
	{
		ShadowNet loaded = findShadowFile(netName);
		if (loaded != null) {
			loaded.switchNetSystem(netSystem);
			return loaded;
		}
		throw new NetNotFoundException(netName);
	}
	
	/**
	 * Tries to find and load a <code>.sns</code> or
	 * <code>.sha</code> file for the given net name.
	 * The shadow net is loaded and returned.
	 * It is not checked whether a shadow net with the given name
	 * is already known.
	 *
	 * @return the fresh shadow net, if one was loaded -
	 *   <code>null</code>, if no matching file could be found.
	 **/	
	ShadowNet findShadowFile(String name) {
		String path;
    File file;
		boolean error;
		for (int i=0; i<netSource.length; ++i) {
			error = false;
			path = netSource[i] + File.separator + name + ".sns";
			try {
				file = new File(path);
				error = !file.canRead();
				if (error) {
					path = netSource[i] + File.separator + name + ".sha";
					file = new File(path);
					error = !file.canRead();
				}
			} catch (Exception e) {
				System.err.println("DefaultShadowNetLoader: " + path
					+ " caused a " + e);
				error = true;
			}
			if (! error) {
				try {
					if (DEBUG) System.out.println("Loading shadow net from " + path + " ... ");
					FileInputStream stream=new FileInputStream(path);
					ObjectInput input=new ObjectInputStream(stream);
					ShadowNetSystem netSystem=(ShadowNetSystem)input.readObject();
					input.close();
					Enumeration nets = netSystem.elements();
					ShadowNet net;
					if (nets.hasMoreElements()) {
						net = (ShadowNet)nets.nextElement();
						if (nets.hasMoreElements()) {
							System.err.println("DefaultShadowNetLoader: " + path
								+ " contains more than one shadow net. Ignored.");
						} else if (net.getName().equals(name)) {
							return net;
						}
					}
					System.err.println("DefaultShadowNetLoader: " + path
						+ " was empty!? Ignored.");
				} catch (Exception e) {
					System.err.println(e);
					e.printStackTrace();
					error = true;
				}
			}
		}
		return null;
	}
}
