package de.renew.shadow;

import collections.*;
import java.util.Hashtable;

public class ShadowNetSystem implements java.io.Serializable {
	static final long serialVersionUID = 4619730346099695518L;

	private UpdatableSet nets;
	private ShadowCompiler compiler=null;
	private transient UpdatableSet compiledNets = new HashedSet();
	private transient UpdatableSet uncompiledNets = new HashedSet();
	private transient ShadowNetLoader netLoader = null;

	public ShadowNetSystem() {
		nets=new HashedSet();
	}

	public ShadowNetSystem(ShadowCompiler compiler) {
		this();
		setCompiler(compiler);
	}

	public ShadowNetSystem(ShadowCompiler compiler, ShadowNetLoader netLoader) {
		this(compiler);
		setNetLoader(netLoader);
	}

	public void setCompiler(ShadowCompiler compiler) {
		 this.compiler=compiler;
	}

	public ShadowCompiler getCompiler() {
		 return compiler;
	}

 /** 
	* Sets a shadow net loader to be asked when a net is missing
	* during compilation.
  **/
  public void setNetLoader(ShadowNetLoader netLoader) {
		this.netLoader = netLoader;
	}

	public ShadowNetLoader getNetLoader() {
		return netLoader;
	}

	// Add and remove nets. These methods should be called
	// indirectly via ShadowNet() and ShadowNet.discard(), so that
	// inconsistencies in the data structure cannot arise.
	protected void add(ShadowNet net) {
		nets.include(net);
		if (!compiledNets.includes(net)) {
			uncompiledNets.include(net);
		}
	}
	protected void remove(ShadowNet net) {
		nets.exclude(net);
		compiledNets.exclude(net);
		uncompiledNets.exclude(net);
	}

	public void markAsCompiled(ShadowNet net) {
		if (nets.includes(net)) {
			compiledNets.include(net);
			uncompiledNets.exclude(net);
		}
	}

	public ShadowLookup compile() throws SyntaxException {
		if (compiler != null) {
			return compiler.compile(this);
		} else {
			return null;
		}
	}

	public ShadowLookup compileMore() throws SyntaxException {
		if (compiler != null) {
			return compiler.compileMore(this);
		}
		return null;
	}

	/**
	 * Returns an enumeration of all nets in the system.
	 * The enumeration will refuse to return more elements if the
	 * shadow net system is modified.
	 **/
	public CollectionEnumeration elements() {
		return nets.elements();
	}

	/**
	 * Returns an enumeration of all nets in the system which are
	 * not marked as compiled yet. The enumeration is independent
	 * from future modifications of the shadow net system.
	 **/
	public CollectionEnumeration uncompiledElements() {
		return uncompiledNets.duplicate().elements();
	}

	/**
	 * Tells whether there are uncompiled nets in the system.
	 **/
	public boolean hasUncompiledElements() {
		return !uncompiledNets.isEmpty();
	}

  /**
   * Deserialization method, behaves like default readObject
   * method.
	 * <p>
	 * Reinitializes the transient fields:
	 * <ul>
	 * <li>  All nets are marked as uncompiled.
	 * </li>
	 * <li>  No net loader is known.
	 * </li>
	 * </ul></p>
   **/
  private void readObject(java.io.ObjectInputStream in)
		throws java.io.IOException, ClassNotFoundException
	{
		in.defaultReadObject();
		uncompiledNets = (UpdatableSet)nets.duplicate();
		compiledNets = new HashedSet();
  }
}
