package de.renew.simulator;

import de.renew.expression.*;
import collections.*;
import de.renew.unify.*;
import de.renew.util.*;

public class ArcOccurrence implements InscriptionOccurrence {
  PlaceInstance placeInstance;
  VariableMapper mapper;
  Variable tokenVar;
  Variable delayVar;
  Arc arc;

  // This is a complete and immutable variable that is used
  // whenever not time specification is explicitly given.
  static Variable theNullTimeVar=new Variable(new Value(new Double(0)),null);

  protected ArcOccurrence(Arc arc, VariableMapper mapper, NetInstance netInstance) {
    this.arc=arc;
    this.mapper=mapper;
    placeInstance=netInstance.getInstance(arc.place);
  }

  protected InhibitorArcBinder getInhibitorArcBinder(Variable tokenVar,
  						     PlaceInstance placeInstance) {
	return new InhibitorArcBinder(tokenVar,placeInstance);
  }

  public CollectionEnumeration makeBinders(Searcher searcher)
    throws Impossible 
  {
    // For the moment, arc expressions are always early.
    // We might make them late for output arcs, but this
    // might cause significant confusion. By introducing a
    // new variable and an action statement a late evaluation
    // can always be enforced within the model.
    tokenVar=new Variable(
      arc.tokenExpr.startEvaluation(mapper,searcher.recorder,
        searcher.calcChecker),
      searcher.recorder);

    if (!arc.isUntimedArc()) {
      delayVar=new Variable(
        arc.timeExpr.startEvaluation(mapper,searcher.recorder,
          searcher.calcChecker),
        searcher.recorder);
    }

    // Prepare a sequence of binders for this arc.
    UpdatableSeq binders=new LinkedList();

    if (arc.arcType==Arc.out) {
      // We must make sure that the variables associated with
      // an output arc are going to be computed before we declare
      // a transition enabled.
      searcher.calcChecker.addLateVariable(tokenVar,
        searcher.recorder);
      searcher.calcChecker.addLateVariable(delayVar,
        searcher.recorder);
    } else if (arc.arcType==Arc.inhibitor) {
      binders.insertFirst(getInhibitorArcBinder(tokenVar,placeInstance));
    } else {
      // If the arc expression is invertable, there is a good
      // reason to try to assign values to it.
      if (arc.tokenExpr.isInvertible()) {
	  binders.insertFirst(new ArcAssignBinder(tokenVar,placeInstance,
			      arc.isTestArc()));
      }

      // Input arcs and double arcs simply check the
      // existence of tokens during binding.
      if (arc.isTestArc()) {
	  binders.insertFirst(new TestArcBinder(tokenVar,placeInstance));
      } else if (arc.arcType==Arc.bothOT) {
	  binders.insertFirst(new InputArcBinder(tokenVar,
						 new Variable(new Integer(0),
							      null),
						 placeInstance));
	  searcher.calcChecker.addLateVariable(delayVar,
					       searcher.recorder);
      } else {
	  binders.insertFirst(new InputArcBinder(tokenVar,delayVar,
						 placeInstance));
      }
    }

    return binders.elements();
  }

  private double getDelay() {
    Object timeObj=delayVar.getValue();
    if (timeObj instanceof Value) {
      timeObj=((Value)timeObj).value;
    }
    return ((Number)timeObj).doubleValue();
  }

  public CollectionEnumeration makeExecutables(Copier copier,
    VariableMapper copiedMapper)
  {
    // There will be no undo past this point, so we need
    // not supply a state recorder. However, we must always
    // make a copy, because the current bindings will be
    // completely rolled back. Even when we make a getValue()
    // call to the old variable, that variable might contain
    // a tuple that is taken apart later on.
    Variable copiedToken=(Variable)copier.copy(tokenVar);

    // Depending on the type of the arc we must create different
    // executables.
    switch (arc.arcType) {
    case Arc.in:
      return new SingletonEnumeration(
        new InputArcExecutable(placeInstance,
          copiedToken.getValue(),getDelay(),
          false,arc.getTrace()));
    case Arc.test:
      TestArcExecutable tester=new TestArcExecutable(placeInstance,
	copiedToken.getValue(),false,arc.getTrace());
      return new PairEnumeration(tester,new UntestArcExecutable(tester));
    case Arc.out:
      // For an output arc, the time might be calculated later.
      Variable copiedDelay=(Variable)copier.copy(delayVar);

      return new SingletonEnumeration(
        new OutputArcExecutable(placeInstance,
          copiedToken,copiedDelay,arc.getTrace()));
    case Arc.both:
      return new PairEnumeration(
	new InputArcExecutable(placeInstance,
          copiedToken.getValue(),getDelay(),false,arc.getTrace()),
        new OutputArcExecutable(placeInstance,
          copiedToken,null,arc.getTrace()));
    case Arc.fastBoth:
      return new SingletonEnumeration(
        new InputArcExecutable(placeInstance,
          copiedToken.getValue(),getDelay(),true,arc.getTrace()));
    case Arc.fastTest:
      return new SingletonEnumeration(
        new TestArcExecutable(placeInstance,
          copiedToken.getValue(),true,arc.getTrace()));
    case Arc.inhibitor:
      return new SingletonEnumeration(
        new InhibitorExecutable(placeInstance,
          copiedToken.getValue(),arc.getTrace()));
    case Arc.bothOT:
      // For an output arc, the time might be calculated later.
      copiedDelay=(Variable)copier.copy(delayVar);

      return new PairEnumeration(
	new InputArcExecutable(placeInstance,
          copiedToken.getValue(),0,false,arc.getTrace()),
        new OutputArcExecutable(placeInstance,
          copiedToken,copiedDelay,arc.getTrace()));
    default:
      throw new RuntimeException("Bad arc type.");
    }
  }
}
