package de.renew.simulator;

import de.renew.unify.Impossible;
import collections.*;
import java.util.Hashtable;
import java.util.Enumeration;

import java.io.Serializable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.StreamCorruptedException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

public class Net implements Serializable {

	static final long serialVersionUID = 4424943946669800287L;

	private static Hashtable netsByName=new Hashtable();
	private static NetLoader netLoader=null;
	private static Object netLoaderLock = new Object();

	/**
	 * Returns the net object with the given name.
	 * If a net is not known, asks the net loader (see {@link
	 * #setNetLoader}) to load the net.
	 *
	 * @param 
	 *   
	 * @returns
	 *   the net object with the given name.
	 *   Never returns <code>null</code>
	 *
	 * @throws NetNotFoundException
	 *   if the net is not loaded and no loader was set or the
	 *   loader could not find the net.
	 **/
	public static Net forName(String name) throws NetNotFoundException {
		Net net=(Net)((Hashtable)netsByName).get(name);
		if (net==null) {
			if (netLoader == null) {
				throw new NetNotFoundException(name);
			} else {
				synchronized(netLoaderLock) {
					// Check again if the net is still unknown.
					// Things may have changed during the synchronization sleep...
					net=(Net)((Hashtable)netsByName).get(name);
					if (net==null) {
						net = netLoader.loadNet(name);
					}
				}
			}
		}
		return net;
	}

	/**
	 * Clears the mapping used by the {@link #forName} method.
	 **/
	public static void forgetAllNets() {
		netsByName.clear();
	}
	/**
	 * Sets the net loader. This request will be ignored when there
	 * is already a net loader registered or there are any known
	 * nets. You are safe if you call it immediately after {@link
	 * #forgetAllNets}...
	 **/
	public static void setNetLoader(NetLoader loader) {
		if ((netLoader == null) && netsByName.isEmpty()) {
			netLoader = loader;
		}
	}

	/**
	 * Gets the net loader.
	 **/
	public static NetLoader getNetLoader() {
		return netLoader;
	}
	/**
	 * Returns an enumeration of all known nets.
	 **/
	public static Enumeration allKnownNets() {
		return netsByName.elements();
	}

	UpdatableSet places=new HashedSet();
	UpdatableSet transitions=new HashedSet();

	String name=null;

	boolean earlyTokens=false;

	private int netCount=0;

	public Net() {
		name=null;
	}

	/**
	 * Creates an empty net body with the given name. Since the net
	 * is not functional yet, it is not announced to the map for
	 * the static {@link #forName} method. This has to be done
	 * separately after full construction of the net by using the
	 * {@link confirmCompilation} method.
	 **/
	public Net(String name) {
		this.name=name;
	}

	/**
	 * Confirms the end of the construction phase of the net. The
	 * net is included in the {@link #forName} lookup.
	 *
	 * @since Renew 1.6
	 **/
	public void confirmCompilation() {
		if (name!=null) {
			netsByName.put(name,this);
		}
	}

	public CollectionEnumeration places() {
		return places.elements();
	}

	public int placeCount() {
		return places.size();
	}

	public CollectionEnumeration transitions() {
		return transitions.elements();
	}

	public int transitionCount() {
		return transitions.size();
	}

	int makeNetNumber() {
		return ++netCount;
	}

	public String toString() {
		return name;
	}

	public String getName() {
		return name;
	}

	/**
	 * Created one instance of this net, but do not yet confirm
	 * its creation. Used for creating instances that may have to be
	 * discarded early.
	 *
	 * @return a newly allocated <code>NetInstance</code>
	 * @exception Impossible if an initial marking expression
	 *            evaluation fails.
	 */
	protected NetInstance makeInstance() throws Impossible {
		// Instead of creating a generic NetInstance, we might also
		// want to create a specialized NetInstance that supports
		// own methods and implements interfaces.
		return new NetInstanceImpl(this);
	}

	/**
	 * Unlike the method {@link #makeInstance},
	 * this publicly accessible
	 * method immediately confirms the creation of the
	 * generated net instance, so that external users do not need
	 * to know about confirmation issues.
	 *
	 * @return a newly allocated <code>NetInstance</code>
	 */
	public NetInstance buildInstance() {
		try {
			NetInstance instance=makeInstance();
			instance.createConfirmation(SilentTracer.INSTANCE);
			return instance;
		} catch (Impossible e) {
			throw new RuntimeException("Could not make net instance.");
		}
	}

	void add(Place place) {
		places.include(place);
	}
	void add(Transition transition) {
		transitions.include(transition);
	}
	void remove(Place place) {
		places.exclude(place);
	}
	void remove(Transition transition) {
		transitions.exclude(transition);
	}

	public void setEarlyTokens(boolean earlyTokens) {
		this.earlyTokens=earlyTokens;
	}

	/**
	 * Serialization method, behaves like default writeObject
	 * method except using the domain trace feature, if the
	 * output is a RenewObjectOutputStream.
	 * @see de.renew.util.RenewObjectOutputStream
	 **/
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		RenewObjectOutputStream rOut = null;
		if (out instanceof RenewObjectOutputStream)
			rOut = (RenewObjectOutputStream)out;
		if (rOut != null) rOut.beginDomain(this);
		out.defaultWriteObject();
		if (rOut != null) rOut.endDomain(this);
	}

	/**
	 * Deserialization method, behaves like default readObject
	 * method and makes the net known to <code>Net.forName()</code>.
	 **/
	private void readObject(java.io.ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		if (name != null) {
			netsByName.put(name, this);
		}
	}

	/**
	 * Writes all currently known <code>Net</code>s
	 * to the given stream. The written information 
	 * describes the static part of all compiled nets.
	 * <p>
	 * If the given <code>ObjectOutput</code> is a <code><b>
	 * de.renew.util.RenewObjectOutputStream</b></code>, its
	 * feature of cutting down the recursion depth by delaying
	 * the serialization of some fields will be used.
	 * </p>
	 *
	 * Added Apr 18 2000	Michael Duvigneau
	 *
	 * @param output target stream (see note about RenewObjectOutputStream)
	 * @see de.renew.util.RenewObjectOutputStream
	 **/
	public static void saveAllNets(ObjectOutput output) 
				 throws IOException {

		Enumeration saveList;
		RenewObjectOutputStream rOut = null;
		if (output instanceof RenewObjectOutputStream)
			rOut = (RenewObjectOutputStream)output;

		if (rOut != null) rOut.beginDomain(Net.class);

		// Save all currently known nets.
		output.writeInt(netsByName.size());
		saveList=netsByName.elements();
		while (saveList.hasMoreElements()) {
			output.writeObject(saveList.nextElement());
		}

		// If a RenewObjectOutputStream is used, write
		// all delayed fields NOW.
		if (rOut != null) rOut.writeDelayedObjects();

		if (rOut != null) rOut.endDomain(Net.class);
	}

	/**
	 * Reads nets stored by <code>saveAllInstances()</code>
	 * and adds the new <code>Net</code>s to the list
	 * of known nets by name.
	 * </p><p>
	 * If the given <code>ObjectInput</code> is a <code>
	 * de.renew.util.RenewObjectInputStream</code>, the
	 * neccessary steps to cover delayed serialization will
	 * be made.
	 * </p><p>
	 * The ObjectInputStream will be read using
	 * <code>de.renew.util.ClassSource</code> to provide
	 * its ability of reloading all user defined classes.
	 * </p>
	 *
	 * Added Apr 18 2000	Michael Duvigneau
	 *
	 * @param input source stream (see notes above)
	 * @see #saveAllNets
	 * @see de.renew.util.ClassSource
	 * @see de.renew.util.RenewObjectInputStream
	 **/
	public static void loadNets(ObjectInput input) 
				 throws IOException, ClassNotFoundException {

		// Read all stored Nets.
		int count = input.readInt();
		Net newNet;
		try {
			for (int i = 0; i < count; i++) {
				newNet=(Net)de.renew.util.ClassSource.readObject(input);
			}
		} catch (ClassCastException e) {
			throw new StreamCorruptedException (
								"Object other than Net found "
								+"when looking for nets: "
								+e.getMessage());
		}
		
		// If a RenewObjectInputStream is used, read
		// all delayed fields NOW.
		if (input instanceof RenewObjectInputStream)
			((RenewObjectInputStream)input).readDelayedObjects();
	}

	// ----------------------------------------------- ID handling ----

	/**
	 * Return the place of the net with the given ID.
	 * Return <code>null</code>, if no such place exists.
	 * <p></p>
	 * Added Mon Jul 17	 2000 by Michael Duvigneau
	 *
	 * @param id the place's ID
	 * @return a place or <code>null</code>
	 */
				public Place getPlaceWithID(NetElementID id)
				{

		boolean corrupted;
		CollectionEnumeration placeEnum;
		Place place;
		Place result = null;

		// Corruption may occur if the set of places changes
		// while copying data into the ID array.
		// As this situation should be very rare, the copying of
		// data is in a loop which will then be restarted.
		// Synchronization is not needed this way.
		// It is important that the first action is getting the
		// enumeration to cover the whole computation with its
		// corruption detection.

		do {
			placeEnum = places();
			corrupted = false;
			try {
				while (placeEnum.hasMoreElements() && (result==null)) {
					place = (Place)placeEnum.nextElement();
																				if (place.getID().equals(id))
																				{
																								result = place;
																				}
				}
			} catch (CorruptedEnumerationException e) {
				System.err.println("Net("+name+"): place enumeration corrupted, restarting."); 
				corrupted = true;
			}
		} while (corrupted);

		return result;
	}

	/**
	 * Return the transition of the net with the given ID.
	 * Return <code>null</code>, if no such place exists.
	 * <p></p>
	 * Added Mon Jul 17	 2000 by Michael Duvigneau
	 *
	 * @param id the transition's ID
	 * @return a transition or <code>null</code>
	 */
				public Transition getTransitionWithID(NetElementID id)
				{

		boolean corrupted;
		CollectionEnumeration transitionEnum;
		Transition transition;
		Transition result = null;

		// Corruption may occur if the set of transitions changes
		// while copying data into the ID array.
		// As this situation should be very rare, the copying of
		// data is in a loop which will then be restarted.
		// Synchronization is not needed this way.
		// It is important that the first action is getting the
		// enumeration to cover the whole computation with its
		// corruption detection.

		do {
			transitionEnum = transitions();
			corrupted = false;
			try {
				while (transitionEnum.hasMoreElements() && (result==null)) {
					transition = (Transition)transitionEnum.nextElement();
																				if (transition.getID().equals(id))
																				{
																								result = transition;
																				}
				}
			} catch (CorruptedEnumerationException e) {
				System.err.println("Net("+name+"): transition enumeration corrupted, restarting."); 
				corrupted = true;
			}
		} while (corrupted);

		return result;
	}

}

