package de.renew.simulator;

import java.io.PrintWriter;
import java.io.PrintStream;


/**
 * Thrown when trying to access a net by its name, but no
 * definition for the net with the specified name could be found.
 * <p>
 * Thrown by {@link Net#forName} or {@link NetLoader#loadNet}.
 * </p>
 * NetNotFoundException.java
 * Created: Tue Dec  4  2001
 * @author Michael Duvigneau
 **/
public class NetNotFoundException extends Exception{

	private Throwable chainedException = null;

	/**
	 * Constructs a <code>NetNotFoundException</code> with no
	 * detail message.
	 **/
	public NetNotFoundException () {
		super();
	}

	/**
	 * Constructs a <code>NetNotFoundException</code> with the 
	 * specified detail message. 
	 *
	 * @param message  the detail message. Typically the name of
	 *                 the missing net.
	 **/
	public NetNotFoundException (String message) {
		super(message);
	}

	/**
	 * Constructs a <code>NetNotFoundException</code> with the
	 * specified detail message and optional exception that was
	 * raised while loading the net.
	 *
	 * @param message  the detail message. Typically the name of
	 *                 the missing net.
	 * @param cause    the exception that was raised while loading
	 *                 the net.
	 **/
	public NetNotFoundException (String message, Throwable cause) {
		super(message);
		this.chainedException = cause;
	}

	/**
	 * Returns the exception that was raised if an error occurred while
	 * attempting to load the net. Otherwise, returns null.
	 **/
	public Throwable getException() {
		return chainedException;
	}


	/**
	 * Prints the stack backtrace. 
	 * 
	 * If an exception occurred during net loading it prints that
	 * exception's stack trace, or else prints the stack backtrace of
	 * this exception.
	 *
	 * @see java.lang.System#err
	 **/
	public void printStackTrace() { 
		printStackTrace(System.err);
	}
    
	/**
	 * Prints the stack backtrace to the specified print stream.
	 *
	 * If an exception occurred during net loading it prints that
	 * exception's stack trace, or else prints the stack backtrace of
	 * this exception.
	 **/
	public void printStackTrace(PrintStream ps) { 
		synchronized (ps) {
	    if (chainedException != null) {
				ps.println(getClass().getName()+": "+getMessage());
				ps.print("Nested exception is: ");
				chainedException.printStackTrace(ps);
	    } else {
				super.printStackTrace(ps);
	    }
		}
	}
    
	/**
	 * Prints the stack backtrace to the specified print writer.
	 *
	 * If an exception occurred during net loading it prints that
	 * exception's stack trace, or else prints the stack backtrace of
	 * this exception.
	 **/
	public void printStackTrace(PrintWriter pw) { 
		synchronized (pw) {
	    if (chainedException != null) {
				pw.println(getClass().getName()+": "+getMessage());
				pw.print("Nested exception is: ");
				chainedException.printStackTrace(pw);
	    } else {
				super.printStackTrace(pw);
	    }
		}
	}
}
