package de.renew.simulator;

import de.renew.database.TransactionSource;
import de.renew.unify.*;
import de.renew.event.ListenerSet;
import java.lang.reflect.InvocationTargetException;


public class SequentialSimulator
  implements Simulator, Runnable
{
  public static void main(String[] args) {
    if (args.length==0) {
      System.out.println(
        "Usage: java de.renew.simulator.SequentialSimulator [CLASS]");
      System.exit(0);
    }

    // Create a net.
    Net net=null;
    try {
      Class clazz=Class.forName(args[0]); // No reload required.
      net=(Net)clazz.getMethod("instance",
	new Class[0]).invoke(null,new Object[0]);
    }
    catch (Exception e) {
      System.out.println(args[0]+
        " isn't a valid net class as it seems.");
      if (e instanceof InvocationTargetException) {
	((InvocationTargetException)e).getTargetException().printStackTrace();
      }
      e.printStackTrace();
      System.exit(1);
    }

    NetInstance primaryInstance=SimulatorHelper.fireInitially(net,
      new ConsoleTracer("Initially","(0) "),true);

    if (primaryInstance!=null) {
      Simulator simulator=new SequentialSimulator();
      simulator.startRun();
    } else {
      System.out.println("The net did not accept the init request.");
    }
  }

  private long cycle=1L;

  private ExecuteFinder finder;
  private Searcher searcher=new Searcher();
  private boolean stepFired=false;
  private boolean isAlive=false;

  private boolean wantBreak=false;
  private Thread runThread=null;
  private TracerFactory tracerFactory;

	/**
	 * If true, the simulator will wait for the event queue
	 * before every firing.
	 **/
	private boolean wantEventQueueDelay;

	public SequentialSimulator()
	{
		this(new ConsoleTracerFactory(""),true);
	}

	public SequentialSimulator(TracerFactory tracerFactory,
														 boolean wantEventQueueDelay)
	{
		this.tracerFactory=tracerFactory;
		this.wantEventQueueDelay=wantEventQueueDelay;

		// If possible, prepare a new binding.
		findNewBinding();
	}

  public boolean isActive() {
    return isAlive;
  }

	public synchronized void startRun()
	{
		// Only create a new thread if none is currently running.
		if (runThread==null)
		{
			// Create a new thread that will run the simulation.
			runThread=new Thread(this);

			// Run the thread in the background.
			runThread.setPriority(Thread.MIN_PRIORITY);

			// Start the thread.
			runThread.start();

			try
			{
				TransactionSource.simulationStateChanged(true, true);
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
	}

	// Gently stop the simulation.
	// This may take a while, because the current search for
	// a activated binding must be completed first.
	// After the completion of this method, no thread is running
	// and runThread==null.
	public synchronized void stopRun()
	{
		while (runThread!=null)
		{
			wantBreak=true;
			try
			{
				wait();
			}
			catch (InterruptedException e)
			{
				// This is expected. But I cannot be sure that
				// I woke up from a notification.
			}
		}

		try
		{
			TransactionSource.simulationStateChanged(true, false);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	// Terminate the simulation once and for all.
	// Do some final clean-up and exit the thread.
	public synchronized void terminateRun()
	{
		// In this case we can simply call stop run, because that
		// is sufficient to terminate all running threads.
		stopRun();

		try
		{
			TransactionSource.simulationStateChanged(false, false);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

  private void fire() {
    Tracer tracer=tracerFactory.getTracer("Synchronously",
      String.valueOf(cycle++));
    stepFired=false;
    while (isAlive && !stepFired) {
			// Enforce flow control.
			if (wantEventQueueDelay) {
				ListenerSet.awaitEmptyQueue();
			}
			// Fire the binding.
      stepFired=finder.execute(tracer,false);
      findNewBinding();
    }
  }

  // To exclude concurrent access to the searcher, this method may only
  // be called from the running thread or from a
  // synchronized method that has ensured that no thread is running.
  // This method is not synchronized, because it could
  // lock up termination requests.
  // If there are no more enabled transitions, this method simply
  // returns without effect.
  private void findNewBinding() {
    // An ExecuteFinder should not be recycled.
    finder=new ExecuteFinder();
    // Let's look for another binding.
    searcher.searchLoop(finder);
    // Did we find a binding?
    isAlive=finder.isCompleted();
  }

  // This method must be synchronized to avoid concurrent
  // accesses to restart another thread after the stop.
  public synchronized int step() {
    if (runThread!=null) {
      // Just stop the current thread.
      stopRun();
      if (isAlive && !stepFired) {
	// The current thread did not make a step.
	fire();
      }
    } else {
      // Try to make another step.
      fire();
    }
    if (stepFired) {
      if (isAlive) {
	return statusStepComplete;
      } else {
	return statusLastComplete;
      }
    } else {
      return statusDisabled;
    }
  }

  // This method is called in threads that are generated on each
  // run request. The simulator class cannot inherit from thread,
  // because this method might be run multiple times.
  public void run() {
    // Do have a prepared binding?
    if (!isAlive) {
      // No, let's try to find one, although there is
      // little hope that we can find it.
      findNewBinding();
    }
    // Make steps until deadlock or until a break is requested.
    while (isAlive && !wantBreak) {
      // Grant other threads a chance to run.
      // Only needed on systems without preemptive multitasking.
      Thread.yield();
      // Fire and search for a new binding.
      fire();
    }
    // Make sure that runThread and wantBreak are updated
    // atomically.
    synchronized(this) {
      // Signal the end of the run.
      runThread=null;
      // Clear the break request.
      wantBreak=false;
      // Notify waiting threads of break.
      notifyAll();
    }
  }

  public synchronized void refresh() {
    if (runThread==null) {
      // No thread is currently running. Make sure to update
      // the enabled bindings and the time manually.
      findNewBinding();
    }
  }
}
