package de.renew.simulator;

import java.util.*;
import de.renew.unify.*;

public class SimulatorHelper {
  public static NetInstance fireInitially(Net net,
    Tracer tracer, boolean asynchronous)
  {
    // Search for the initial transition and fire it.
    // We make sure that the primary instance exists
    // only after a successful initial synchronisation.
    // We create a searcher especially for this purpose.
    // This is not overly efficient, but we do not expect
    // to fire a primary transition often.
    Searcher searcher=new Searcher();

    // Create the primary instance.
    NetInstance primaryInstance;
    try {
      primaryInstance=net.makeInstance();
    }
    catch (Impossible e) {
      primaryInstance=null;
    }

    // If the creation of the primary instance succeeded,
    // try to perform the initial synchronisation.
    ExecuteFinder finder=new ExecuteFinder();
    if (primaryInstance!=null) {
      // Fire the primary transition to get an
      // initial marking. The parameters could be
      // dependent on the parameters of this method.
      // No need to undo this action.
      Variable variable=new Variable(new Tuple(0),null);

      // Prepare the execution of the net. If no channel with
      // name "new" is present, simply ignore the synchronisation.
      searcher.initiatedSearch(primaryInstance,
	"new",variable,true,finder,null);

      // Did we find a binding?
      if (!finder.isCompleted()) {
	primaryInstance=null;
      }
    }

    // Execute the binding.
    if (primaryInstance!=null) {
      Runnable runnable=new InitialBindingRunnable(primaryInstance,
						   finder,tracer);
      de.renew.util.Detacher.possiblyStartAttached(runnable,asynchronous);
    }

    return primaryInstance;
  }

  // This method checks a searchable for enabledness.
  // A searcher must be provided for this method. The method
  // uses the given triggerable to store a new set of triggers.
  public static boolean isFirable(Searchable searchable,
    Triggerable triggerable, Searcher searcher)
  {
    EnablednessFinder finder=new EnablednessFinder();
    searcher.searchOnce(finder,searchable,triggerable);
    return finder.isEnabled();
  }

  // This method checks a searchable for enabledness. It
  // uses the given triggerable to store a new set of triggers.
  public static boolean isFirable(Searchable searchable,
    Triggerable triggerable)
  {
    Searcher searcher=new Searcher();
    return isFirable(searchable,triggerable,searcher);
  }

	// This method fires one binding. It resets the triggers
	// of the transition instance.
	public static boolean fireOneBinding(TransitionInstance instance,
		Tracer tracer, boolean asynchronous, Searcher searcher)
	{
		ExecuteFinder finder=new ExecuteFinder();
		searcher.searchOnce(finder,instance,instance);

		if (finder.isCompleted())
		{
			finder.execute(tracer,asynchronous);
			return true;
		}
		return false;
	}

  public static boolean fireOneBinding(TransitionInstance instance,
    Tracer tracer, boolean asynchronous)
  {
    Searcher searcher=new Searcher();
    return fireOneBinding(instance,tracer,asynchronous,searcher);
  }

  /**
   * This method collects all bindings for a searchable.
   * A searcher must be provided for this method. The method
   * uses the given triggerable to store a new set of triggers.
   * By using triggerables other than those associated with the
   * searchable, it is possible to keep more triggers than
   * with the searchable's triggers. This is required because the
   * searchable might only be interested in its own enabledness,
   * but not in all its bindings.
   *
   * @see de.renew.simulator.Binding
   *
   * @param searchable the searchable whose binding will be determined
   * @param triggerable the triggerable that get notified as soon as
   *   the search result becomes invalid
   * @param searcher the searcher that performs the search
   * @return an enumeration of bindings
   */
  public static Collection findAllBindings(Searchable searchable,
    Triggerable triggerable, Searcher searcher)
  {
    CollectingFinder finder=new CollectingFinder();
    searcher.searchOnce(finder,searchable,triggerable);
    return finder.bindings();
  }

  /**
   * This method collects all bindings for a searchable. It
   * uses the given triggerable to store a new set of triggers.
   */
  public static Collection findAllBindings(Searchable searchable,
    Triggerable triggerable)
  {
    Searcher searcher=new Searcher();
    return findAllBindings(searchable,triggerable,searcher);
  }
}
