package de.renew.simulator;

import collections.*;
import de.renew.util.Null;
import java.io.Serializable;

// I allow nulls to be stored in a place. I do this by
// converting nulls to and from an internal representation using
// the class Null.

public class TestTokenBag implements Serializable {
  // I prepare an array with small integers to avoid
  // creating excessive amounts of temporary integer objects.
  static final int maxStaticMult=15;
  static final Integer[] multiplicities;
  static {
    multiplicities=new Integer[maxStaticMult+1];
    for (int i=0;i<=maxStaticMult;i++) {
      multiplicities[i]=new Integer(i);
    }
  }

  private int size;
  private UpdatableMap testCount;
  private UpdatableMap orgTime;
  
  TestTokenBag() {
    testCount=new HashedMap();
    orgTime=new HashedMap();
    size = 0;
  }

  // Synchronize just to be safe. Maybe this is overly cautious.
  public synchronized int getUniqueSize() {
    return testCount.size();
  }

  public synchronized int getTestMultiplicity(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (testCount.includesKey(elem)) {
      return ((Integer)testCount.at(elem)).intValue();
    } else {
      return 0;
    }
  }

  private final static Integer toInteger(int result) {
    if (result<=maxStaticMult) {
      return multiplicities[result];
    } else {
      return new Integer(result);
    }
  }

  // Unfortunately, we have to make a copy of the
  // elements, because the hashtable might be updated
  // asynchronously.
  public synchronized CollectionEnumeration uniqueElements() {
    UpdatableSet elements=new HashedSet();
    elements.includeElements(testCount.keys());
    
    // Convert nulls to external representation.
    return Null.convertEnumeration(elements.elements());
  }

  public synchronized boolean includesTested(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    return testCount.includesKey(elem);
  }

  // A token bag should only be updated under control of its place.
  // Therefore this method is not made public.
  //
  // If this is supposed to change, a place instance may not
  // expose its token bag any longer.
  synchronized void addTested(Object elem, double time) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (testCount.includesKey(elem)) {
      Integer num=(Integer)testCount.at(elem);
      testCount.putAt(elem,toInteger(num.intValue()+1));
    } else {
      // We are testing the first token.
      // For this token, we record its time.
      testCount.putAt(elem,toInteger(1));
      orgTime.putAt(elem,new Double(time));
    }
    ++size;
  }

  // A token bag should only be updated under control of its place.
  // Therefore this method is not made public.
  //
  // If this is supposed to change, a place instance may not
  // expose its token bag any longer.
  synchronized double removeTested(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (testCount.includesKey(elem)) {
      int newMult=((Integer)testCount.at(elem)).intValue()-1;
      double time=((Double)orgTime.at(elem)).doubleValue();
      if (newMult==0) {
	testCount.removeAt(elem);
      } else {
	testCount.putAt(elem,toInteger(newMult));
      }
      --size;
      return time;
    } else {
      throw new RuntimeException("Negative number of tokens detected.");
    }
  }

  public String toString() {
    StringBuffer buffer = new StringBuffer();
    buffer.append("TestTokenBag(size: ");
    buffer.append(size);
    buffer.append("; count'token@time:");
    CollectionEnumeration enum = testCount.keys();
    while (enum.hasMoreElements()) {
      Object token = enum.nextElement();
      buffer.append(' ');
      buffer.append(testCount.at(token));
      buffer.append('\'');
      buffer.append(token);
      buffer.append('@');
      buffer.append(orgTime.at(token));
    }
    buffer.append(')');
    return buffer.toString();
  }
}
