package de.renew.simulator;

import collections.*;
import de.renew.util.Null;
import java.io.Serializable;
import java.util.NoSuchElementException;

// I allow nulls to be stored in a place. I do this by
// converting nulls to and from an internal representation using
// the class Null.

public class TokenBag implements Serializable {
  private int size;
  private UpdatableMap map;
  
  TokenBag() {
    map=new HashedMap();
    size = 0;
  }

  // Returns the delay until the availability of the tokens.
  // Positive infinity signals that the tokens will never
  // become available by simply waiting.
  public double computeEarliestTime(Object elem, TimeSet delays) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);
    if (map.includesKey(elem)) {
      double earliest=((TimeSet)map.at(elem)).computeEarliestTime(delays);
      return earliest;
    } else if (delays.getSize()==0) {
      // This should not happen, but we are always enabled.
      return 0;
    } else {
      return Double.POSITIVE_INFINITY;
    }
  }

  public synchronized TimeSet getTimeSet(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (map.includesKey(elem)) {
      return (TimeSet)map.at(elem);
    } else {
      return TimeSet.EMPTY;
    }
  }

  /**
   *  <code>getMultiplicity</code> returns the multiplicity of the
   * elem in the tokenBag.
   *
   * @param elem an <code>Object</code> value
   * @return an <code>int</code> value
   */
  public synchronized int getMultiplicity(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (map.includesKey(elem)) {
      return ((TimeSet)map.at(elem)).getSize();
    } else {
      return 0;
    }
  }

  // Unfortunately, we have to make a copy of the
  // elements, because the hashtable might be updated
  // asynchronously.
  public synchronized CollectionEnumeration uniqueElements() {
    UpdatableSet elements=new HashedSet();
    elements.includeElements(map.keys());
    
    // Convert nulls to external representation.
    return Null.convertEnumeration(elements.elements());
  }

  /**
   *  <code>includesAnytime</code> ignores timestamps on tokens.
   *
   * @param elem an <code>Object</code> value
   * @return a <code>boolean</code> value
   */
  public synchronized boolean includesAnytime(Object elem) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    return map.includesKey(elem);
  }

  public synchronized boolean includesBefore(Object elem, double time) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (!map.includesKey(elem)) {
	return false;
    }
    TimeSet times=(TimeSet)map.at(elem);
    return time>=times.earliestTime();
  }

  // A token bag should only be updated under control of its place.
  // Therefore this method is not made public.
  //
  // If this is supposed to change, a place instance may not
  // expose its token bag any longer.
  synchronized void add(Object elem, double time) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (map.includesKey(elem)) {
      TimeSet times=(TimeSet)map.at(elem);
      map.putAt(elem,times.including(time));
    } else {
      map.putAt(elem,TimeSet.make(time));
    }
    ++size;
  }

  // A token bag should only be updated under control of its place.
  // Therefore this method is not made public.
  //
  // If this is supposed to change, a place instance may not
  // expose its token bag any longer.
  synchronized void removeOneOf(Object elem, double time) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (map.includesKey(elem)) {
      TimeSet times=((TimeSet)map.at(elem)).excluding(time);
      if (times.isEmpty()) {
	map.removeAt(elem);
      } else {
	map.putAt(elem,times);
      }
      --size;
    } else {
      throw new RuntimeException("Negative number of tokens detected.");
    }
  }

  // A token bag should only be updated under control of its place.
  // Therefore this method is not made public.
  // 
  // This method returns the times stamp of the token that is actually
  // removed so that the token can be put back easily.
  synchronized double removeWithDelay(Object elem, double delay) {
    // Convert nulls to internal representation.
    elem=Null.convert(elem);

    if (map.includesKey(elem)) {
      TimeSet times=(TimeSet)map.at(elem);
      double time=times.latestWithDelay(delay,SearchQueue.getTime());
      times=times.excluding(time);
      if (times.isEmpty()) {
	map.removeAt(elem);
      } else {
	map.putAt(elem,times);
      }
      --size;
      return time;
    } else {
      throw new NoSuchElementException();
    }
  }

  public String toString() {
    StringBuffer buffer = new StringBuffer();
    buffer.append("TokenBag(size: ");
    buffer.append(size);
    buffer.append("; token@timeset:");
    CollectionEnumeration enum = map.keys();
    while (enum.hasMoreElements()) {
      Object token = enum.nextElement();
      buffer.append(' ');
      buffer.append(token);
      buffer.append('@');
      buffer.append(map.at(token));
    }
    buffer.append(')');
    return buffer.toString();
  }
}
