package de.renew.simulator;

import collections.*;
import de.renew.util.Null;

// A token reservation keeps track of the number of tokens
// that would have to be tested and removed from a single
// place instance.
//
// Accesses to objects of this class are unsynchronized.
// It is assumed that any access is guarded by a lock on the
// associated token reserver.

class TokenReservation {
  private PlaceInstance instance;
  private int reservations;

  private TestTokenBag testedTokens;
  private TokenBag removedTokenDelays;

  TokenReservation(PlaceInstance instance) {
    this.instance=instance;
    reservations=0;
    
    testedTokens=new TestTokenBag();
    removedTokenDelays=new TokenBag();
  }

  boolean isRemovable() {
    return reservations==0;
  }

  // Return the set of registered delays for the given token.
  // If there are also tests, but the tests are not
  // targeted at an already tested token, an addional zero delay is
  // included.
  private TimeSet getRemovedDelaySet(Object token, boolean isTested) {
    TimeSet delays=removedTokenDelays.getTimeSet(token);
    if (isTested && !instance.containsTestedToken(token)) {
      delays=delays.including(0);
    }
    return delays;
  }

  public double computeEarliestTime() {
    instance.lock.lock();
    double result=0;
    UpdatableSet tokens=new HashedSet();
    tokens.includeElements(
      Null.convertEnumeration(testedTokens.uniqueElements()));
    tokens.includeElements(
      Null.convertEnumeration(removedTokenDelays.uniqueElements()));
    CollectionEnumeration enum=Null.convertEnumeration(tokens.elements());
    while (enum.hasMoreElements()) {
      Object token=enum.nextElement();
      TimeSet delays=getRemovedDelaySet(token,
	testedTokens.getTestMultiplicity(token)>0);
      double time=instance.computeEarliestTime(token,delays);
      if (time>result) {
	result=time;
      }
    }
    instance.lock.unlock();
    return result;
  }

  public boolean containsRemovableToken(Object token, double delay) {
    instance.lock.lock();

    TimeSet delays=getRemovedDelaySet(token,
      testedTokens.getTestMultiplicity(token)>0).including(delay);
    double time=instance.computeEarliestTime(token,delays);

    instance.lock.unlock();
    return time<Double.POSITIVE_INFINITY;
  }

  public boolean containsTestableToken(Object token) {
    instance.lock.lock();

    TimeSet delays=getRemovedDelaySet(token,true);
    double time=instance.computeEarliestTime(token,delays);

    instance.lock.unlock();
    return time<Double.POSITIVE_INFINITY;
  }

  public boolean removeToken(Object token, double delay) {
    instance.lock.lock();

    boolean result=containsRemovableToken(token,delay);
    if (result) {
      removedTokenDelays.add(token,delay);
      reservations++;
    }

    instance.lock.unlock();
    return result;
  }

  public boolean testToken(Object token) {
    instance.lock.lock();

    boolean result=containsTestableToken(token);
    if (result) {
      // We can ignore the time. It is only used to
      // restore a previous state in the case of places.
      testedTokens.addTested(token,0);
      reservations++;
    }

    instance.lock.unlock();
    return result;
  }

  public void unremoveToken(Object token, double delay) {
    removedTokenDelays.removeOneOf(token,delay);
    reservations--;
  }

  public void untestToken(Object token) {
    testedTokens.removeTested(token);
    reservations--;
  }

  public CollectionEnumeration getRemovableTokens(Object pattern) {
    instance.lock.lock();

    CollectionEnumeration tokens=instance.getDistinctTokens(pattern);
    UpdatableSeq result=new LinkedList();
    while (tokens.hasMoreElements()) {
      Object token=tokens.nextElement();
      // The token must be removable sometimes, not neccessarily
      // now, because the arc might specify a negative time delay.
      if (containsRemovableToken(token,Double.NEGATIVE_INFINITY)) {
	result.insertFirst(Null.convert(token));
      }
    }

    instance.lock.unlock();
    return Null.convertEnumeration(result.elements());
  }

  public CollectionEnumeration getTestableTokens(Object pattern) {
    instance.lock.lock();

    CollectionEnumeration tokens=instance.getDistinctTestableTokens(pattern);
    UpdatableSeq result=new LinkedList();
    while (tokens.hasMoreElements()) {
      Object token=tokens.nextElement();
      if (containsTestableToken(token)) {
	result.insertFirst(Null.convert(token));
      }
    }

    instance.lock.unlock();
    return Null.convertEnumeration(result.elements());
  }
}

