package de.renew.util;

import java.io.Serializable;
import java.io.IOException;
import java.io.NotSerializableException;

// In essence this class duplicates the functionality of the
// synchronized statement. But locks can be used without
// a textual matching in the source code, which allows
// more flexible (and hence more dangerous) synchronisation
// schemes.

public class Lock implements Serializable {
  private transient Thread lockingThread;
  private transient int lockCount;

  public Lock() {
    lockingThread=null;
    lockCount=0;
  }

  public synchronized void lock() {
    Thread currentThread=Thread.currentThread();
    while (lockingThread!=null && lockingThread!=currentThread) {
      try {
	wait();
      } catch (InterruptedException e) {}
    }

    lockingThread=currentThread;
    lockCount++;
  }

  public synchronized void unlock() {
    verify();
    lockCount--;
    if (lockCount==0) {
      lockingThread=null;
    }
    notify();
  }

	public synchronized void verify()
	{
		if (Thread.currentThread() != lockingThread)
		{
			if (lockingThread == null)
			{
				throw new IllegalStateException("The lock is not locked.");
			}
			else
			{
				throw new IllegalStateException("The lock was locked by another thread.");
			}
		}
	}

  /**
   * Serialization method, behaves like default writeObject
   * method except checking additional error conditions.
   * Throws NotSerializableException if the lock is in use.
   **/
  private void writeObject(java.io.ObjectOutputStream out)
      throws IOException {
      synchronized(this) {
	  if ( (lockCount==0) && (lockingThread==null) ) {
	      out.defaultWriteObject();
	  } else {
	      throw new NotSerializableException(
                    "de.renew.util.Lock: "+this+" is in use.");
	  }
      }
  }

  /**
   * Deserialization method, behaves like default readObject
   * method except restoring additional transient fields.
   * Resets lockCount and lockingThread to null.
   **/
  private void readObject(java.io.ObjectInputStream in)
               throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    lockCount = 0;
    lockingThread = null;
  }
}
