package de.renew.util;

import collections.CircularList;
import java.util.NoSuchElementException;
import java.io.Serializable;



/**
 *  <code>Queue</code> implements a simple FIFO-Queue.
 *
 * @author <a href="mailto:roelke@informatik.uni-hamburg.de">Heiko Roelke</a>
 * @version 1.0
 * @since Renew 1.5
 */

public class Queue implements Serializable {
    static final long serialVersionUID = 3620540645530043407L;

    private CircularList circularList;

    public Queue() {
	circularList = new CircularList();
    }
    

    /**
     *  <code>enqueue</code> inserts an element at the queue's
     *  end.
     *
     * @param element an <code>Object</code> value
     */
    public void enqueue(Object element) {
      synchronized(circularList) {
	circularList.insertLast(element);
	circularList.notifyAll();
      }
    }

    /**
     *  <code>dequeue</code> returns oldest element and removes
     *  it from the queue.
     *
     * @return an <code>Object</code> value
     * @exception NoSuchElementException if queue is empty
     */
    public Object dequeue() throws NoSuchElementException {
      synchronized(circularList) {
	Object element = circularList.first(); 
	circularList.removeFirst();
	return element;
      }
    }

    /**
     *  <code>front</code> returns the front element
     *  <b>without</i> removing it from the queue.
     *
     * @return an <code>Object</code> value
     * @exception NoSuchElementException if queue is empty.
     */
    public Object front() throws NoSuchElementException {
	return circularList.first();
    }

    /**
     * Waits until the queue is non-empty, then returns the
     * oldest element and removes it from the queue.
     * <p>
     * If the queue contains an element in the moment this method
     * is called, it will return immediately (as {@link #dequeue}
     * would. But if the queue is empty, this method will block
     * until there is an element.
     * </p>
     * Added 25 Jun 2001 by Michael Duvigneau
     *
     * @return an <code>Object</code> value
     * @exception InterruptedException 
     *     if another thread has interrupted the current thread.
     */
    public Object waitAndDequeue() throws InterruptedException {
      synchronized(circularList) {
	while (circularList.isEmpty()) {
	  circularList.wait();
	}
	Object element = circularList.first(); 
	circularList.removeFirst();
	return element;
      }
    }

    /**
     * Tells whether the queue doesn't contain any elements.
     *
     * @return <code>true</code>  if the queue is empty, <br>
     *         <code>false</code> if there is at least one
     *                            element in the queue.
     */
    public boolean isEmpty() {
	return circularList.isEmpty();
    }
}
