package de.renew.util;

import collections.*;

/**
 * The Scheduler class allows to delay the execution of runnables.
 * Every runnable is executed in the scheduler thread.
 */
public class Scheduler implements Runnable {
    private static Scheduler instance;
    public static synchronized Scheduler instance() {
	if (instance==null) {
	    instance=new Scheduler();
	}
	return instance;
    }

    private Scheduler() {
	new Thread(this).start();
    }

    /**
     * A map from time stamps represented as longs
     * to pairs of time stamp and queued elements. This map is always
     * accessed by take and insert methods.
     */
    private RBMap queues=new RBMap();

    private SchedulerPair takeEarliest() {
	if (queues.isEmpty()) {
	    return null;
	} else {
	    return (SchedulerPair)queues.take();
	}
    }

    private SchedulerPair takePairAt(long timeMillis) {
	Long key=new Long(timeMillis);
	try {
	    SchedulerPair pair=(SchedulerPair)queues.at(key);
	    queues.removeAt(key);
	    return pair;
	} catch (java.util.NoSuchElementException e) {
	    return new SchedulerPair(key);
	}
    }

    /**
     * Reinsert the given scheduler pair if it contains
     * at least one runnable.
     *
     * @param pair the <code>SchedulerPair</code> to be inserted
     */
    private void putPair(SchedulerPair pair) {
	if (!pair.list.isEmpty()) {
	    queues.putAt(pair.key,pair);
	}
    }
    
    /**
     * Delays the execution of a runnable for a given period of time
     * A runnable that is executed may call this method to reinsert itself
     * or to schedule other runnables.
     *
     * @param runnable the <code>Runnable</code> to be executed
     * @param deltaMillis the number of milliseconds to delay the execution.
     */
    public synchronized void executeIn(Runnable runnable, long deltaMillis) {
	executeAt(runnable,System.currentTimeMillis()+deltaMillis);
    }

    /**
     * Delays the execution of a runnable until a given point of time
     * A runnable that is executed may call this method to reinsert itself
     * or to schedule other runnables.
     *
     * @param runnable the <code>Runnable</code> to be executed
     * @param timeMillis the time in milliseconds of desired the execution.
     */
    public synchronized void executeAt(Runnable runnable, long timeMillis) {
	SchedulerPair pair=takePairAt(timeMillis);
	pair.list.insertFirst(runnable);
	putPair(pair);
	// Make sure that the scheduler thread wakes up to process
	// the new runnnable, if it happens to be the earliest runnable.
	notifyAll();
    }

    public synchronized void run() {
	while (true) {
	    SchedulerPair pair=takeEarliest();
	    if (pair!=null) {
		long timeMillis=pair.key.longValue();
		long now=System.currentTimeMillis();
		if (timeMillis>now) {
		    // No runnables are ready yet.
		    putPair(pair);
		    try {
			wait(timeMillis-now);
		    } catch (InterruptedException e) {
			// This is expected.
		    }
		} else {
		    // Some runnables have become executable.
		    // After this point of time, we must not assume
		    // that the map is unchanged, because the
		    // invoked runnables might have inserted further
		    // items into the queue.
		    while (!pair.list.isEmpty()) {
			Runnable runnable=(Runnable)pair.list.take();
			runnable.run();
		    }
		    // No wait required. Check next time stamp immediately.
		}
	    } else {
		// No runnables are available.
		try {
		    wait();
		} catch (InterruptedException e) {
		    // This is expected.
		}
	    }
	}
    }
}
